/*
 * Copyright (c) 1997-2002 Akihiko YAMAMOTO (star@eos.hokudai.ac.jp)
 * All rights reserved.
 *
 * The user should define -DPW_CHECK, -DPW_CHECK_DNS or -DPW_CHECK_FILE
 * as a compiling flag to use the modules below.  See README file
 * for further explanations.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Specify your resolver library here (modify the next line). 
 * Or comment out the next line if you don't need it.
 * $Libraries: -lresolv$
 *
 */

/* 
 * Force to define PW_CHECK here to compile this module!
 *
 * If you  want to specify  this as  a compiling flag
 * (e.g.  PW_CHECK_FILE, PW_CHECK_DNS,...)  in "make"
 * stage, comment  out the following  three lines and
 * add your compiling flag to "CFLAGS"(e.g. -DPW_CHECK)
 * in Make.rules to suit your needs.   Note that this
 * module accepts following six flags.
 *
 * PW_CHECK            .......... compile all sources
 * PW_CHECK_FILE       .......... compile "file" check related sources
 * PW_CHECK_DNS        .......... compile "dns"  check related sources
 * PW_CHECK_DEBUG      .......... compile and set debugging flag to true
 * PW_CHECK_FILE_DEBUG .......... compile and set debugging flag to true
 * PW_CHECK_DNS_DEBUG  .......... compile and set debugging flag to true
 */
#ifndef PW_CHECK
#define PW_CHECK	/* By default, compile all sources */
#endif



/*
 * ======================================================
 * NOTHING BELOW THIS POINT SHOULD REQUIRE MANUAL EDITING
 * ======================================================
 */



#ifdef  PW_CHECK_DEBUG
#define PW_CHECK
#endif

#ifdef  PW_CHECK_FILE_DEBUG
#define PW_CHECK_FILE
#define PW_CHECK_DEBUG
#endif

#ifdef  PW_CHECK_DNS_DEBUG
#define PW_CHECK_DNS
#define PW_CHECK_DEBUG
#endif

#ifdef  PW_CHECK
#define PW_CHECK_DNS
#define PW_CHECK_FILE
#endif

#if defined(PW_CHECK_FILE) || defined(PW_CHECK_DNS)

/*
 * Password-check module for ProFTPD
 */
#ifndef lint
static char pck_id[] = "$Id: mod_pwcheck.c,v 1.3 2002/10/24 09:30:14 star Exp $";
static char pck_author_name[] = "Akihiko YAMAMOTO";
static char pck_author_addr[] = "star@eos.hokudai.ac.jp";
#endif /* lint */

#include "pwcheck.h"

PCK		pck;

extern int	core_display_file(const char *,const char *);
extern char*	fmt_time(time_t);

/* Prototypes */
void		pck_decompose_line(char [], char [], int);
char*		pck_extract_class(char *, int);
char*		pck_get_domain_name(char *);
int		pck_isdigitad(const char *);
int		pck_isdigitad2(const char *);
int		pck_istrrstr(int, char *, char *);
int		pck_istrstr (int, char *, char *);
int		pck_need_to_check(int valid,
				  char (*)[],
				  char *directive, 
				  char *logging_mark);
int		pck_password_check(int valid, 
				   char *passwd, 
				   char (*)[],
				   char *directive, 
				   char *logging_mark);
int		pck_proof_dns(char *, char *, char *, int);
int		pck_proof_domain_name(char *, char *, int, int);
int		pck_proof_file(char *, char *, char *, int);
char*		pck_remove_bracket(char *);
char*		pck_remove_control(char *);
char*		pck_remove_one_char(const char *, char);
void		pck_server_error(int, char *, char *, char *);
char*		pck_shift_class_list(char *, int);
int		pck_validate_password(char *, int, char *);
static char*	pck_rangematch(char *pattern, char test, int flags);
int		pck_fnmatch(char *pattern, char *string, int flags);
void		pck_reply(int, char *fmt,...);
void		pck_lreply(int, char *fmt,...);
void		pck_vreply(long flags, int n, char *fmt, va_list ap);
void		pck_perror_reply(int code, char *string);
void		pck_upper(char *s);
void		pck_pr_mesg(int msgcode, char *msgfile);
void		pck_msg_massage(const char *inbuf, 
				char *outbuf, 
				size_t outlen);
int		pck_lookup_name(char *name, char **ip);
int		pck_lookup_ip(char *ip, char **fqdn);
int		pck_check_name_for_ip(char *name, char *ip);
void		pck_init(struct conn_struc  *config, 
			 struct class_struc *class);
int		pck_has_reverse_dns();
int		pck_has_matching_dns();
int		pck_has_reverse_dns_class_kickout();
int		pck_has_reverse_dns_host_kickout();
int		pck_has_matching_dns_class_kickout();
int		pck_has_matching_dns_host_kickout();
int		pck_goaway_class_kickout();
int		pck_goaway_host_kickout();
int		pck_server_dead_class_kickout();
int		pck_server_dead_host_kickout();
void		pck_entrance();
int		pck_exit();
char**		pck_comma_split( char *text );
void		pck_block_out(char *directive,
			      char *logging_mark,
			      char *index);
void		pck_path_define(char *directive, const char *path);
void		pck_path_define_pck(char *directive, 
				    char dest[],
				    const char *path);
void		pck_override_define(char *directive, const char *pck_argv);
int		pck_core_handler(struct conn_struc  *config, 
				 struct class_struc *class,
				 config_rec *c);
void		pck_block_sub1(char *directive,const char **pck_buf2);
void		pck_block_sub2(char *directive,int check_flag);
void		pck_block_sub3(char *directive,
			       int check_flag,
			       char *logging_mark);
int		pck_override_display_file(const char *code, char *file);
int		pck_core_display_file(const char *numeric, const char *fn);
MODRET		pck_add_bool (cmd_rec * cmd);
MODRET		pck_add_bool_str (cmd_rec * cmd);
MODRET		pck_add_bool_str_ugly (cmd_rec * cmd);
MODRET		pck_add_bool_str2 (cmd_rec * cmd);
MODRET		pck_add_bool_str2_ugly (cmd_rec * cmd);
MODRET		pck_add_bool_str3 (cmd_rec * cmd);
MODRET		pck_add_bool_str3_ugly (cmd_rec * cmd);
MODRET		pck_add_str1 (cmd_rec * cmd);
MODRET		pck_add_str2 (cmd_rec * cmd);
MODRET		pck_add_str3 (cmd_rec * cmd);
MODRET		pck_add_str4 (cmd_rec * cmd);
MODRET		pck_add_str5 (cmd_rec * cmd);
MODRET		pck_add_str45 (cmd_rec * cmd);
MODRET		pck_command_handler(cmd_rec *command_rec);;
static void	pck_build_group_arrays(pool *pool, 
				       struct passwd *pw, 
				       char *name,
				       array_header **gids, 
				       array_header **groups);
static 
config_rec*	pck_anonymous_group(pool *pool, char *user);
config_rec*	pck_resolve_user(pool *pool, 
				 char **user, 
				 char **ournamep,
				 char **anonnamep);
void		pck_respond_grant_message(cmd_rec *command_rec);
void		pck_respond_deny_message(cmd_rec *command_rec);
char*		pck_make_filename( const char *first_part, ... );
char*		pck_stpcpy( char *a, const char *b );
char*		pck_get_home_dir(char *user_name);
char*		pck_get_system_dir();
char*		pck_expand_tilde(char *path);
void		pck_display_login_message();
int		pck_block_pck(config_rec *c,
			      char *directive, 
			      char *logging_mark,
			      char *index);
int		pck_dns_check(char only_domain[],
			      char *name_server,
			      int valid);
void		pck_init_checkmode();
void		pck_dump(PCK_PTR x);
void		pck_xfree(void *ptr);
void		pck_free();

char		check_mode[5][MAXPATHLEN];
char		pw_pwderrmsg2[BUFSIZ];
char		pck_buf0[BUFSIZ];
char		pck_buf1[BUFSIZ];
const char*	pck_buf2;
const char*	pck_buf3;
char		pck_temp[BUFSIZ];
int		no_more = 0;
char		hostname[MAXHOSTNAMELEN];
int		pw_check_debug;
extern char	ServerType;
int		_getshort(u_char *msg_ptr);
int		_getlong (u_char *msg_ptr);

#endif

/*
 * ########################
 * Modules for PW_CHECK_DNS
 * ########################
 */
#ifdef PW_CHECK_DNS

#include <arpa/nameser.h>
#include <resolv.h>

int
pck_proof_dns( char *passwd, char *arg , char *path, int valid )
{
  char only_domain[BUFSIZ];
  char *name_server;
  int file_flag;
  char *save;

  /*
   * return if already rejected
   */
  if ( !valid )
    return (valid);

  /*
   * return if NO_MORE_CHECK flag is set or CHECKED_ONCE_DNS flag is set
   */
  if ( (valid & NO_MORE_CHECK) || (valid & CHECKED_ONCE_DNS) )
    return (valid);

  file_flag  =  (valid & CHECKED_ONCE_FILE);

  /*
   * return with CHECKED_ONCE_DNS flag if CHECKED_ONCE_FILE flag is set
   * and REJECT flag is set
   */
  if (file_flag && !(valid & ACCEPT) )
    return (valid|CHECKED_ONCE_DNS);

  if ( pw_check_debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: pck_proof_dns executing...");

  /*
#ifdef HAVE_SYSINFO
  sysinfo(SI_HOSTNAME, hostname, sizeof(hostname));
#else
  (void) gethostname(hostname, sizeof(hostname));
#endif
  name_server = hostname;
  */

  (void) strcpy(hostname,pck.local.name);
  name_server = (char *)pck.local.name_server;

  if ( pw_check_debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: name_server = [%s]", name_server);

  /*
   * If mail address(password) is empty, we leave here
   * with RECORD_NOT_FOUND and CHECKED_ONCE_DNS flags.
   */
  (void) strcpy(only_domain,passwd);
  if ( !strlen(only_domain) ) {
    log_pri (L_ERROR, "ERR: Failed (NO ADDRESS): [%s]", only_domain);
    valid = (RECORD_NOT_FOUND|CHECKED_ONCE_DNS);
    valid |= file_flag;
    return (valid);
  }
  save = only_domain;

  if ( pw_check_debug & PCK_DEBUG_NORMAL )
    log_pri (L_TRACE, "---: Address(org) = [%s]", only_domain);

  if (arg && (!strcasecmp(arg, "remove")                  ||
              !strcasecmp(arg, "remove-control-bracket")  ||
              !strcasecmp(arg, "remove-bracket-control"))) {

    (void) strcpy(only_domain,pck_remove_control(only_domain));
    (void) strcpy(only_domain,pck_remove_bracket(only_domain));

  } else if (arg && (!strcasecmp(arg, "remove-control") ||
                     !strcasecmp(arg, "remove-cntrl"))) {

    (void) strcpy(only_domain,pck_remove_control(only_domain));

  } else if (arg && !strcasecmp(arg, "remove-bracket")) {

    (void) strcpy(only_domain,pck_remove_bracket(only_domain));

  }

  if ( pw_check_debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: Address(rem) = [%s]", only_domain);

  /*
   * Domain name part of the password is defined here.
   * If we cannot define legal domain name, we leave here
   * with RECORD_NOT_FOUND and CHECKED_ONCE_DNS flags.
   */
  (void) strcpy(only_domain,pck_get_domain_name(only_domain));
  if( !strlen(only_domain)) {
    log_pri (L_ERROR, "ERR: ADDRESS INVALID: [%s] <%s>", only_domain, passwd);
    valid = (RECORD_NOT_FOUND|CHECKED_ONCE_DNS);
    valid |= file_flag;
    return (valid);
  }

  if ( pw_check_debug & PCK_DEBUG_NORMAL )
    log_pri (L_TRACE, "---: Domain Name  = [%s]", only_domain);

  /*
   * DNS check ("A","MX") according to the order by PCKdnsorder.
   */
  valid = pck_dns_check(only_domain, name_server, valid);

  if ( !valid ) {
    log_pri (L_INFO,
	   "FTP LOGIN REFUSED: Bogus Domain [%s] (`A'/`MX' record NOT FOUND)", 
	   passwd);
  } else {
    log_pri (L_INFO,
           "ACCEPTED(pck_proof_dns): Address Proofed [%s]", passwd);
  }

  /*
   * we leave here with CHECKED_ONCE_DNS flag and original file_flag.
   */
  valid |= CHECKED_ONCE_DNS;
  valid |= file_flag;
  if ( pw_check_debug & PCK_DEBUG_FULL )
    log_pri (L_TRACE, "---: valid_3   = [%d]", valid    );
  return (valid);
}

int pck_dns_check(char only_domain[],
		  char *name_server,
		  int valid)
{

  config_rec	*pck_tmp;
  char		*temp;
  int		valid_save = 0;
  int		i,count    = 0;
  int		result_A   = 0;
  int		result_MX  = 0;
  int		max_check  = 2;
  const char	*done_A  = "NONE";
  const char	*done_MX = "NONE";

  /* save *original* valid */
  valid_save = valid;
  
  /*
   * If "PCKdnsorder" is included in "proftpd.conf",
   * each DNS check should be done in the specified order
   * by entries in "PCKdnsorder".
   *
   * If not, each check will be done according to the order
   * initialized by "pwcheck_child_init()"
   * (see pck.dns_order[][])
   *
   * (Inspired by a conversation with Masayuki FUKUI)
   */
  pck_tmp = find_config(main_server->conf,
			CONF_PARAM,
			"PCKdnsorder",
			TRUE);

  for( i=0; i<max_check; i++ ) {
    /*
     * When each entry begins with "!", we ignore this check.
     *
     * Example: (case-insensitive!)
     *
     * PCKdnsorder   MX  !A
     *
     */
    if( pck_tmp ) {
      temp = pck_tmp->argv[i];
      if( pck.debug & PCK_DEBUG_NORMAL )
	log_pri (L_TRACE, "---: DNS Check Entry [%s]",temp);
      if( *temp=='!' ) {
	count++;
	while( *temp=='!' ) temp++;
	if( pck.debug & PCK_DEBUG_FULL )
	  log_pri (L_TRACE, "---: !_match DNS CHECK NOT DONE (%s)...", temp);
	continue;
      }
    }

    /*
     * `pck_proof_domain_name' returns one of the two flags
     * < RECORD_FOUND / RECORD_NOT_FOUND > and
     * `pck_proof_dns' returns this result with CHECKED_ONCE_DNS flag
     * and file_flag.
     */   
    if( ( pck_tmp && !strcasecmp(pck_tmp->argv[i], "A")) ||
	(!pck_tmp && !strcasecmp(pck.dns_order[i], "A")) ) {

      /* This check should be done only *once* ! */
      if( strcasecmp(done_A, "NONE") ) {
	log_pri (L_INFO, "WRN: A_Record check already done!");
	continue;
      }
      done_A    = "DONE";
      result_A  = pck_proof_domain_name(only_domain, 
					name_server, 
					T_A,  
					result_MX );
      if ( pck.debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: Result_A   = [%d]", result_A );

    } else if( ( pck_tmp && !strcasecmp(pck_tmp->argv[i], "MX")) ||
	       (!pck_tmp && !strcasecmp(pck.dns_order[i], "MX")) ) {

      /* This check should be done only *once* ! */
      if( strcasecmp(done_MX, "NONE") ) {
	log_pri (L_INFO, "WRN: MX_Record check already done!");
	continue;
      }
      done_MX   = "DONE";
      result_MX = pck_proof_domain_name(only_domain, 
					name_server, 
					T_MX, 
					result_A );
      if ( pck.debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: Result_MX  = [%d]", result_MX );

    } else {
      /*
       * _ONLY_ two entries are allowed here.
       *
       * PCKdnsorder  MX  A
       *
       */
      if( pck_tmp ) {
	count++;
	log_pri (L_ERROR, "ERR: Invalid entry in PCKdnsorder? [%s]", 
	       (char *)pck_tmp->argv[i]);
      } else {
	/* NOTREACHED */
	count++;
	log_pri (L_ERROR, "ERR: Invalid entry in dns_order? [%s]", 
	       pck.dns_order[i]);
      }
    }
  }
  
  if ( pw_check_debug & PCK_DEBUG_FULL )
    log_pri (L_TRACE, "--: valid_1    = [%d]", valid);
  
  /* Set "valid" according to the result by "MX" and "A" checks */
  valid = (result_A|result_MX);
  
  if ( pw_check_debug & PCK_DEBUG_FULL )
    log_pri (L_TRACE, "---: valid_2    = [%d]", valid);
  
  if( !(max_check - count) ) {
    log_pri (L_ERROR, "ERR: No DNS check Done? (check your PCKdnsorder)");
    return valid_save;
  } else {
    return valid;	/* return with DNS check result! */
  }
  
}

int
pck_proof_domain_name( char *domain_name,
		       char *name_server,
		       int query_type,
		       int valid)
{

  union {
    HEADER hdr;
    u_char buf[PACKETSZ];
  } query, response;

  int query_buffer;
  int response_buffer;
  int ns_count;
  int i;
  struct in_addr addr_org[MAXNS];
  struct hostent *host;
  char *qtype = NULL;
  char real_server[MAXHOSTNAMELEN];

  /*
   * return if DNS record already found
   */
  if ( valid )
    return (valid);

  if ( pw_check_debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: pck_proof_domain_name executing...");

  if ( query_type == T_A  ) qtype = "Proof__A";
  if ( query_type == T_MX ) qtype = "Proof_MX";

  (void) res_init();

  /*
   * Save the original information from _res structure.
   */
  ns_count = _res.nscount;
  /*
   * Default name server not found?
   */
  if ( ns_count == 0 ) {
    log_pri (L_ERROR, "ERR: [%s] No Default Server? (nscount=0)\n",qtype);
    (void) res_init();
    return(RECORD_FOUND);
  }
  
  if ( pw_check_debug & PCK_DEBUG_FULL )
    log_pri (L_TRACE, "---: ns_count = %d (%s)\n", ns_count, qtype);

  /*
   * Save original information from _res structure.
   */
  for(i = 0; i < ns_count; i++) {
    if ( pw_check_debug & PCK_DEBUG_FULL ) {
      log_pri (L_TRACE, "---: nsaddr_list[%d] = [%s]\n", i,
	     inet_ntoa(_res.nsaddr_list[i].sin_addr));
    }
    addr_org[i] = _res.nsaddr_list[i].sin_addr;
  }

  /*
  _res.options |=  RES_DEBUG;
   */

  /*
   * Turn on recursion-desired bit in queries.
   * res_send do iterative queries here and
   * expects the name server to handle recursion.
   */
  _res.options |=  RES_RECURSE;

  /*
   * Turn off RES_DNSRCH bit in queries.
   * res_search will not search for host names
   * in the current domain and in parent domains.
   */
  _res.options &= ~RES_DNSRCH;

  /*
   * Turn off RES_DEFNAMES bit in queries.
   * res_search will not append the default domain name
   * to single-component names (those that do not contain a dot).
   */
  _res.options &= ~RES_DEFNAMES;

  /*
   * number of retries.
   */
  _res.retry = 2;

  /*
   * Restore from original information.
   */
  _res.nscount = ns_count;

  for(i = 0; i < ns_count; i++)
    _res.nsaddr_list[i].sin_addr = addr_org[i];

  /*
   * We define the name server for DNS queries.
   * Actual query server should be re-defined in 
   * `_res.nsaddr_list[0].sin_addr'.
   */

  /*
   * We use the localhost as a name server.
   */
  if ( !strcasecmp(name_server, "localhost") ) {

    if ((host = gethostbyname(hostname)) == NULL ) {
      log_pri (L_ERROR, "ERR: no address found for [%s]", hostname);
      (void) res_init();
      return(RECORD_FOUND);
    }
    (void) memcpy((void *)&_res.nsaddr_list[0].sin_addr,
		  (void *)host->h_addr_list[0],
		  (size_t)host->h_length);
    (void) strcpy(real_server, hostname);

    /*
     * We use a host listed in resolv.conf as a name server.
     */
  } else if ( !strcasecmp(name_server, "resolver") ) {

    _res.nsaddr_list[0].sin_addr = addr_org[0];
    if (( host = gethostbyaddr((char *) &_res.nsaddr_list[0].sin_addr,
			       sizeof (struct in_addr),
			       AF_INET)) == NULL ) {
      log_pri (L_ERROR, "ERR: no FQDN found for [%s]",
	     inet_ntoa(_res.nsaddr_list[0].sin_addr));
      (void) res_init();
      return(RECORD_FOUND);
    } else {
      (void) strcpy(real_server, host->h_name);
    }

    /*
     * We use a special host as a name server which is defined
     * in `PCKnameserver' parameter in proftpd.conf.
     */
  } else {

    if ((host = gethostbyname(name_server)) == NULL ) {
      log_pri (L_ERROR, "ERR: no address found for [%s]", name_server);
      (void) res_init();
      return(RECORD_FOUND);
    }
    (void) memcpy((void *)&_res.nsaddr_list[0].sin_addr,
                  (void *)host->h_addr_list[0],
                  (size_t)host->h_length);
    (void) strcpy(real_server, name_server);

  }

  if ( pw_check_debug & PCK_DEBUG_HIGH ) {
    log_pri (L_TRACE, "---: query_server   = [%s]\n", real_server);
  }

  if ( pw_check_debug & PCK_DEBUG_FULL ) {
    log_pri (L_TRACE, "---: proftpd.conf   = [%s]\n", name_server);
    log_pri (L_TRACE, "---: nsaddr_list[0] = [%s]\n",
	   inet_ntoa(_res.nsaddr_list[0].sin_addr));
  }

  /*
   * We construct a standard query message and place it in query_buffer.
   * It returns the size of the query, or -1 if the query is larger
   * than buffer length.
   */
  _res.nscount = 1;
  if (( query_buffer = res_mkquery ( QUERY,
				     domain_name,
				     C_IN,
				     query_type,
				     (const u_char *)NULL,
				     0,
				     (const u_char *)NULL,
				     (u_char *)&query,
				     sizeof(query))) < 0 ) {
    /*
     * If address is invalid, we leave here with
     * RECORD_NOT_FOUND flag.
     */
    log_pri (L_INFO, "DNS: ADDRESS Invalid: %s [%s] (%s)",
	   real_server, qtype, domain_name);
    (void) res_init();
    return(RECORD_NOT_FOUND);
  }

  /*
   * We send a pre-formatted query packet and get an answer here.
   * Error flag is set on `error' parameter and this flag should be
   * initialized before query.
   */
  errno = 0;
  if (( response_buffer = res_send( (const u_char *)&query,
				    query_buffer,
				    (u_char *)&response,
				    sizeof(response))) < 0 ) {
    /*
     * We found name server is not running on host which is
     * specified by `name-server' parameter.
     * Note that we leave here with RECORD_FOUND flag
     * to enable user's anonymous login.
     */
    if(errno == ECONNREFUSED) {
      log_pri (L_ERROR, "ERR: NameServer is not running?: %s [%s] (%s)",
	     real_server, qtype, domain_name);
      /*
       * Name server did not give us any responses.
       * Note that we leave here with RECORD_FOUND flag
       * to enable user's anonymous login.
       */
    } else {
      log_pri (L_ERROR, "ERR: No response from NameServer?: %s [%s] (%s)",
	     real_server, qtype, domain_name);
    }
    (void) res_init();
    /*
     * We should turn on RECORD_FOUND flag here because we should
     * put this trouble down to a system(name server).
     */
    return(RECORD_FOUND);
  }

  /*
   * If the answer includes some errors, we leave here with
   * special error messages defined in the `pck_server_error' module.
   * Note that RECORD_NOT_FOUND flag is set to disable user's login.
   */
  if(response.hdr.rcode != NOERROR){
    (void) pck_server_error( (int)response.hdr.rcode,
			     real_server,
			     domain_name,
			     qtype);
    (void) res_init();
    return(RECORD_NOT_FOUND);
  }
  
  /*
   * How many answers did we get from query server?
   * If the number of answers is zero, we leave here with
   * RECORD_NOT_FOUND flag and disable login.
   */
  if(ntohs(response.hdr.ancount) == 0){

    log_pri (L_INFO, "DNS: [%s] NOT Found (%s)", qtype, domain_name);
    (void) res_init();
    return(RECORD_NOT_FOUND);

    /*
     * How many answers did we get from query server?
     * If we got more than one answer, we simply say
     * how many records we found and get no action.
     */
  } else if(ntohs(response.hdr.ancount) > 1){

    log_pri (L_INFO,
	   "DNS: more than 1 record [%d] found.",
	   ntohs(response.hdr.ancount));

  }

  log_pri (L_INFO, "DNS: [%s] Found (%s)", qtype, domain_name);
  (void) res_init();
  return(RECORD_FOUND);

}

char *
pck_get_domain_name(char *address)
{
  int user, addrlen;

  addrlen  = strlen(address);
  user = pck_istrstr (0, address, "@") + 1;
  /*printf("addrlen,user [%d] [%d] \n",addrlen,user);*/

  if ( (user-1) == addrlen || user == addrlen) {
    *address = '\0';
    return(address);
  }

  while(user--) address++;
  return (address);
}

void
pck_server_error( int return_code,
		  char *name_server,
		  char *domain_name,
		  char *message)
{
  switch(return_code){
  case FORMERR:
    log_pri (L_INFO, "Bogus: FORMERR (%s) [%s]",  message, domain_name);
    break;
  case SERVFAIL:
    log_pri (L_INFO, "Bogus: SERVFAIL (%s) [%s]", message, domain_name);
    break;
  case NXDOMAIN:
    log_pri (L_INFO, "Bogus: NXDOMAIN (%s) [%s]", message, domain_name);
    break;
  case NOTIMP:
    log_pri (L_INFO, "Bogus: NOTIMP (%s) [%s]",   message, domain_name);
    break;
  case REFUSED:
    log_pri (L_INFO, "Bogus: REFUSED (%s) [%s]",  message, domain_name);
    break;
  default:
    log_pri (L_INFO, "Bogus: Unexpected (%s) [%s]", message, domain_name);
    log_pri (L_INFO, "Bogus: Unexpected Return Code = [%d]", return_code);
    break;
  }
}

#endif /* end of PW_CHECK_DNS */

/*
 * #########################
 * Modules for PW_CHECK_FILE
 * #########################
 */
#ifdef PW_CHECK_FILE

int
pck_proof_file( char *passwd, char *arg, char *path, int valid )
{
  int dns_flag;

  /*
   * return if already rejected
   */
  if ( !valid )
    return (valid);

  /*
   * return if NO_MORE_CHECK flag is set or CHECKED_ONCE_FILE flag is set
   */
  if ( (valid & NO_MORE_CHECK) || (valid & CHECKED_ONCE_FILE) )
    return (valid);

  dns_flag  =  (valid & CHECKED_ONCE_DNS);

  /*
   * return with CHECKED_ONCE_FILE flag if CHECKED_ONCE_DNS flag is set
   * and REJECT flag is set
   */
  if (dns_flag && !(valid & ACCEPT) )
    return (valid|CHECKED_ONCE_FILE);

  if ( pw_check_debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: pck_proof_file executing...");

  if (arg && (!strcasecmp(arg, "remove")                  ||
	      !strcasecmp(arg, "remove-control-bracket")  ||

	      !strcasecmp(arg, "remove-bracket-control"))) {

    valid = pck_validate_password(passwd,REMOVE_CONTROL|REMOVE_BRACKET,path);

  } else if (arg && (!strcasecmp(arg, "remove-control") ||
		     !strcasecmp(arg, "remove-cntrl"))) {

    valid = pck_validate_password(passwd,REMOVE_CONTROL, path);

  } else if (arg && !strcasecmp(arg, "remove-bracket")) {

    valid = pck_validate_password(passwd,REMOVE_BRACKET, path);

  } else {

    valid = pck_validate_password(passwd,DONT_REMOVE, path);

  }

  valid |= dns_flag;
  return (valid);

}

int
pck_validate_password(char *passwd, int flags, char *path_invalidpw)
{

  register FILE *fd;
  register char *p;
  char line[BUFSIZ];
  char *passwducase, *passwd_save;
  char *lineucase, *line_save;
  char *linetemp;
  char *save1, *save2, *save3, *save4;
  char *pw_invalidpw = NULL;
  char *dot_check = NULL;
  int flag_exclude = NONE_DEFINED;
  int case_sensitive = 0;

  if ( pw_check_debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: pck_validate_password executing...");

  if ((passwducase = (char *)malloc((unsigned)(strlen(passwd)+1))) == NULL) {
    pck_perror_reply(421, "Local resource failure (passwducase): malloc");
    return (ACCEPT|CHECKED_ONCE_FILE);
  }
  save1 = passwducase;

  if ((passwd_save = (char *)malloc((unsigned)(strlen(passwd)+1))) == NULL) {
    pck_perror_reply(421, "Local resource failure (passwd_save): malloc");
    return (ACCEPT|CHECKED_ONCE_FILE);
  }
  save2 = passwd_save;

  /*
   * Save `passwd' (case-insensitive)
   * This `passwducase' should be used in upper-case.
   */
  (void) strcpy(passwducase, passwd);

  /*
   * Save `passwd' (case-sensitive)
   */
  (void) strcpy(passwd_save, passwd);

  if ( (flags & REMOVE_CONTROL) ) {
    (void) strcpy(passwducase,pck_remove_control(passwducase));
    (void) pck_upper(passwducase);
    (void) strcpy(passwd_save,pck_remove_control(passwd_save));
  }

  if ( (flags & REMOVE_BRACKET) ) {
    (void) strcpy(passwducase,pck_remove_bracket(passwducase));
    (void) pck_upper(passwducase);
    (void) strcpy(passwd_save,pck_remove_bracket(passwd_save));
  }

  /* Change case-insensitive passwd into upper case */
  if ( !(flags & REMOVE_CONTROL) && !(flags & REMOVE_BRACKET) ){
    (void) pck_upper(passwducase);
  }

  if ( pw_check_debug & PCK_DEBUG_FULL )
    log_pri (L_TRACE, "---: PWD u_case ul_save [%s] [%s]",
	   passwducase, passwd_save);

  /*
   * We eliminate the leading `-' signs from user's password.
   */
  while (*passwducase == '-') { passwducase++; passwd_save++; }

  if( (flag_exclude & NONE_DEFINED) ) flag_exclude = NONE_DEFINED;

  /*
   * We don't want to accept any password without `@', `.'
   * as an email address.   Can we ignore these passwords?
   *
   * The `PCKdotcheck true' entry in proftpd.conf enables us to check to see
   * if user's password has `@' and `.' signs, before we read the regular
   * expression file (ftpinvalid file). Unless you set `dot-check do' in
   * proftpd.conf, I strongly recommend you to add the expression `!*@*[.]*'
   * as a last entry of ftpinvalid file.
   */
  /*
   * Default value of `dot-check' is `do'!!
   * for backward compatibility.
   */
  if (pck.dot_check) {
    dot_check = "do";
  } else {
    dot_check = "none";
  }

  if ( pw_check_debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: dot_check [%s]", dot_check);

  if ( !strcasecmp(dot_check, "do" ) || !strcasecmp(dot_check, "yes" ) ) {
    if (strchr(passwducase, '.') == NULL ||
	strchr(passwducase, '@') == NULL) {
      if(!no_more) {
	/*
	 * If no-more checking is not specified, we leave here with
	 * REJECT flag and refuse anonymous login immediately.
	 */
	log_pri (L_INFO,
	       "FTP LOGIN REFUSED: Invalid  Password [%s] (no `@' or no `.')",
	       passwd);
	log_pri (L_INFO,
	       "FTP LOGIN REFUSED: Compared Password [%s]", passwducase);
	passwducase = save1;
	free(passwducase);
	passwd_save = save2;
	free(passwd_save);
	return (REJECT|CHECKED_ONCE_FILE);
      } else {
	/*
	 * If no-more checking is specified here, we only set REJECT and
	 * CHECKED_ONCE_FILE flags.  Note that we don't leave here and
	 * continue more password checking.
	 */
	log_pri (L_INFO, "NMC: Invalid Password? (no `@' or no `.') [%s]",
	       passwd);
	if( (flag_exclude & NONE_DEFINED) )
	  flag_exclude = (REJECT|CHECKED_ONCE_FILE);
      }
    }
  }

  /*
   * If `path_invalidpw' is already defined in `PCKcheckmode*' entry,
   * `path_invalidpw' overrides `pw_invalidpw'.
   */
  if ( path_invalidpw && *path_invalidpw ) {
    if ( pw_check_debug & PCK_DEBUG_HIGH ) {
      log_pri (L_TRACE, "---: pw_invalidpw [%s]", path_invalidpw);
    }
    pw_invalidpw = pck_expand_tilde((char*)path_invalidpw);
  }

  if ( pw_check_debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: pw_invalidpw(final) [%s]",pw_invalidpw);

  if ((fd = fopen(pw_invalidpw, "r")) != NULL) {  /* if-loop-1 */
    while (fgets(line, sizeof(line), fd) != NULL)
      if ((p = strchr(line, '\n')) != NULL) {  /* if-loop-2 */
	*p = '\0';

	/*
	 * A Line beginning with a pound sign (`#') is a comment line.
	 */
	if (line[0] == '#' || line[0] == '\0')
	  continue;

	if ((lineucase = (char *)malloc((unsigned)(strlen(line)+1))) == NULL) {
	  pck_perror_reply(421, "Local resource failure (lineucase): malloc");
	  passwducase = save1;
	  free(passwducase);
	  passwd_save = save2;
	  free(passwd_save);
	  (void) fclose(fd);
	  return (ACCEPT|CHECKED_ONCE_FILE);
	}
	save3 = lineucase;

	if ((line_save = (char *)malloc((unsigned)(strlen(line)+1))) == NULL) {
	  pck_perror_reply(421, "Local resource failure (line_save): malloc");
	  passwducase = save1;
	  free(passwducase);
	  passwd_save = save2;
	  free(passwd_save);
	  (void) fclose(fd);
	  return (ACCEPT|CHECKED_ONCE_FILE);
	}
	save4 = line_save;

	/*
	 * Decompose `line' and obtain a *correct* regular expression
	 * and an error-message file name. When white spaces or tabs
	 * are contained in regular expressions, we must treat them
	 * carefully. Note that the entry in `ftpinvalidpw' would look
	 * like this:
	 *
	 * guest@*	 /some/where/in/system/errmsg.for.guest
	 * *[ ]*@*	 /some/where/in/system/errmsg.for.wspc
	 * *@*[ ]*	 /some/where/in/system/errmsg.for.wspc
	 *
	 * Note that when you use an expression like "*@*[ ]*", you
	 * SHOULD NOT specify "remove" as the third argument of
	 * "PCKcheckmode*" because "remove" (or "remove-control")
	 * actually removes all white spaces of regular expressions in
	 * `line'. The expressions including "*[ ]*" become effective
	 * when you use "remove-bracket" in "PCKcheckmode*".
	 *
	 * The regular expression is stored into `lineucase' from
	 * the entry `line' here.
	 */
	(void) pck_decompose_line(line, lineucase, 1);
	(void) strcpy(line_save, lineucase);
	(void) pck_upper(lineucase);
	/*
	log_pri (L_TRACE, "---: line,lineu,save = <%s> <%s> <%s>",
	       line, lineucase, line_save);
	       */

	/*
	 * If `line' is starting with `%', we should not ignore cases
	 * in regular expression. (Do case-sensitive comparison)
	 */

	/*
	 * Case: %FooBar*@*
	 */
	if ( *lineucase == '%' ) {  /* if-loop-3 */
	  while (*lineucase == '%') { lineucase++; line_save++;	}
	  case_sensitive = 1;
	  /* Pointer SHOULD NOT be restored here */
	  /* lineucase = save; */
	} else {
	  if ( *lineucase == '!' ) {
	    while (*lineucase == '!') lineucase++;
	    /*
	     * Case: !%FooBar*@*
	     */
	    if ( *lineucase == '%' ) {
	      while (*lineucase == '%') lineucase++;
	      case_sensitive = 1;
	    } else {
	      /*
	       * Case: !FooBar*@*
	       */
	      case_sensitive = 0;
	    }
	    /* Pointer SHOULD be restored here */
	    lineucase = save3;
	  } else if ( *lineucase == '$' ) {
	    while (*lineucase == '$') lineucase++;
	    /*
	     * Case: $%FooBar*@*
	     */
	    if ( *lineucase == '%' ) {
	      while (*lineucase == '%') lineucase++;
	      case_sensitive = 1;
	    } else {
	      /*
	       * Case: $FooBar*@*
	       */
	      case_sensitive = 0;
	    }
	    /* Pointer SHOULD be restored here */
	    lineucase = save3;
	  } else {
	    /*
	     * Case: FooBar*@*
	     */
	    case_sensitive = 0;
	  }
	}      /* end of if-loop-3 */

	if( (flag_exclude & NONE_DEFINED) ) flag_exclude = NONE_DEFINED;

	/*
	 * Password matches the line beginning with `!'.
	 */
	if ( *lineucase == '!' ) {  /* if-loop-4 */

	  while (*lineucase == '!') { lineucase++; line_save++; }
	  while (*lineucase == '%') { lineucase++; line_save++; }

	  /*
	   * We allow comma-separated domain name lists as an entry of
	   * `inclusion lists' beginning with "!" sign in ftpinvalid.
	   * The entry would look like:
	   * 
	   * !*@*.hokudai.ac.jp,*@*.gov,*@*.edu   /some/where/in/system/errmsg1
	   *
	   * or
	   *
	   * !%*@*.Hokudai.Ac.Jp,*@*.Gov,*@*.EDU  /some/where/in/system/errmsg2
	   *
	   * and check if user's address (password) matches each domain name.
	   *
	   */
	  if ( pw_check_debug & PCK_DEBUG_FULL )
	    log_pri (L_TRACE, "---: case_sensitive [%d]", case_sensitive);

	  linetemp = (case_sensitive)? line_save : lineucase;

	  if ( pw_check_debug & PCK_DEBUG_FULL )
	    log_pri (L_TRACE, "---: linetemp [%s]",linetemp);

	  /*
	   * Loop to separate each domain name.
	   */
	  do { /* do-while-loop-1 */

	    /*
	     * Remove leading '%' for each regular expression
	     * decomposed by 'pck_extract_class' routine.
	     */
	    while (*linetemp == '%') linetemp++;

	    if ( case_sensitive ) {
	      (void) strcpy(line_save,linetemp);
	      linetemp = pck_extract_class(line_save,0);
	    } else {
	      (void) strcpy(lineucase,linetemp);
	      linetemp = pck_extract_class(lineucase,0);
	    }

	    if ( pw_check_debug & PCK_DEBUG_FULL )
	      log_pri (L_TRACE, "---: !_Regexp-t,u,s  <%s> <%s> <%s>",
		     linetemp, lineucase, line_save);

	    if ( linetemp != NULL ) { /* if-loop-5 */

	      if ( !pck_fnmatch( linetemp,
			     (case_sensitive)? passwd_save : passwducase,
			      FNM_PERIOD2) ) {  /* if-loop-6 */
		  if( (flag_exclude & NONE_DEFINED) ) {  /* if-loop-7 */
		    /*
		     * If no-more checking is not specified here and password
		     * matches `!' expression, we leave here with ACCEPT flag
		     * and continue more checking.
		     */
		    flag_exclude = (ACCEPT|CHECKED_ONCE_FILE);
		    if (!no_more) {  /* if-loop-8 */
		      log_pri (L_INFO, "CHK: !_Match     <%s> <%s> <%s>",
			     (case_sensitive)? passwd_save : passwducase,
			     linetemp,
			     (case_sensitive)? line_save : lineucase);
		      log_pri (L_INFO,
			     "ACCEPTED: Password accepted [%s] <%s>",
			     passwd,line);
		      log_pri (L_INFO,
			     "ACCEPTED(pck_validate_password): Compared Password [%s] <%s> <%s>",
			     (case_sensitive)? passwd_save : passwducase,
			     linetemp,
			     (case_sensitive)? line_save : lineucase);
		      passwducase = save1;
		      free(passwducase);
		      passwd_save = save2;
		      free(passwd_save);
		      lineucase = save3;
		      free(lineucase);
		      line_save = save4;
		      free(line_save);
		      (void) fclose(fd);
		      return(flag_exclude);
		    } else {
		      log_pri (L_INFO, "NMC: !_Match     <%s> <%s> <%s>",
			     (case_sensitive)? passwd_save : passwducase,
			     linetemp,
			     (case_sensitive)? line_save : lineucase);
		    }      /* end of if-loop-8 */
		  }        /* end of if-loop-7 */
		}          /* end of if-loop-6 */
	    }              /* end of if-loop-5 */

	    linetemp = (case_sensitive)? line_save : lineucase;

	  } while ( (linetemp = pck_shift_class_list(linetemp,0)) != NULL);
	  /* end of do-while-loop-1 */

	  /*
	   * User's password that reached here doesn't match any
	   * regular expression beginning with "!" sign.
	   */
	  if ( pw_check_debug & PCK_DEBUG_FULL )
	    log_pri (L_TRACE, "---: !_Regexp-Reject <%s> <%s> <%s>",
		   linetemp, lineucase, line_save);

	  if( (flag_exclude & NONE_DEFINED) ) {  /* if-loop-9 */
	    /*
	     * If no-more checking is not specified here and password
	     * does not match `!' expression, we leave here with REJECT
	     * flag and refuse anonymous login.
	     */
	    flag_exclude = (REJECT|CHECKED_ONCE_FILE);

	    /*
	     * Redefine `lineucase' and `line_save' here for
	     * syslog outputs.
	     */
	    lineucase = save3;
	    line_save = save4;
	    (void) pck_decompose_line(line, pw_pwderrmsg2, 0);
	    (void) pck_decompose_line(line, lineucase, 1);
	    (void) strcpy(line_save, lineucase);
	    (void) pck_upper(lineucase);
	    while (*lineucase == '!') { lineucase++; line_save++; }
	    while (*lineucase == '%') { lineucase++; line_save++; }

	    if (!no_more) {  /* if-loop-10 */
	      log_pri (L_INFO, "CHK: !_Not Match <%s> <%s>",
		     (case_sensitive)? passwd_save : passwducase,
		     (case_sensitive)? line_save   : lineucase);
	      log_pri (L_INFO,
		     "REFUSED: Password rejected [%s] <%s>",passwd,line);
	      log_pri (L_INFO,
		     "REFUSED: Compared Password [%s] <%s>", 
		     (case_sensitive)? passwd_save : passwducase,
		     (case_sensitive)? line_save   : lineucase);
	      passwducase = save1;
	      free(passwducase);
	      passwd_save = save2;
	      free(passwd_save);
	      lineucase = save3;
	      free(lineucase);
	      line_save = save4;
	      free(line_save);
	      (void) fclose(fd);
	      return(flag_exclude);
	    } else {
	      log_pri (L_INFO, "NMC: !_Not Match <%s> <%s>",
		     (case_sensitive)? passwd_save : passwducase,
		     (case_sensitive)? line_save   : lineucase);
	    }      /* end of if-loop-10 */
	  }        /* end of if-loop-9 */

	  /*
	   * Password matches the line beginning with `$'.
	   */
	} else if ( *lineucase == '$' ) {

	  while (*lineucase == '$') { lineucase++; line_save++; }
	  while (*lineucase == '%') { lineucase++; line_save++; }

          if ( pw_check_debug & PCK_DEBUG_FULL )
            log_pri (L_TRACE, "---: case_sensitive [%d]", case_sensitive);

          linetemp = (case_sensitive)? line_save : lineucase;

          if ( pw_check_debug & PCK_DEBUG_FULL )
            log_pri (L_TRACE, "---: linetemp <%s>",linetemp);

          /*
           * Loop to separate each domain name.
           */
          do { /* do-while-loop-4 */

	    /*
	     * Remove leading '%' for each regular expression
	     * decomposed by 'pck_extract_class' routine.
	     */
	    while (*linetemp == '%') linetemp++;

            if ( case_sensitive ) {
              (void) strcpy(line_save,linetemp);
              linetemp = pck_extract_class(line_save,0);
            } else {
              (void) strcpy(lineucase,linetemp);
              linetemp = pck_extract_class(lineucase,0);
            }

            if ( pw_check_debug & PCK_DEBUG_FULL )
              log_pri (L_TRACE, "---: $_Regexp-t,u,s  <%s> <%s> <%s>",
                     linetemp, lineucase, line_save);

	    if ( linetemp != NULL ) { /* if-loop-19 */

	      if ( !pck_fnmatch( linetemp,
			     (case_sensitive)? passwd_save : passwducase,
			     FNM_PERIOD2) ) {  /* if-loop-11 */
		if( !(flag_exclude & NO_MORE_CHECK) ) {  /* if-loop-12 */
		  /*
		   * If no-more checking is specified here and password
		   * matches `$' expression, we turn on ACCEPT flag,
		   * leave here and accept anonymous login.
		   * Note that NO_MORE_CHECK flag is also defined here and
		   * we ignore any more checking for this password.
		   */
		  if (no_more) {  /* if-loop-13 */
		    log_pri (L_INFO, "NMC: $_Match <%s> <%s> <%s>",
			   (case_sensitive)? passwd_save : passwducase,
			   linetemp,
			   (case_sensitive)? line_save : lineucase);
		    log_pri (L_INFO,
			   "ACCEPTED: Password accepted [%s] <%s>",
			   passwd,line);
		    log_pri (L_INFO,
			   "ACCEPTED(pck_validate_password): Compared Password [%s] <%s> <%s>",
			   (case_sensitive)? passwd_save : passwducase,
			   linetemp,
			   (case_sensitive)? line_save : lineucase);
		    log_pri (L_INFO,
			   "NMC: Checks Stopped Here. Login Accepted.");
		    flag_exclude = (ACCEPT|NO_MORE_CHECK);
		    flag_exclude |= CHECKED_ONCE_FILE;
		    passwducase = save1;
		    free(passwducase);
		    passwd_save = save2;
		    free(passwd_save);
		    lineucase = save3;
		    free(lineucase);
		    line_save = save4;
		    free(line_save);
		    (void) fclose(fd);
		    return(flag_exclude);
		  } else {
		    log_pri (L_INFO, "CHK: $_Match <%s> <%s> <%s>",
			   (case_sensitive)? passwd_save : passwducase,
			   linetemp,
			   (case_sensitive)? line_save : lineucase);
		    flag_exclude = NONE_DEFINED;
		  }      /* end of if-loop-13 */
		}        /* end of if-loop-12 */
	      }          /* end of if-loop-11 */
	    }            /* end of if-loop-19 */

	    linetemp = (case_sensitive)? line_save : lineucase;

	  } while ( (linetemp = pck_shift_class_list(linetemp,0)) != NULL);
	  /* end of do-while-loop-4 */

	  /*
	   * Password matches one of regular expressions other than
	   * `!' and `$' in file `pw_invalidpw'.
	   */
	} else {

	  /*
	   * We allow comma-separated `regex' lists as an entry of
	   * `exclusion lists' in ftpinvalid.
	   * The entry would look like:
	   * 
	   * @*,*@,*@*.,*..*,*@*@*      /some/where/in/system/errmsg1
	   *
	   * or
	   *
	   * *[ ]*@*,*@*[ ]*            /some/where/in/system/errmsg2
	   *
	   * and check if user's address (password) matches each domain name.
	   *
	   */

          if ( pw_check_debug & PCK_DEBUG_FULL )
            log_pri (L_TRACE, "---: case_sensitive [%d]", case_sensitive);

	  linetemp = (case_sensitive)? line_save : lineucase;

          if ( pw_check_debug & PCK_DEBUG_FULL )
            log_pri (L_TRACE, "---: linetemp <%s>",linetemp);

          /*
           * Loop to separate each regular expression.
           */
	  do { /* do-while-loop-3 */

            if ( case_sensitive ) {
              (void) strcpy(line_save,linetemp);
              linetemp = pck_extract_class(line_save,0);
            } else {
              (void) strcpy(lineucase,linetemp);
              linetemp = pck_extract_class(lineucase,0);
            }

            if ( pw_check_debug & PCK_DEBUG_FULL )
              log_pri (L_TRACE, "---: Regexp-t,u,s  <%s> <%s> <%s>",
                     linetemp, lineucase, line_save);

	    if ( linetemp != NULL ) { /* if-loop-18 */

	      if ( !pck_fnmatch( linetemp,
			     (case_sensitive)? passwd_save : passwducase,
			     FNM_PERIOD2) ) {  /* if-loop-14 */
		if( (flag_exclude & NONE_DEFINED) ) {  /* if-loop-15 */
		  /*
		   * If no-more checking is not specified here, we turn on
		   * REJECT flag, leave here and refuse anonymous login.
		   */
		  flag_exclude = (REJECT|CHECKED_ONCE_FILE);
		  if (!no_more) {  /* if-loop-16 */
		    log_pri (L_INFO, "CHK: Match   <%s> <%s> <%s>",
			   (case_sensitive)? passwd_save : passwducase,
			   linetemp,
			   (case_sensitive)? line_save : lineucase);
		    log_pri (L_INFO,
			   "REFUSED: Password rejected [%s] <%s>",passwd,line);
		    log_pri (L_INFO,
			   "REFUSED: Compared Password [%s] <%s> <%s>", 
			   (case_sensitive)? passwd_save : passwducase,
			   linetemp,
			   (case_sensitive)? line_save : lineucase);
		    passwducase = save1;
		    free(passwducase);
		    passwd_save = save2;
		    free(passwd_save);
		    lineucase = save3;
		    free(lineucase);
		    line_save = save4;
		    free(line_save);
		    (void) fclose(fd);
		    (void) pck_decompose_line(line, pw_pwderrmsg2, 0);
		    return(flag_exclude);
		  } else {
		    log_pri (L_INFO, "NMC: Match   <%s> <%s>",
			   (case_sensitive)? passwd_save : passwducase,
			   (case_sensitive)? line_save : lineucase);
		    (void) pck_decompose_line(line, pw_pwderrmsg2, 0);
		  }      /* end of if-loop-16 */
		}        /* end of if-loop-15 */
	      }          /* end of if-loop-14 */
	    } /* end of if-loop-18 */

	    linetemp = (case_sensitive)? line_save : lineucase;

	  } while ( (linetemp = pck_shift_class_list(linetemp,0)) != NULL);
	  /* end of do-while-loop-3 */

	}            /* end of if-loop-4 */
	lineucase = save3;
	free(lineucase);
	line_save = save4;
	free(line_save);
      }              /* end of if-loop-2 */

  } else {
    log_pri (L_ERROR, "ERR: Cannot Open INVALIDPW file %s: %s",
	   pw_invalidpw, strerror(errno));
    passwducase = save1;
    free(passwducase);
    passwd_save = save2;
    free(passwd_save);
    /*
    lineucase = save3;
    free(lineucase);
    line_save = save4;
    free(line_save);
    */
    return (ACCEPT|CHECKED_ONCE_FILE);
  }      /* end of if-loop-1 */

  if ( pw_check_debug & PCK_DEBUG_FULL )
    log_pri (L_TRACE, "---: flag_exclude [%d]",flag_exclude);

  if ( (flag_exclude & NONE_DEFINED) ) {  /* if-loop-17 */
    /*
     * If no regular expression hit for given password
     * we turn on _ONLY_ ACCEPT flag and return with
     * CHECKED_ONCE_FILE flag.
     */
    flag_exclude = (ACCEPT|CHECKED_ONCE_FILE);
    if (!no_more) {
      log_pri (L_INFO, "ACCEPTED: password -> [%s]",passwd);
    }
  } else {
    /*
     * In other cases, we turn on only CHECKED_ONCE_FILE flag 
     * and strip other flags off.
     */
    flag_exclude &= ~CHECKED_ONCE_FILE;
    flag_exclude &= ~NO_MORE_CHECK;
    flag_exclude &= ~NONE_DEFINED;
    flag_exclude |=  CHECKED_ONCE_FILE;
  }      /* end of if-loop-17 */

  if ( pw_check_debug & PCK_DEBUG_FULL )
    log_pri (L_TRACE, "---: flag_exclude [%d]",flag_exclude);
  passwducase = save1;
  free(passwducase);
  passwd_save = save2;
  free(passwd_save);
  (void) fclose(fd);
  return (flag_exclude);
}

void
pck_decompose_line(char input[], char output[], int regular)
{
  /*
   * INPUT:
   *        input ..... strings which must be in the format like
   *           *[ ]*@*[.]*           /some/where/in/the/system/error.msg1
   *           !*@*[.]*              /some/where/in/the/system/error.msg2
   *
   *              Note that we accept white spaces in regular expressions.
   *
   *
   *        regular ... integer flag to decompose `input'
   *
   * OUTPUT:
   *        output .... a regular expression when regular=1,
   *                    an error-message file name when regular=0
   */
  char *save_1, *save_2;
  int ispc, itab;

  /*
  if ( pw_check_debug)
    log_pri (L_TRACE, "---: ORG LINE <%s>\n", input);
    */

  (void) strcpy(output,input);
  save_1 = output;
  save_2 = save_1;

  ispc = pck_istrstr(0, input, " ");	/* Where is the first SPACE code? */
  itab = pck_istrstr(0, input, "\t");	/* Where is the first TAB code? */

  /*
   * If the original entry does not contain TABs nor WHITE SPACEs,
   * we quit here.
   */
  if ( (ispc == strlen(input)) && (itab == strlen(input)) ) {
    if ( !regular ) *save_1 = '\0';     /* File name is not given here. */
    if ( pw_check_debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: Regexp_1 <%s>\n", output);
    return;
  }

  /*
   * Remove strings for regular expressions.
   */
  while( !isspace(*save_2) ) save_2++;

  /*
   * Remove TABs and WHITE SPACEs.
   */
  while(  isspace(*save_2) ) save_2++;

  /*
   * When white spaces or tabs are contained in regular expressions,
   * we must treat them carefully to extract a *correct* regular
   * expression and/or file name from the input string.
   * Note that white spaces or tabs are actually specified between
   * "[" and "]" signs.
   */

  /*
  if ( pw_check_debug)
    log_pri (L_TRACE, "---: Temp-1   <%s>\n", save_2);
    */

  while ( strstr(save_2,"]") != NULL ) {
    while( !isspace(*save_2) && *save_2 != '\0' ) save_2++;
    while(  isspace(*save_2) && *save_2 != '\0' ) save_2++;
  }

  /*
  if ( pw_check_debug)
    log_pri (L_TRACE, "---: Temp-2   <%s>\n", save_2);
    */
  if ( regular ) {
    while( isspace(*--save_2) );	/* Remove trailing space/tab codes */
    *++save_2 = '\0';
    if ( pw_check_debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: Regexp_2 <%s>\n", output);
    return;
  }

#ifdef USG
  (void) memmove(save_1, save_2, strlen(save_2)+1);
#else
  (void)   bcopy(save_2, save_1, strlen(save_2)+1);
#endif

  if ( strlen(save_1) ) {
    if ( pw_check_debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: ERRMSG [%s] (size:%d)\n",
	     output, strlen(save_1));
  }
  return;
}

#endif /* end of PW_CHECK_FILE */


/*
 * ####################################################
 * Common Modules for PW_CHECK_FILE and/or PW_CHECK_DNS
 * ####################################################
 */

#if defined(PW_CHECK_FILE) || defined(PW_CHECK_DNS)

int
pck_need_to_check(int valid, 
		  char (*how_to_check)[MAXPATHLEN],
		  char *directive, 
		  char *logging_mark)
{
  char *target_class;
  char check_class[BUFSIZ];
  int check_flag;
  config_rec	*pck_tmp;

  /*
   * return if already rejected
   */
  if ( !valid )
    return (valid);

  /*
   * return if NO_MORE_CHECK flag is already set
   */
  if ( (valid & NO_MORE_CHECK) )
    return (valid);

  if ( pw_check_debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: pck_need_to_check executing...");

  /*
   * We initialize valid flag.
   */
  check_flag  =  NO_NEED_TO_CHECK;
  check_flag |=  NONE_DEFINED;
  check_flag &= ~NEED_TO_CHECK;
  check_flag &= ~CHECKED_ONCE_FILE;
  check_flag &= ~NO_MORE_CHECK;

  /*
   * "pck_buf0" is reduced to a remote access class name.
   */
  if ( !strcasecmp(pck.index, "class") ) {
    (void) strcpy(pck_buf0,pck.remote.class);
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: CLASS check!!");
    /*
     * "pck_buf0" is reduced to a remote host name.
     */
  } else if ( !strcasecmp(pck.index, "host" ) ) {
    (void) strcpy(pck_buf0,pck.remote.name);
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: HOST  check!!");
    /*
     * By default, "pck_buf0" is reduced to a remote acess class name.
     */
  } else {
    (void) strcpy(pck_buf0,pck.remote.class);
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: CLASS check!! <default>");
  }

  (void) pck_upper(pck_buf0);


  /*
   * We search all `PCKcheckmode*' entries in proftpd.conf and check if
   * user's class matches one of their 4-th entries. If we judge
   * user's class needs password checking, we set 1st(arg_0),
   * 2nd(arg_1) and 3rd(arg_2) parameters of `PCKcheckmode*' and leave
   * here with NEED_TO_CHECK flag. If we judge user's class needs no
   * password checking, we leave here with NO_NEED_TO_CHECK flag.
   */
  pck_tmp = find_config(main_server->conf,
			CONF_PARAM,
			directive,
			TRUE);
  if(pck_tmp) do { /* while-loop-1 */

#if 1
    if ( pw_check_debug & PCK_DEBUG_FULL ) {
      log_pri (L_TRACE, "---: No. of arguments [%d]", pck_tmp->argc);
      log_pri (L_TRACE, "---: acess_class [%s]",pck.remote.class);
      log_pri (L_TRACE, "---: remote_name [%s]",pck.remote.name);
      if( pck_tmp->argc > 4 ) {
	log_pri (L_TRACE, "---: ARG0 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[0]);
	log_pri (L_TRACE, "---: ARG1 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[1]);
	log_pri (L_TRACE, "---: ARG2 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[2]);
	log_pri (L_TRACE, "---: ARG3 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[3]);
	log_pri (L_TRACE, "---: ARG4 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[4]);
      } else if( pck_tmp->argc > 3 ) {
	log_pri (L_TRACE, "---: ARG0 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[0]);
	log_pri (L_TRACE, "---: ARG1 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[1]);
	log_pri (L_TRACE, "---: ARG2 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[2]);
	log_pri (L_TRACE, "---: ARG3 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[3]);
      } else{
	log_pri (L_ERROR, "ERR: Argument error? (PCKcheckmodehost/class)");
      }
    }

    if( pck_tmp->argc < 4 )
      log_pri (L_ERROR, "ERR: Argument error? (PCKcheckmodehost/class)");

#else
    if ( pw_check_debug & PCK_DEBUG_FULL ) {
      log_pri (L_TRACE, "---: acess_class [%s]",pck.remote.class);
      if(pck_tmp->argv[0]) 
	log_pri (L_TRACE, "---: ARG0 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[0]);
      if(pck_tmp->argv[1]) 
	log_pri (L_TRACE, "---: ARG1 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[1]);
      if(pck_tmp->argv[2]) 
	log_pri (L_TRACE, "---: ARG2 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[2]);
      if(pck_tmp->argv[3]) 
	log_pri (L_TRACE, "---: ARG3 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[3]);
      if(pck_tmp->argv[4]) 
	log_pri (L_TRACE, "---: ARG4 (NeedToCheck) [%s]",
	       (char *)pck_tmp->argv[4]);
    }
#endif

    /*
     * We cannot find the 4-th entry (ARG3) in proftpd.conf.
     * In this case, we allow login.
     */
    if(!pck_tmp->argv[3]) {
      if( (check_flag & NONE_DEFINED) ) {
        check_flag &= ~NEED_TO_CHECK;
        if(pck_tmp->argv[0]) 
	  (void)strcpy(how_to_check[0],pck_tmp->argv[0]);
        if(pck_tmp->argv[1]) 
	  (void)strcpy(how_to_check[1],pck_tmp->argv[1]);
        if(pck_tmp->argv[2]) 
	  (void)strcpy(how_to_check[2],pck_tmp->argv[2]);
      }
      log_pri (L_TRACE, "---: ARG3 Not Found");
      continue;
    }
    /*
     * We save `class' information (ARG3) of each `PCKcheckmode*' entry.
     * Note that we accept regular expression of `class'.
     */
    if(pck_tmp->argv[3]) 
      (void)strcpy(check_class, pck_tmp->argv[3]);
    (void) pck_upper(check_class);
    target_class = check_class;

    if ( pw_check_debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: list  [%s]",check_class);

    /*
     * Since 4th argument is found, we strip NONE_DEFINED flag off.
     */
    check_flag &= ~NONE_DEFINED;

    do { /* do-while-loop-2 */
      /*
       * We extract each check class successively from the 4-th argument
       * until we meet `!' entry or the end of all entries.
       */
      (void) strcpy(check_class,target_class);
      target_class = pck_extract_class(check_class,1);

      if ( pw_check_debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: shift [%s] [%s<-%s]",
	       pck.remote.class,target_class,check_class);

      if ( target_class != NULL ) {

	if ( *target_class == '!' ) {
	  while ( *target_class == '!' ) target_class++;
	  /*
	   * We do not have to do password check against this access
	   * class, because this class matches `!' expression.
	   */
	  if ( !pck_fnmatch(target_class, 
			    IsDigitAd(target_class, pck.index)?
			    (char*)pck.remote.ip : pck_buf0,
			    FNM_PERIOD2) ) {
	    log_pri (L_INFO, "%s match [!%s] <%s> NO NEED TO CHECK",
		   logging_mark,
		   target_class,
		   IsDigitAd(target_class, pck.index)?
		   (char*)pck.remote.ip : pck_buf0);
	    if( !(check_flag & CHECKED_ONCE_FILE) ) {
	      check_flag |=  CHECKED_ONCE_FILE;
	      check_flag &= ~NEED_TO_CHECK;
	    }

	    /*
	     * We have to do password check against this access class,
	     * because this class does not match `!' expression.
	     */
	  } else {
	    log_pri (L_INFO, "%s match [!%s] <%s>",
		   logging_mark,
		   target_class,
		   IsDigitAd(target_class, pck.index)?
		   (char*)pck.remote.ip : pck_buf0);
	    if( !(check_flag & CHECKED_ONCE_FILE) ) {
	      (void)strcpy(how_to_check[0],pck_tmp->argv[0]);
	      (void)strcpy(how_to_check[1],pck_tmp->argv[1]);
	      (void)strcpy(how_to_check[2],pck_tmp->argv[2]);
	      if(pck_tmp->argv[3])
		(void)strcpy(how_to_check[3],pck_tmp->argv[3]);
	      pck_path_define_pck(directive,
				  how_to_check[4],
				  pck_tmp->argv[4]);
	      check_flag |=  CHECKED_ONCE_FILE;
	      check_flag |=  NEED_TO_CHECK;
	    }
	  }

	  /*
	   * We do not have to do password check against this access
	   * class, because this class matches `$' expression.
	   * If user's class does not match `$' expression,
	   * we do nothing and continue searching `PCKcheckmode*'
	   * entries in proftpd.conf.
	   */
	} else if ( *target_class == '$' ) {
	  while ( *target_class == '$' ) target_class++;
	  if ( !pck_fnmatch(target_class, 
			    IsDigitAd(target_class, pck.index)?
			    (char*)pck.remote.ip : pck_buf0,
			    FNM_PERIOD2) ) {
	    log_pri (L_INFO, "%s match [$%s] <%s> NO NEED TO CHECK",
		   logging_mark,
		   target_class,
		   IsDigitAd(target_class, pck.index)?
		   (char*)pck.remote.ip : pck_buf0);
	    if( !(check_flag & NO_MORE_CHECK) ) {
	      check_flag |=  CHECKED_ONCE_FILE;
	      check_flag |=  NO_MORE_CHECK;
	      check_flag &= ~NEED_TO_CHECK;
	    }
	  }
	} else {
	  /*
	   * We have to do password check against this access class.
	   */
	  if ( !pck_fnmatch(target_class, 
			    IsDigitAd(target_class, pck.index)?
			    (char*)pck.remote.ip : pck_buf0,
			    FNM_PERIOD2) ) {
	    log_pri (L_INFO, "%s match [%s] <%s>",
		   logging_mark,
		   target_class,
		   IsDigitAd(target_class, pck.index)?
		   (char*)pck.remote.ip : pck_buf0);
	    if( !(check_flag & CHECKED_ONCE_FILE) ) {
	      (void)strcpy(how_to_check[0],pck_tmp->argv[0]);
	      (void)strcpy(how_to_check[1],pck_tmp->argv[1]);
	      (void)strcpy(how_to_check[2],pck_tmp->argv[2]);
	      if(pck_tmp->argv[3])
		(void)strcpy(how_to_check[3],pck_tmp->argv[3]);
	      pck_path_define_pck(directive,
				  how_to_check[4],
				  pck_tmp->argv[4]);
	      check_flag |=  CHECKED_ONCE_FILE;
	      check_flag |=  NEED_TO_CHECK;
	    }
	  }
	}
      }
      target_class = check_class;
    } while ( (target_class = pck_shift_class_list(target_class,1)) != NULL);
    /* end of do-while-loop-2 */

    /* end of while-loop-1 */
  } while((pck_tmp = find_config_next(pck_tmp,
				      pck_tmp->next,
				      CONF_PARAM,
				      directive,
				      TRUE)) != NULL);

  /*
   * User's class matches `$' expression.  We ignore all check routines
   * hereafter and allow login.
   */
  if( (check_flag & NO_MORE_CHECK) ) {
    log_pri (L_INFO, (!strcasecmp(pck.index, "class"))?
	   "%s class [%s] match (`$') NO NEED TO CHECK" :
	   "%s host  [%s] match (`$') NO NEED TO CHECK",
	   logging_mark, pck_buf0);
	   check_flag  = NO_NEED_TO_CHECK;
    return(check_flag);
  }

  if( (check_flag & NONE_DEFINED) ) {
    /*
     * When `PCKcheckmode*' does exist in proftpd.conf without 4th argument,
     * we leave here and allow login.
     */
    log_pri (L_ERROR, "ERR: \"PCKcheckmode*\" does not appear in proftpd.conf");
    log_pri (L_ERROR, "ERR: or class [%s] is not given... check it.",
	   pck_buf0);
    check_flag  = NO_NEED_TO_CHECK;
    return(check_flag);
  } else {
    /*
     * User's class matches some regular expression.
     * Access permission is stored in check_flag here.
     * Other flags should be stripped off before return.
     */
    if( (check_flag & CHECKED_ONCE_FILE) ) {
      
      log_pri (L_INFO, (!strcasecmp(pck.index, "class"))?
	     "%s class [%s] match" :
	     "%s host  [%s] match",
	     logging_mark, pck_buf0);
      check_flag &= ~NONE_DEFINED;
      check_flag &= ~NO_MORE_CHECK;
      check_flag &= ~CHECKED_ONCE_FILE;
      if ( pw_check_debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: return flag? [%d]", check_flag);
      if ( (check_flag & NEED_TO_CHECK) ) {
	if ( pw_check_debug & PCK_DEBUG_FULL )
	  log_pri (L_TRACE, "---: check mode?  [%s]",how_to_check[0]);
      }
      return(check_flag);
    } else {
      /*
       * Although 4th paratemer (classes) is specified in `PCKcheckmode*'
       * of proftpd.conf, we cannot find user's access class there.
       * In this case we leave here and allow login.
       */
      log_pri (L_INFO, (!strcasecmp(pck.index, "class"))?
	     "%s class [%s] does not match" :
	     "%s host  [%s] does not match",
	     logging_mark, pck_buf0);
      check_flag = NO_NEED_TO_CHECK;
      return(check_flag);
    }
  }
}

char *
pck_extract_class(char *class, int mod)
{
  char extract[BUFSIZ], *temp;
  int i, comma, len;
  int bracket_1, bracket_2;
  /* 
   * We use 'return code' (0x0D) as a replace sign.
   */
  static char replace_sign = '\r';

  /*printf("		EXT=<%s>\n", class);*/
  temp = extract;
  while ( *class == ',' ) class++;
  len  = strlen(class);

  bracket_1 = pck_istrstr (0, class, "["); /* Where is the first "[" ? */
  bracket_2 = pck_istrstr (0, class, "]"); /* Where is the first "]" ? */

  /*
   * If square brackets are found, commas (',') lying between them
   * should be replaced with 'replace_sign' to correctly extract
   * the regular expression including commas.
   */
  if ( len != bracket_1 && len != bracket_2 ) {
    for ( i = bracket_1; i <= bracket_2; i++ ) {
      if ( class[i] == ',' ) class[i] = replace_sign;
      if ( i == bracket_1 ) class[i] = '[';
      if ( i == bracket_2 ) class[i] = ']';
    }
  }

  /*printf("		EXT=<%s>\n", class);*/
  comma = pck_istrstr(0, class, ",");

  if ( comma == len && len == 0 ) {
    /*
     * Comma is not included any more. (quit this routine)
     */
    return (NULL);
  } else {
    /*
     * If square brackets found, commas (',') replaced with
     * 'replace_sign' revive here !
     */
    for ( i = 0; i < comma; i++ ) {
      if ( len != bracket_1 && len != bracket_2 &&
           i   >= bracket_1 && i   <= bracket_2 &&
           *class == replace_sign ) {
        *temp++ = ',';
        class++;
      } else {
        *temp++ = *class++;
      }
    }
    *temp = '\0';
    if(mod) (void) pck_upper(extract);
    return ( strlen(extract) == 0 ? NULL : extract );
  }

}

char *
pck_shift_class_list(char *class_list, int mod)
{
  char list_buf[BUFSIZ], *temp;
  int comma, len, i;
  int bracket_1, bracket_2;
  /* 
   * We use 'return code' (0x0D) as a replace sign.
   */
  static char replace_sign = '\r';

  while ( *class_list == ',' ) class_list++;

  /*printf("		SHT=<%s>\n",class_list);*/
  len  = strlen(class_list);
  comma = pck_istrstr (0, class_list, ",");
  if ( comma == len ) {
    return (NULL);
  }

  bracket_1 = pck_istrstr (0, class_list, "["); /* Where is the first "]" ? */
  bracket_2 = pck_istrstr (0, class_list, "]"); /* Where is the first "[" ? */

  /*
   * If square brackets are found, commas (',') lying between them
   * should be replaced with 'replace_sign' to correctly extract
   * the regular expression including commas.
   */
  if ( len != bracket_1 && len != bracket_2 ) {
    for ( i = bracket_1; i <= bracket_2; i++ ) {
      if ( class_list[i] == ',' ) class_list[i] = replace_sign;
      if ( i == bracket_1 ) class_list[i] = '[';
      if ( i == bracket_2 ) class_list[i] = ']';
    }
  }

  (void) strcpy(list_buf, class_list);
  temp = list_buf;

  len  = strlen(list_buf);
  for ( i = 0; i < len; i++ ) {
    if ( i > comma ) {
      *temp = *class_list;
      temp++;
    }
    class_list++;
  }
  *temp = '\0';
  if(mod) (void) pck_upper(list_buf);
  return ( strlen(list_buf) == 0 ? NULL : list_buf );
}

char *
pck_remove_control(char *input)
{
  char output[BUFSIZ], *temp, *save;
  int len;

  len = strlen(input);
  if ( len == 0)
    return (input);

  (void) strcpy(output, input);
  temp = output;
  save = temp;

  while ( *input != '\0' ) {
    if (!iscntrl(*input)  &&
        !isspace(*input) ) {
      *temp = *input;
      temp++;
    }
    input++;
  }
  *temp = '\0';
  /*
  (void) pck_upper(output);
  */
  return (save);
}

char *
pck_remove_bracket(char *input)
{
  char output[BUFSIZ], *temp, *save;
  int i, len;
  int first, last;

  len = strlen(input);
  if ( len == 0)
    return (input);

  first = pck_istrstr (0, input, "(");
  last  = pck_istrrstr(0, input, ")");
  if ( first == len || last == len )
    return (input);

  (void) strcpy(output, input);
  temp = output;
  save = temp;

  for ( i = 0; i < len; i++ ) {
    if ( i < first || i > last ) {
      *temp = *input;
      temp++;
    }
    input++;
  }
  *temp = '\0';
  /*
  (void) pck_upper(output);
  */
  return (save);
}

char *
pck_remove_one_char(const char *input, char remove)
{
  char *ptr, *temp;
  
  if (input) {
    ptr = (char *) malloc((unsigned) (strlen(input)+1));
    if (ptr == NULL) {
      pck_perror_reply(421,"Local resource failure (remove_one_char): malloc");
      return ((char *)input);
      /*fprintf(stderr, "\"malloc\" botched. Out of memory...");
	exit(1);*/
    }
    (void) strncpy(ptr, input, strlen(input));
    temp = ptr;
  } else {
    /*ptr = (char *) NULL;*/
    return ((char *)input);
  }
  
  while ( *input != '\0' ) {
    if (  *input != remove ) {
      *temp = *input;
      temp++;
    }
    input++;
  }
  *temp = '\0';
  
  return (ptr);
}

int
pck_istrstr(int location, char *origin, char *search)
{
  int i;
  int origin_len = strlen(origin);
  int search_len = strlen(search);

  if ( location < 0 || location >= origin_len ) {
    return(origin_len);
  } else {
    for ( i = location; i < origin_len; i++ )
      if (!strncasecmp(origin+i, search, search_len))
	return(i);
  }
  return(origin_len);
}

int
pck_istrrstr(int location, char *origin, char *search)
{
  int i;
  int origin_len = strlen(origin);
  int search_len = strlen(search);

  if ( location < 0 || location >= origin_len ) {
    return(origin_len);
  } else {
    for ( i = origin_len-location; i > 0; i-- )
      if (!strncasecmp(origin+i, search, search_len))
	return(i);
    if (!strncasecmp(origin, search, search_len))
      return(0);
  }
  return(origin_len);
}

int
pck_isdigitad(const char *address)
{
  int i, temp_len, location;
  int count = 0;
  char *temp;

  temp = (char *)address;
  temp_len = strlen(temp);
  location = pck_istrrstr(0, temp, "@");
  if ( temp_len == location ) location = 0;

  temp += location;
  for ( i = location; i < temp_len; i++ ) {
    if ( isdigit(*temp) || *temp == ' ' ||
	 *temp == '@'   || *temp == '.' ||
	 *temp == '['   || *temp == ']' ||
	 *temp == '('   || *temp == ')' )
      count++;
    temp++;
  }
  return( (temp_len-location)==count ? 1 : 0 );
}

int
pck_isdigitad2(const char *address)
{
  int i, temp_len, location;
  int count = 0;
  char *temp;

  temp = (char *)address;
  temp_len = strlen(temp);
  /*
   * If "address" contains only one asterisk(*),
   * we treat it as _NON_ digitad.
   */
  if(temp_len==1 && *temp=='*') {
    /*printf("* HIT!!\n");*/
    return 0;
  }
  location = pck_istrrstr(0, temp, "@");
  if ( temp_len == location ) location = 0;

  temp += location;
  for ( i = location; i < temp_len; i++ ) {
    if ( isdigit(*temp) || *temp == ' ' ||
	 *temp == '@'   || *temp == '.' ||
	 *temp == '['   || *temp == ']' ||
	 *temp == '*'   || *temp == '?' ||
	 *temp == '('   || *temp == ')' )
      count++;
    temp++;
  }
  return( (temp_len-location)==count ? 1 : 0 );
}

int
pck_password_check(int valid, 
		   char *passwd, 
		   char (*check_mode)[MAXPATHLEN],
		   char *directive, 
		   char *logging_mark)
{

  char *arg_0, *arg_1, *arg_2, *arg_3, *arg_4;
  int valid_save = 1;

  /*
   * We initialize `valid' for `password-check-kit'
   */
  valid = 1;

  /*
   * First let us check if we have to do `no-more-check'.
   * Note that `PCKnomorecheck true' or omitting `PCKnomorecheck'
   * in proftpd.conf _ONLY_ enables no-more checking.
   * Other cases do not allow no-more checking.
   */
  no_more = pck.no_more_check;

  valid_save = valid;

  /*
   * Let us check if we have to do password checking against user's
   * access class.  When user's access class is included in 4th
   * parameter of `PCKcheckmode*' entry, user's password requires
   * password checking. In this case, `pck_need_to_check' module returns
   * NEED_TO_CHECK flag with 1st, 2nd and 3rd parameters of
   * `PCKcheckmode*' entry.  Note that arg_0 plays the most important
   * role of this password check kit. Also note that `pck_need_to_check'
   * module returns NO_NEED_TO_CHECK flag in case of no need to check.
   */

  if ( pw_check_debug & PCK_DEBUG_FULL )
    log_pri (L_TRACE, "---: check_mode [%s]", check_mode[3]);
  
  if( pck_need_to_check(valid,
			check_mode,
			directive,
			logging_mark) ) {
    /*
     * arg_0 is the 1st parameter (ex. "file+dns")
     * arg_1 is the 2nd parameter (ex. "enforce")
     * arg_2 is the 3rd parameter (ex. "remove")
     * arg_3 is the 4th parameter (ex. "*")
     * arg_4 is the 5th parameter (ex. "/some/where/in/system/msg.file")
     */
    arg_0 = check_mode[0];
    arg_1 = check_mode[1];
    arg_2 = check_mode[2];
    arg_3 = check_mode[3];
    arg_4 = check_mode[4];

    /*
     * If `pw_invalidpw'(=arg_4) is not defined yet,
     * we use `pck.path.invalidpw' defined in `pck_entrance()'.
     */
    if ( !(arg_4 && *arg_4) ) {
      if ( pw_check_debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: pw_invalidpw(arg_4) [%s]", arg_4);
      arg_4 = pck_expand_tilde((char*)pck.path.invalidpw);
      if ( pw_check_debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: pw_invalidpw(arg_4) [%s]", arg_4);
    }

    pck.path.invalidpw = arg_4;
    
    log_pri (L_INFO, "%s mode  [%s] [%s] [%s] [%s] [%s]",
	   logging_mark,
	   arg_0, arg_1, arg_2, arg_3, arg_4);

#ifdef PW_CHECK_FILE
    /*
     * no-more checking uses `pck_proof_file' module and returns
     * NO_MORE_CHECK, CHECKED_ONCE_FILE, ACCEPT/REJECT flags.
     */
    if (no_more) {
      log_pri (L_INFO, "%s (NMC) no-more check executing...",
	     logging_mark);
      valid |= NONE_DEFINED;
      valid_save = valid;
      valid = pck_proof_file ( passwd, arg_2, arg_4, valid );

      if ( pw_check_debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: flag   [%d->%d]",valid_save,valid);

      if (valid & ACCEPT) {
	log_pri (L_INFO, "%s (NMC) ACCEPTED.  password -> [%s]",
	       logging_mark, passwd);
      } else {
	log_pri (L_INFO, "%s (NMC) REJECTED.  password -> [%s]",
	       logging_mark, passwd);
      }
    }
    valid_save = valid;
#endif /* end of PW_CHECK_FILE */

    if ( pw_check_debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: actual check starts here.");

    if ( !strcasecmp(arg_0, "file" )) {
#ifdef PW_CHECK_FILE
      valid = pck_proof_file ( passwd, arg_2, arg_4, valid );
#endif
    } else if ( !strcasecmp(arg_0, "dns" )) {
#ifdef PW_CHECK_DNS
      valid = pck_proof_dns  ( passwd, arg_2, arg_4, valid );
#endif
    } else if ( !strcasecmp(arg_0, "file+dns" )) {
#ifdef PW_CHECK_FILE
      valid = pck_proof_file ( passwd, arg_2, arg_4, valid );
#endif
#ifdef PW_CHECK_DNS
      valid = pck_proof_dns  ( passwd, arg_2, arg_4, valid );
#endif
    } else if ( !strcasecmp(arg_0, "dns+file" )) {
#ifdef PW_CHECK_DNS
      valid = pck_proof_dns  ( passwd, arg_2, arg_4, valid );
#endif
#ifdef PW_CHECK_FILE
      valid = pck_proof_file ( passwd, arg_2, arg_4, valid );
#endif
    }
    if ( strcasecmp(arg_0, "file")     &&
	 strcasecmp(arg_0, "dns")      &&
	 strcasecmp(arg_0, "file+dns") &&
	 strcasecmp(arg_0, "dns+file") ) {
      log_pri (L_ERROR, "ERR: NO PASSWORD-CHECK DONE !! (invalid mode)");
      log_pri (L_ERROR, "ERR: check mode = [%s] ??",arg_0);
      log_pri (L_ERROR, "ERR: CHECK the 1st ARGUMENT of \"PCKcheckmodehost/class\"");
    } else {
      log_pri (L_INFO, "%s password-check done", logging_mark);
    }
  } else {
    /*
     * `pck_need_to_check' module gives NO_NEED_TO_CHECK flag.
     * We allow login in this case.
     */
    if( !strcasecmp(pck.index, "class") ) {
      log_pri (L_INFO, "%s access class [%s]",
	     logging_mark, pck.remote.class);
      log_pri (L_INFO, "%s no password-check done",
	     logging_mark);
    } else {
      log_pri (L_INFO, "%s remote host  [%s]",
	     logging_mark, pck.remote.name);
      log_pri (L_INFO, "%s no password-check done",
	     logging_mark);
    }
  }

  /*
   * Any other flags besides ACCEPT/REJECT should be stripped off
   * here and final result(valid) of this password-check routines
   * is decided here.
   */
  if ( pw_check_debug & PCK_DEBUG_FULL )
    log_pri (L_TRACE, "---: flag_1 [%d->%d]",valid_save,valid);
  valid_save = valid;
  valid &= ~NONE_DEFINED;
  valid &= ~NO_MORE_CHECK;
  valid &= ~CHECKED_ONCE_DNS;
  valid &= ~CHECKED_ONCE_FILE;
  if ( pw_check_debug & PCK_DEBUG_FULL )
    log_pri (L_TRACE, "---: flag_2 [%d->%d]",valid_save,valid);
  /*
   * Return value is the final result:   1 (accept) or 0 (reject)
   */
  return(valid);
}

/*
 * Copyright (c) 1989, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Guido van Rossum.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)fnmatch.c	8.2 (Berkeley) 4/16/94";
#endif /* LIBC_SCCS and not lint */

/*
 * Function fnmatch() as specified in POSIX 1003.2-1992, section B.6.
 * Compares a filename or pathname to a pattern.
 */

#include <ctype.h>
#include <stdio.h>
#ifdef BSD
#include <strings.h>
#else
#include <string.h>
#endif

#define	FNM_NOMATCH	1	/* Match failed. */

#ifndef FNM_PATHNAME
#define	FNM_PATHNAME	0x01	/* Slash must be matched by slash. */
#endif
#ifndef	FNM_NOESCAPE
#define	FNM_NOESCAPE	0x02	/* Disable backslash escaping. */
#endif
#ifndef FNM_PERIOD
#define	FNM_PERIOD	0x04	/* Period must be matched by period. */
#endif
#ifndef FNM_PERIOD2
#define FNM_PERIOD2     0x08
#endif

#ifndef _POSIX_SOURCE
#ifndef FNM_LEADING_DIR
#define FNM_LEADING_DIR 0x08    /* Ignore /<tail> after Imatch. */
#endif
#ifndef FNM_CASEFOLD
#define FNM_CASEFOLD    0x10    /* Case insensitive search. */
#endif
#endif

#define	EOS	'\0'

static char *
pck_rangematch(char *pattern, char test, int flags)
{
	int negate, ok;
	char c, c2;

	/*
	 * A bracket expression starting with an unquoted circumflex
	 * character produces unspecified results (IEEE 1003.2-1992,
	 * 3.13.2).  This implementation treats it like '!', for
	 * consistency with the regular expression syntax.
	 * J.T. Conklin (conklin@ngai.kaleida.com)
	 */
	if ( (negate = (*pattern == '!' || *pattern == '^')) )
		++pattern;

#ifdef FNM_CASEFOLD
	if (flags & FNM_CASEFOLD)
		test = tolower((unsigned char)test);
#endif
	for (ok = 0; (c = *pattern++) != ']';) {
		if (c == '\\' && !(flags & FNM_NOESCAPE))
			c = *pattern++;
		if (c == EOS)
			return (NULL);

#ifdef FNM_CASEFOLD
		if (flags & FNM_CASEFOLD)
			c = tolower((unsigned char)c);
#endif
		if (*pattern == '-'
		    && (c2 = *(pattern+1)) != EOS && c2 != ']') {
			pattern += 2;
			if (c2 == '\\' && !(flags & FNM_NOESCAPE))
				c2 = *pattern++;
			if (c2 == EOS)
				return (NULL);

#ifdef FNM_CASEFOLD
			if (flags & FNM_CASEFOLD)
				c2 = tolower((unsigned char)c2);
#endif
#ifdef notdef
			if (   __collate_range_cmp(c, test) <= 0
			    && __collate_range_cmp(test, c2) <= 0
			       )
#else
			  /* this is a hack */
			  if ((c <= test) && (test <= c2))
#endif
				ok = 1;
		} else if (c == test)
			ok = 1;
	}
	return (ok == negate ? NULL : pattern);
}


int
pck_fnmatch(char *pattern, char *string, int flags)
{

  char *stringstart;
  char c, test;

  for (stringstart = string;;)
    switch (c = *pattern++) {
    case EOS:
#ifdef FNM_LEADING_DIR
      if ((flags & FNM_LEADING_DIR) && *string == '/')
	return (0);
#endif
      return (*string == EOS ? 0 : FNM_NOMATCH);
    case '?':
      if (*string == EOS)
	return (FNM_NOMATCH);
      if (*string == '/' && (flags & FNM_PATHNAME))
	return (FNM_NOMATCH);
      if (*string == '.' && (flags & FNM_PERIOD2))
	return (FNM_NOMATCH);
      if (*string == '.' && (flags & FNM_PERIOD) &&
	  (string == stringstart ||
	   ((flags & FNM_PATHNAME) && *(string - 1) == '/')))
	return (FNM_NOMATCH);
      ++string;
      break;
    case '*':
      c = *pattern;
      /* Collapse multiple stars. */
      while (c == '*')
	c = *++pattern;
      
      if (*string == '.' && (flags & FNM_PERIOD) &&
	  (string == stringstart ||
	   ((flags & FNM_PATHNAME) && *(string - 1) == '/')))
	return (FNM_NOMATCH);
      
      /* Optimize for pattern with * at end or before /. */
      if (c == EOS)
	if (flags & FNM_PATHNAME)
#ifdef FNM_LEADING_DIR
	  return ((flags & FNM_LEADING_DIR) ||
		  strchr(string, '/') == NULL ?
		  0 : FNM_NOMATCH);
#else
      return (strchr(string, '/') == NULL ?
	      0 : FNM_NOMATCH);
#endif
	else
	  return (0);
      else if (c == '/' && flags & FNM_PATHNAME) {
	if ((string = strchr(string, '/')) == NULL)
	  return (FNM_NOMATCH);
	break;
      }
      
      /* General case, use recursion. */
      while ((test = *string) != EOS) {
	if (!pck_fnmatch(pattern, string, flags & ~FNM_PERIOD))
	  return (0);
	if (test == '/' && flags & FNM_PATHNAME)
	  break;
	++string;
      }
      return (FNM_NOMATCH);
    case '[':
      if (*string == EOS)
	return (FNM_NOMATCH);
      if (*string == '/' && flags & FNM_PATHNAME)
	return (FNM_NOMATCH);
      if ((pattern =
	   pck_rangematch(pattern, *string, flags)) == NULL)
	return (FNM_NOMATCH);
      ++string;
      break;
    case '\\':
      if (!(flags & FNM_NOESCAPE)) {
	if ((c = *pattern++) == EOS) {
	  c = '\\';
	  --pattern;
	}
      }
      /* FALLTHROUGH */
    default:
      if (c == *string)
	;
#ifdef FNM_CASEFOLD
      else if ((flags & FNM_CASEFOLD) &&
	       (tolower((unsigned char)c) ==
		tolower((unsigned char)*string)))
	;
#endif
      else
	return (FNM_NOMATCH);
      string++;
      break;
    }
  /* NOTREACHED */
}

char *autospout = NULL;
int autospout_free = 0;
int dolreplies = 1;
int restricted_user = 0;
char *home;
int debug = 0;

void pck_reply(int n, char *fmt,...)
{
    va_list ap = NULL;	/* Avoid compiler warning */

    if (autospout != NULL) {	/* deal with the autospout stuff... */
      char *p, *ptr = autospout;

      while (*ptr) {
	if ((p = strchr(ptr, '\n')) != NULL)	/* step through line by line */
	  *p = '\0';

	/* send a line...(note that this overrides dolreplies!) */
	pck_vreply(USE_REPLY_LONG | USE_REPLY_NOTFMT, n, ptr, ap);

	if (p)
	  ptr = p + 1;	/* set to the next line... (\0 is handled in the while) */
	else
	  break;		/* oh, we're done; drop out of the loop */
      }

      if (autospout_free) {	/* free autospout if necessary */
	(void) free(autospout);
	autospout_free = 0;
      }
      autospout = 0;		/* clear the autospout */
    }

    va_start(ap, fmt);

    /* send the reply */
    pck_vreply(0L, n, fmt, ap);

    va_end(ap);
}

void pck_lreply(int n, char *fmt,...)
{
  va_list ap;

    if (!dolreplies)	/* prohibited from doing long replies? */
      return;

  va_start(ap, fmt);

    /* send the reply */
    pck_vreply(USE_REPLY_LONG, n, fmt, ap);

  va_end(ap);
}

void pck_perror_reply(int code, char *string)
{
  /*
   * If restricted user and string starts with home dir path, strip it off
   * and return only the relative path.
   */
  if (restricted_user && (home != NULL) && (home[0] != '\0')) {
    size_t len = strlen (home);
    if (strncmp (home, string, len) == 0) {
      if (string[len - 1] == '/')
	string += len - 1;
      else if (string[len] == '/')
	string += len;
      else if (string[len] == '\0')
	string = "/";
    }
  }
  pck_reply(code, "%s: %s.", string, strerror(errno));
}


void pck_vreply(long flags, int n, char *fmt, va_list ap)
{
    char buf[BUFSIZ];

    flags &= USE_REPLY_NOTFMT | USE_REPLY_LONG;

    if (n)			/* if numeric is 0, don't output one; use n==0 in place of printf's */
      sprintf(buf, "%03d%c", n, flags & USE_REPLY_LONG ? '-' : ' ');

    /* This is somewhat of a kludge for autospout.  I personally think that
     * autospout should be done differently, but that's not my department. -Kev
     */
    if (flags & USE_REPLY_NOTFMT)
	snprintf(buf + (n ? 4 : 0), n ? sizeof(buf) - 4 : sizeof(buf), "%s", fmt);
    else
	vsnprintf(buf + (n ? 4 : 0), n ? sizeof(buf) - 4 : sizeof(buf), fmt, ap);

    if (debug)			/* debugging output :) */
      log_pri (L_TRACE, "<--- %s", buf);

    /* Yes, you want the debugging output before the client output; wrapping
     * stuff goes here, you see, and you want to log the cleartext and send
     * the wrapped text to the client.
     */

    printf("%s\r\n", buf);	/* and send it to the client */
#ifdef TRANSFER_COUNT
    byte_count_total += strlen(buf);
    byte_count_out += strlen(buf);
#endif
    fflush(stdout);
}

void pck_upper(char *s)
{
    while (*s != '\0') {
	if (islower(*s))
	    *s = toupper(*s);
	s++;
    }
}

/*************************************************************************/
/* FUNCTION  : pck_pr_mesg                                               */
/* PURPOSE   : Display a message to the user                             */
/* ARGUMENTS : message code, name of file to display                     */
/*************************************************************************/

void pck_pr_mesg(int msgcode, char *msgfile)
{
    FILE *infile;
    char inbuf[1024], outbuf[1024], *cr;

    if (msgfile && (int) strlen(msgfile) > 0) {
	infile = fopen(msgfile, "r");
	if (infile) {
	    while (fgets(inbuf, sizeof(inbuf), infile) != NULL) {
		if ((cr = strchr(inbuf, '\n')) != NULL)
		    *cr = '\0';
		/*pck_msg_massage(inbuf, outbuf, sizeof(outbuf));*/
		(void) strcpy(outbuf,inbuf);
		pck_lreply(msgcode, "%s", outbuf);
	    }
	    fclose(infile);
	}
    }
}

#if defined(PW_CHECK_DNS)

typedef union {
    HEADER qb1;
    u_char qb2[PACKETSZ];
} querybuf;

/****************************************************************************
 * pck_lookup_name()
 *   This routine takes a FQDN and tries to find a valid IP address.
 *   If a CNAME is returned, it looks further on in the reply for an
 *   A record. It will return as soon as it finds an A record, so hosts
 *   with multiple A records will have the first A record returned.
 ***************************************************************************/
int pck_lookup_name(char *name, char **ip)
{
#ifndef NO_DNS
    u_char *msg_end, *msg_ptr;
#ifdef USE_RES_SEND
    querybuf question, answer;
#else
    querybuf answer;
#endif
    int rc, num_answers, num_query, q_type, q_class, q_ttl, q_len;
    char query_name[MAXDNAME + 1];
    struct in_addr inaddr;
    char *result;

    /* res_mkquery+res_send seem to function identically to res_query, so I'm */
    /* using res_query. If there's some advantage to using res_mkquery that I */
    /* don't know about (which wouldn't surprise me :) then you can just      */
    /* define USE_RES_SEND and it will use that method                        */

#ifdef USE_RES_SEND
    rc = res_mkquery(QUERY, name, C_IN, T_A, NULL, 0, NULL, question.qb2, sizeof(querybuf));

    if (rc < 0) {
	return FALSE;
    }

    rc = res_send(question.qb2, rc, answer.qb2, sizeof(querybuf));
#else
    rc = res_query(name, C_IN, T_A, answer.qb2, sizeof(querybuf));
#endif

    if (rc < 0) {
	return FALSE;
    }

    msg_end = (u_char *) & answer + rc;

    num_answers = ntohs(answer.qb1.ancount);
    num_query = ntohs(answer.qb1.qdcount);

    if (num_answers < 1) {
	/* No answers mean that this hostname doesn't exist..                 */
	return FALSE;
    }

    msg_ptr = answer.qb2 + HFIXEDSZ;

    /* skip over the query */
    for (; (num_query > 0) && (msg_ptr < msg_end); num_query--) {
	msg_ptr += dn_skipname(msg_ptr, msg_end) + QFIXEDSZ;
    }

    for (; (num_answers > 0) && (msg_ptr < msg_end); num_answers--) {

	rc = dn_expand(answer.qb2, msg_end, msg_ptr, (void *) query_name, MAXDNAME);

	msg_ptr += rc;

	q_type = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	q_class = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	q_ttl = _getlong(msg_ptr);
	msg_ptr += INT32SZ;

	q_len = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	/* look at the type of response that we recieved. If it's a CNAME then */
	/* we need to find out what the CNAME's IP is.                         */
	switch (q_type) {
	case T_CNAME:
	    /* Got a CNAME - hope that there are other answers with the A */
	    msg_ptr += q_len;
	    break;
	case T_A:
	    bcopy(msg_ptr, (char *) &inaddr, INADDRSZ);
	    msg_ptr += q_len;
	    result = (char *) inet_ntoa(inaddr);

	    *ip = (char *) malloc(strlen(result) + 1);
	    if (*ip == NULL)
		return FALSE;
	    strcpy(*ip, result);
	    return TRUE;
	default:
	    /* what the ?? - we don't expect this response type */
	    return FALSE;
	}
    }

    /* oh dear, there was no A's in the response */
#endif /* !NO_DNS */
    return FALSE;
}

/****************************************************************************
 * pck_lookup_ip()
 *   This routine takes an IP address in the format a.b.c.d and returns the
 *   hostname.
 ***************************************************************************/
int pck_lookup_ip(char *ip, char **fqdn)
{
#ifndef NO_DNS
    u_char *msg_end, *msg_ptr;
#ifdef USE_RES_SEND
    querybuf question, answer;
#else
    querybuf answer;
#endif
    int rc, num_answers, num_query, q_type, q_class, q_ttl, q_len;
    char query_name[MAXDNAME + 1];
    char in_addr[MAXDNAME + 1];
    unsigned int a, b, c, d;

    a = b = c = d = 0;
    sscanf(ip, "%u.%u.%u.%u", &a, &b, &c, &d);
    sprintf(in_addr, "%u.%u.%u.%u.in-addr.arpa", d, c, b, a);

#ifdef USE_RES_SEND
    rc = res_mkquery(QUERY, in_addr, C_IN, T_PTR, NULL, 0, NULL, question.qb2, sizeof(querybuf));

    if (rc < 0) {
	return FALSE;
    }

    rc = res_send(question.qb2, rc, answer.qb2, sizeof(querybuf));
#else
    rc = res_query(in_addr, C_IN, T_PTR, answer.qb2, sizeof(querybuf));
#endif

    if (rc < 0) {
	return FALSE;
    }

    msg_end = (u_char *) & answer + rc;

    num_answers = ntohs(answer.qb1.ancount);
    num_query = ntohs(answer.qb1.qdcount);

    if (num_answers < 1) {
	/* No answers mean that this hostname doesn't exist..                 */
	return FALSE;
    }

    msg_ptr = answer.qb2 + HFIXEDSZ;

    /* skip over the query */
    for (; (num_query > 0) && (msg_ptr < msg_end); num_query--) {
	msg_ptr += dn_skipname(msg_ptr, msg_end) + QFIXEDSZ;
    }

    for (; num_answers > 0 && (msg_ptr < msg_end); num_answers--) {

	rc = dn_expand(answer.qb2, msg_end, msg_ptr, (void *) query_name, MAXDNAME);

	/* increment our pointer */
	msg_ptr += rc;

	/* read off the various answer information */
	q_type = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	q_class = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	q_ttl = _getlong(msg_ptr);
	msg_ptr += INT32SZ;

	q_len = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	switch (q_type) {
	case T_PTR:
	    /* this is our answer, expand the name, allocate space for it and */
	    /* set the return information                                     */
	    rc = dn_expand(answer.qb2, msg_end, msg_ptr, (void *) query_name, MAXDNAME);

	    if (rc >= 0) {
		*fqdn = (char *) malloc(strlen(query_name) + 1);
		if (*fqdn == NULL)
		    return FALSE;

		strcpy(*fqdn, query_name);
		return TRUE;
	    }
	    break;
	default:
	    /* unknown response type.. keep looking */
	    msg_ptr += q_len;
	}
    }

#endif /* !NO_DNS */

    return FALSE;
}

/****************************************************************************
 * pck_check_name_for_ip()
 *   This routine checks to see if a given IP address is a valid IP for a
 *   given name. We need this because pck_lookup_name() only returns the first
 *   IP address in the response, and if a user is coming from a different
 *   IP address for the same machine, we want to make sure that we match
 *   them.
 ***************************************************************************/
int pck_check_name_for_ip(char *name, char *ip)
{
#ifndef NO_DNS
#ifdef USE_RES_SEND
    querybuf question, answer;
#else
    querybuf answer;
#endif
    u_char *msg_end, *msg_ptr;
    char query_name[MAXDNAME + 1];
    int rc, num_answers, num_query, q_type, q_class, q_ttl, q_len;
    struct in_addr inaddr, qaddr;

    /* convert the passed IP address into an in_addr (for comparison later) */
    /* not all systems have inet_aton(), so use inet_addr() */
    if ((qaddr.s_addr = inet_addr(ip)) == (u_long) - 1) {
	return FALSE;
    }

#ifdef USE_RES_SEND
    rc = res_mkquery(QUERY, name, C_IN, T_A, NULL, 0, NULL, question.qb2, sizeof(querybuf));

    if (rc < 0) {
	return FALSE;
    }

    rc = res_send(question.qb2, rc, answer.qb2, sizeof(querybuf));
#else
    rc = res_query(name, C_IN, T_A, answer.qb2, sizeof(querybuf));
#endif

    if (rc < 0) {
	return FALSE;
    }

    msg_end = (u_char *) & answer + rc;

    num_answers = ntohs(answer.qb1.ancount);
    num_query = ntohs(answer.qb1.qdcount);

    if (num_answers < 1) {
	/* No answers mean that this hostname doesn't exist..                 */
	return FALSE;
    }

    msg_ptr = answer.qb2 + HFIXEDSZ;

    /* skip over the query */
    for (; (num_query > 0) && (msg_ptr < msg_end); num_query--) {
	msg_ptr += dn_skipname(msg_ptr, msg_end) + QFIXEDSZ;
    }

    for (; (num_answers > 0) && (msg_ptr < msg_end); num_answers--) {
	rc = dn_expand(answer.qb2, msg_end, msg_ptr, (void *) query_name, MAXDNAME);

	msg_ptr += rc;

	q_type = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	q_class = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	q_ttl = _getlong(msg_ptr);
	msg_ptr += INT32SZ;

	q_len = _getshort(msg_ptr);
	msg_ptr += INT16SZ;

	/* look at the type of response that we recieved. If it's a CNAME then */
	/* we need to find out what the CNAME's IP is.                         */
	switch (q_type) {
	case T_CNAME:
	    /* Got a CNAME - hope that there are other answers with the A */
	    msg_ptr += q_len;
	    break;
	case T_A:
	    bcopy(msg_ptr, (char *) &inaddr, INADDRSZ);
	    msg_ptr += q_len;
	    if (memcmp(&inaddr, &qaddr, sizeof(struct in_addr)) == 0) {
		return TRUE;
	    }
	    break;
	default:
	    /* what the ?? - we don't expect this response type */
	    return FALSE;
	}
    }

#endif /* !NO_DNS */

    /* no matching IP's */
    return FALSE;
}

#endif /* end of "defined(PW_CHECK_DNS)" */


/*
 * PCK initialization
 */
void pck_init(struct conn_struc  *config, 
	      struct class_struc *class)
{
  config_rec *pck_tmp;
  char *temp_pointer;
  
  /*
   * "PCK" directive is alive ?
   */
  switch(get_param_int(main_server->conf,"PCK",FALSE))
    {
    case FALSE: 
      pck.enable = FALSE; 
      break;
    case TRUE: 
      pck.enable = TRUE; 
      break;
    default: 
      pck.enable = FALSE;	/* By default, disable PCK checking */
      break;
    }
  
  if( !pck.remote.anonname || !pck.enable )	/* anonymous and "PCK" is true? */
    return;

  /*
   * Initialize debug parameter for `password-check'
   */
  if((pck_tmp = find_config( main_server->conf,
			     CONF_PARAM, 
			     "PCKdebug", 
			     FALSE)) != NULL &&
     !pck_tmp->argv[0]) {
#if defined(PCK_AY_DEBUG)	/* (Debugging purpose:@@) */
    if(pck.enable)
      log_pri (L_INFO, "PCK: YES");
    else
      log_pri (L_INFO, "PCK: NO");
#endif
    if( pck_tmp->argc > 1 ) {
      pck.debug = PCK_DEBUG_NONE;
      if(!strcasecmp(pck_tmp->argv[1], "level-0") ||
	 !strcasecmp(pck_tmp->argv[1], "none"))   {
	pck.debug = PCK_DEBUG_NONE;
      } else if(!strcasecmp(pck_tmp->argv[1], "level-1") ||
		!strcasecmp(pck_tmp->argv[1], "normal")) {
	pck.debug = PCK_DEBUG_NORMAL;
      } else if(!strcasecmp(pck_tmp->argv[1], "level-2") ||
		!strcasecmp(pck_tmp->argv[1], "high"))   {
	pck.debug = PCK_DEBUG_HIGH;
      } else if(!strcasecmp(pck_tmp->argv[1], "level-3") ||
		!strcasecmp(pck_tmp->argv[1], "full"))   {
	pck.debug = PCK_DEBUG_FULL;
      } else {
	log_pri (L_ERROR, "ERR: Unrecognized Debug Level? [%s]", 
	       (char *)pck_tmp->argv[1]);
	log_pri (L_ERROR, "ERR: Debug Level set to \"none\"");
	pck.debug = PCK_DEBUG_NONE;
      }
    } else {
      pck.debug = PCK_DEBUG_NORMAL;	/* Default debug level: normal! */
    }
  } else {
    /* 
     * By default, "PW_CHECK_DEBUG" flag gives debugging syslogs in
     * "normal" level.  A `PCKdebug' directive of "proftpd.conf"
     * overrides this and we can control log level dynamically
     * from "proftpd.conf".  
     */
#if defined(PW_CHECK_DEBUG)
    pck.debug = PCK_DEBUG_NORMAL;
#else
    pck.debug = PCK_DEBUG_NONE;
#endif
  }
  /*
   * Turn on debug bit against lower levels!
   */
  if( pck.debug & PCK_DEBUG_FULL ) {
    pck.debug |= PCK_DEBUG_NORMAL;
    pck.debug |= PCK_DEBUG_HIGH;
  } else if( pck.debug & PCK_DEBUG_HIGH ) {
    pck.debug |= PCK_DEBUG_NORMAL;
  }

  if( pck.debug ) {
    if(pck.enable)
      log_pri (L_INFO, "PCK: YES");
    else
      log_pri (L_INFO, "PCK: NO");
    if( pck.debug & PCK_DEBUG_FULL ) {
      log_pri (L_INFO, "PCK: DEBUG ON (LEVEL: FULL )");
      log_pri (L_INFO, "PCK: Module ID: [%s]", pck_id);
      log_pri (L_INFO, "PCK: Author Name:  [%s]", pck_author_name);
      log_pri (L_INFO, "PCK: Author Email: [%s]", pck_author_addr);
    } else if( pck.debug & PCK_DEBUG_HIGH ) {
      log_pri (L_INFO, "PCK: DEBUG ON (LEVEL: HIGH )");
    } else if( pck.debug & PCK_DEBUG_NORMAL ) {
      log_pri (L_INFO, "PCK: DEBUG ON (LEVEL: NORMAL )");
    }
    if( pck.debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: DEBUG LEVEL [%d]",pck.debug);
  }
  
  pw_check_debug = pck.debug;

  /*
   * Initialize local hostname! (stored in "pck.local.name")
   */
#ifdef HAVE_SYSINFO
  sysinfo(SI_HOSTNAME, pck.local.name, sizeof(pck.local.name));
#else
  (void) gethostname((char *)pck.local.name, sizeof(pck.local.name));
#endif

  /* Save remote name */
  temp_pointer    = config->remote_name;
  pck.remote.name = (char *) malloc(strlen(temp_pointer) + 1);
  if( pck.remote.name == NULL ) {
    pck.remote.name = "Remote_Name";
  } else {
    (void) strcpy((char *)pck.remote.name, temp_pointer);
  }

  /* Save remote IP */
  temp_pointer  = inet_ntoa(*config->remote_ipaddr);
  pck.remote.ip = (char *) malloc(strlen(temp_pointer) + 1);
  if( pck.remote.ip == NULL ) {
    pck.remote.ip = "0.0.0.0";
  } else {
    (void) strcpy((char *)pck.remote.ip, temp_pointer);
  }

  /* Save local IP */
  temp_pointer  = inet_ntoa(*config->local_ipaddr);
  pck.local.ip  = (char *) malloc(strlen(temp_pointer) + 1);
  if( pck.local.ip == NULL ) {
    pck.local.ip = "0.0.0.0";
  } else {
    (void) strcpy((char *)pck.local.ip, temp_pointer);
  }
  
  if( pck.debug & PCK_DEBUG_HIGH ) {
    log_pri (L_TRACE, "---: local_name  = [%s]",pck.local.name);
    log_pri (L_TRACE, "---: local_ip    = [%s]",pck.local.ip);
    log_pri (L_TRACE, "---: remote_name = [%s]",pck.remote.name);
    log_pri (L_TRACE, "---: remote_ip   = [%s]",pck.remote.ip);
  }
  
  if( pck.debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, (ServerType)? 
	   "---: ServerType   [standalone]" :
	   "---: ServerType   [inetd]");
  
  if( pck.debug & PCK_DEBUG_HIGH ) {
    log_pri (L_TRACE, "---: local__name [%s]", pck.local.name);
    log_pri (L_TRACE, "---: remote_name [%s]", pck.remote.name);
  }
  
  /*
   * `pck.local.cname' is defined here.
   * If `PCKlocalcname' directive is not found in `conf',
   * `hostname' is set as `pck.local.cname'.
   */
  if(get_param_int(main_server->conf,"PCKlocalcname",FALSE) != -1) {
    pck_tmp = find_config(main_server->conf, 
			  CONF_PARAM, 
			  "PCKlocalcname", 
			  TRUE);
    if(pck_tmp->argv[0]) {
      pck.local.cname = pck_tmp->argv[0];
    }
  }
  
  if( !(pck.local.cname && *pck.local.cname )) {
    pck.local.cname = pck.local.name;
  }
  
  if( pck.debug & PCK_DEBUG_HIGH ) {
    log_pri (L_TRACE, "---: local_cname   is [%s]", pck.local.cname);
    log_pri (L_TRACE, "---: hostname      is [%s]", pck.local.name);
  }

  /*
   * `pck.path.msg_prefix' file is defined here.
   * If `PCKmsgprefix' directive is not found in `conf',
   * _PATH_MSGPREFIX in pwcheck.h is set as `pck.path.msg_prefix'.
   */
  pck.msg_prefix = FALSE;
  
  if((pck_tmp = find_config( main_server->conf,
			     CONF_PARAM, 
			     "PCKmsgprefix", 
			     FALSE)) != NULL &&
     !pck_tmp->argv[0]) {
    
    pck.msg_prefix = TRUE;
    if( pck_tmp->argc > 1 ) {
      pck.path.msg_prefix = (char*)pck_tmp->argv[1];
    } else {
      pck.path.msg_prefix = _PATH_MSGPREFIX;
    }
  }
  
  if( !(pck.path.msg_prefix && *pck.path.msg_prefix )) {
    pck.path.msg_prefix = _PATH_MSGPREFIX;
  }

  if( pck.debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: PCKmsgprefix is [%s]", pck.path.msg_prefix);

  /*
   * `pck.path.display_login' file is defined here.
   * If `PCKdisplaylogin' directive is not found in `conf',
   * _PATH_DISPLAYLOGIN in pwcheck.h is set as `pck.path.display_login'.
   */
  pck.display_login = FALSE;

  if((pck_tmp = find_config( main_server->conf,
			     CONF_PARAM, 
			     "PCKdisplaylogin", 
			     FALSE)) != NULL &&
     !pck_tmp->argv[0]) {

    pck.display_login = TRUE;
    if( pck_tmp->argc > 1 ) {
      pck.path.display_login = pck_expand_tilde((char*)pck_tmp->argv[1]);
    } else {
      pck.path.display_login = pck_expand_tilde(_PATH_DISPLAYLOGIN);
    }
  }

  if( !(pck.path.display_login && *pck.path.display_login )) {
    pck.path.display_login = pck_expand_tilde(_PATH_DISPLAYLOGIN);
  }

  if( pck.debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: PCKdisplaylogin is [%s]", pck.path.display_login);

  /*
   * `pck.path.dump' file is defined here.
   * If `PCKdump' directive is not found in `conf',
   * _PATH_DUMP in pwcheck.h is set as `pck.path.dump'.
   */
  pck.dump = FALSE;

  if((pck_tmp = find_config( main_server->conf,
			     CONF_PARAM, 
			     "PCKdump", 
			     FALSE)) != NULL &&
     !pck_tmp->argv[0]) {

    pck.dump = TRUE;
    if( pck_tmp->argc > 1 ) {
      pck.path.dump = pck_expand_tilde((char*)pck_tmp->argv[1]);
    } else {
      pck.path.dump = pck_expand_tilde(_PATH_DUMP);
    }
  }

  if( !(pck.path.dump && *pck.path.dump )) {
    pck.path.dump = pck_expand_tilde(_PATH_DUMP);
  }

  if( pck.debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: PCKdump is [%s]", pck.path.dump);

  /*
   * Define user's access class here.
   */
  if((pck.remote.class_defined = get_param_int(main_server->conf,
					       "Classes",
					       FALSE)) < 0)
    pck.remote.class_defined = 0;
  
  if(pck.remote.class_defined) {
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: Original-class: [%s]", 
	     class->name);

    /* Save remote class */
    temp_pointer     = class->name;
    pck.remote.class = (char *) malloc(strlen(temp_pointer) + 1);
    if( pck.remote.class == NULL ) {
      pck.remote.class = "Default";
    } else {
      (void) strcpy((char *)pck.remote.class, temp_pointer);
    }
    (void) pck_upper((char *)pck.remote.class);
    
    if( pck.debug & PCK_DEBUG_NORMAL )
      log_pri (L_TRACE, "---: ACCESS-CLASS: [%s]", 
	     pck.remote.class);
  } else {
    if( pck.debug & PCK_DEBUG_NORMAL ) {
      log_pri (L_ERROR, "WRN: I can not define remote access class!!");
      log_pri (L_ERROR, "WRN: If you want to do class-based access control,");
      log_pri (L_ERROR, "WRN: Set \"Classes on\" in your proftpd.conf.");
    }
    log_pri (L_INFO, "PCK: Access Class [(not defined)]");
    return;
  }

  log_pri (L_INFO, "PCK: Access Class [%s]", pck.remote.class);
  
  return;

}

/*
 * PCK core part to block non-legit users out.
 *
 * Return value (int):
 *               Not_Valid.....deny login
 *               Valid ........accept login
 */
int pck_core_handler(struct conn_struc  *config, 
		     struct class_struc *class,
		     config_rec *c)
{
  config_rec	*pck_tmp;
  int		max_check = 5;
  int		i,count;
  char		*temp;

  /*
   * Initialize PCK and some parameters.
   */
  (void) pck_init(config, class);
  
  /*
   * If non-anonymous login, quit PCK here.
   */
  if( !pck.remote.anonname ) {
    if( pck.enable )
      log_pri (L_ERROR, "ERR: Quit PCK. (Non-Anonymous login?)");
    return Valid;
  }

  /*
   * If class-name is not defined, quit PCK here.
   */
#if 0
  if( !pck.remote.class_defined ) {
    if( pck.enable )
      log_pri (L_ERROR, "ERR: Quit PCK. (Class-Name not defined?)");
    return Valid;
  }
#endif

  /*
   * If "PCKcheckorder" is included in "proftpd.conf",
   * each check should be done in the specified order
   * by entries in "PCKcheckorder".
   *
   * If not, each check will be done according to the order
   * initialized by "pwcheck_child_init()"
   * (see pck.check_order[][])
   */
  pck_tmp = find_config(main_server->conf,
			CONF_PARAM,
			"PCKcheckorder",
			TRUE);
  
  count = 0;
  
  for( i=0; i<max_check; i++ ) {	/* for-loop-1 */
    
    /*
     * When each entry begins with "!", we ignore this check.
     * This never fails to override the entries specified by
     * each directive in proftpd.conf.
     *
     * Example:
     *
     * PCKcheckorder  dead  !reverse  !match  goaway  password
     *
     */
    if( pck_tmp ) {
      temp = pck_tmp->argv[i];
      if( pck.debug & PCK_DEBUG_NORMAL )
	log_pri (L_TRACE, "---: Check Entry [%s]",temp);
      if( *temp=='!' ) {
	count++;
	while( *temp=='!' ) temp++;
	if( pck.debug & PCK_DEBUG_FULL )
	  log_pri (L_TRACE, "---: !_match CHECK NOT DONE (%s)...", temp);
	continue;
      }
    }

    if( ( pck_tmp && !strcasecmp(pck_tmp->argv[i],   "dead")) ||
	(!pck_tmp && !strcasecmp(pck.check_order[i], "dead")) ) {
      (void) strcpy(pck.check_order[i],"DEAD");
      if( pck.debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: >>>>> DEAD check");
      (void) pck_block_out("PCKserverdeadhost", "SDH:", "host");
      if(pck.server_dead.host)
	if(pck_server_dead_host_kickout())
	  return Not_Valid;
      (void) pck_block_out("PCKserverdeadclass", "SDC:", "class");
      if(pck.server_dead.class)
	if(pck_server_dead_class_kickout())
	  return Not_Valid;
    } else if( ( pck_tmp && !strcasecmp(pck_tmp->argv[i],   "goaway")) ||
	       (!pck_tmp && !strcasecmp(pck.check_order[i], "goaway")) ) {
      (void) strcpy(pck.check_order[i],"GOAWAY");
      if( pck.debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: >>>>> GOAWAY check");
      (void) pck_block_out("PCKgoawayhost", "GAH:", "host");
      if(pck.goaway.host)
	if(pck_goaway_host_kickout())
	  return Not_Valid;
      (void) pck_block_out("PCKgoawayclass", "GAC:", "class");
      if(pck.goaway.class)
	if(pck_goaway_class_kickout())
	  return Not_Valid;
    } else if( ( pck_tmp && !strcasecmp(pck_tmp->argv[i],   "reverse")) ||
	       (!pck_tmp && !strcasecmp(pck.check_order[i], "reverse")) ) {
      (void) strcpy(pck.check_order[i],"REVERSE");
      if( pck.debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: >>>>> REVERSE check");
#if defined(PW_CHECK_DNS)
      (void) pck_block_out("PCKreverseDNShost", "RVH:", "host");
      if(!pck_has_reverse_dns())
	if(pck_has_reverse_dns_host_kickout())
	  return Not_Valid;
      (void) pck_block_out("PCKreverseDNSclass", "RVC:", "class");
      if(!pck_has_reverse_dns())
	if(pck_has_reverse_dns_class_kickout())
	  return Not_Valid;	  
#endif
    } else if( ( pck_tmp && !strcasecmp(pck_tmp->argv[i],   "match")) ||
	       (!pck_tmp && !strcasecmp(pck.check_order[i], "match")) ) {
      (void) strcpy(pck.check_order[i],"MATCH");
      if( pck.debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: >>>>> MATCH check");
#if defined(PW_CHECK_DNS)
      (void) pck_block_out("PCKmatchDNShost", "MTH:", "host");
      if(!pck_has_matching_dns())
	if(pck_has_matching_dns_host_kickout())
	  return Not_Valid;
      (void) pck_block_out("PCKmatchDNSclass", "MTC:", "class");
      if(!pck_has_matching_dns())
	if(pck_has_matching_dns_class_kickout())
	  return Not_Valid;
#endif
    } else if( ( pck_tmp && !strcasecmp(pck_tmp->argv[i],   "password")) ||
	       (!pck_tmp && !strcasecmp(pck.check_order[i], "password")) ) {
      (void) strcpy(pck.check_order[i],"PASSWORD");
      if( pck.debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: >>>>> PASSWORD check");
      
      if( !pck_block_pck(c,"PCKcheckmodeclass", "CMC:", "class") )
	return Not_Valid;
      
      if( !pck_block_pck(c,"PCKcheckmodehost",  "CMH:", "host") )
	return Not_Valid;

    } else {
      /*
       * _ONLY_ five entries are allowed here.
       *
       * PCKcheckorder  dead  reverse  match  goaway  password
       *
       */
      if( pck_tmp ) {
	log_pri (L_ERROR, "ERR: Invalid entry in PCKcheckorder? [%s]", 
	       (char *)pck_tmp->argv[i]);
      } else {
	/* NOTREACHED */
	log_pri (L_ERROR, "ERR: Invalid entry in check_order? [%s]", 
	       pck.check_order[i]);
      }
    }
  }	/*end of for-loop-1 */

  if( !(max_check - count) )
    log_pri (L_ERROR, "ERR: No PCK Done? (check your PCKcheckorder)");
  
  log_pri (L_INFO, "PCK: %s",pck.grantmsg); /* accept */
  
  return Valid;
  
}

int pck_block_pck( config_rec *c,
		   char *directive, 
		   char *logging_mark,
		   char *index)
{
  config_rec *pck_tmp;
  const char *dummy = "DUMMY";
  int authcode = Valid;       /* valid = 1 (accept) 0 (reject) */ 
  
  /*
   * If index="class" and class-name is not defined, quit here.
   */
  if( !pck.remote.class_defined && !strcasecmp(index, "class") ) {
    if( pck.enable )
      log_pri (L_ERROR, "%s Quit. Class not defined?",
	     logging_mark);
    return authcode;
  }

  /*
   * Look for specified "directive" in "proftpd.conf".
   */
  pck_tmp = find_config( main_server->conf, 
			 CONF_PARAM,
			 directive,
			 FALSE);
  
  /*
   * When specified "directive" is _NOT_ found,
   * return normally without doing anything.
   * Initialization is already done by pwcheck_child_init().
   */
  if(pck_tmp) {
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: \"%s\" FOUND.", directive);
  } else {
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: \"%s\" NOT FOUND ??", directive);
    pck.index = dummy;	/* Avoid excess syslog messages */
    return authcode;
  }

  pck.index = index;
    
  /*
   * do `password-check' if the following conditions are met,
   *
   * (1) the user tries to gain entry as anonymous login,
   * (2) "AnonRequirePassword" directive is not defined, and
   * (3) "PCK" directive is activated.
   *
   *	 pck.valid = 1 (accept)
   *                 0 (reject) 
   */
  if( pck.remote.anonname              && 
      pck.enable                       &&
      ( pck.debug & PCK_DEBUG_HIGH ) )  {
    if(!c || get_param_int(c->subset,"AnonRequirePassword",FALSE) == 1) {
      log_pri (L_TRACE, "---: AnonRequirePassword is ON");
    } else {
      log_pri (L_TRACE, "---: AnonRequirePassword is OFF");
    }
  }
  
  if( pck.remote.anonname              &&	/* anonymous? */
      pck.enable                       &&	/* "PCK" is true? */
      (c && get_param_int(c->subset,"AnonRequirePassword",FALSE) != 1)) {
    pck_entrance();
    pck.valid = pck_password_check(pck.valid,
				   (char*)pck.remote.pass,
				   pck.check_mode,
				   directive,
				   logging_mark);
    authcode = pck_exit();
  } else if( pck.enable ) {
    log_pri (L_ERROR, "ERR: \"password-check\" not done??");
    log_pri (L_ERROR, "ERR: PCK=on? Anonymous? AnonRequirePassword=off?");
  }

  return authcode;
}

void pck_path_define(char *directive, const char *path)
{

  if(!strcasecmp(directive, "PCKserverdeadclass")) {
    if(path) {
      pck.server_dead.class_path  = pck_expand_tilde((char*)path);
    } else {
      pck.server_dead.class_path  = pck_expand_tilde(_PATH_SERVERDEADCLASS);
    }
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: server_dead_class_path is [%s]", 
	     pck.server_dead.class_path);
  } else if(!strcasecmp(directive, "PCKserverdeadhost")) {
    if(path) {
      pck.server_dead.host_path   = pck_expand_tilde((char*)path);
    } else {
      pck.server_dead.host_path   = pck_expand_tilde(_PATH_SERVERDEADHOST);
    }
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: server_dead_host_path is [%s]", 
	     pck.server_dead.host_path);
  } else if(!strcasecmp(directive, "PCKreverseDNSclass")) {
    if(path) {
      pck.reverse_dns.class_path  = pck_expand_tilde((char*)path);
    } else {
      pck.reverse_dns.class_path  = pck_expand_tilde(_PATH_REVERSEDNSCLASS);
    }
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: reverse_dns_class_path is [%s]", 
	     pck.reverse_dns.class_path);
  } else if(!strcasecmp(directive, "PCKreverseDNShost")) {
    if(path) {
      pck.reverse_dns.host_path   = pck_expand_tilde((char*)path);
    } else {
      pck.reverse_dns.host_path   = pck_expand_tilde(_PATH_REVERSEDNSHOST);
    }
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: reverse_dns_host_path is [%s]", 
	     pck.reverse_dns.host_path);
  } else if(!strcasecmp(directive, "PCKmatchDNSclass")) {
    if(path) {
      pck.matching_dns.class_path = pck_expand_tilde((char*)path);
    } else {
      pck.matching_dns.class_path = pck_expand_tilde(_PATH_MATCHDNSCLASS);
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: matching_dns_class_path is [%s]", 
	       pck.matching_dns.class_path);
    }
  } else if(!strcasecmp(directive, "PCKmatchDNShost")) {
    if(path) {
      pck.matching_dns.host_path  = pck_expand_tilde((char*)path);
    } else {
      pck.matching_dns.host_path  = pck_expand_tilde(_PATH_MATCHDNSHOST);
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: matching_dns_host_path is [%s]", 
	       pck.matching_dns.host_path);
    }
  } else if(!strcasecmp(directive, "PCKgoawayclass")) {
    if(path) {
      pck.goaway.class_path       = pck_expand_tilde((char*)path);
    } else {
      pck.goaway.class_path       = pck_expand_tilde(_PATH_GOAWAYCLASS);
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: goaway_class_path is [%s]", 
	       pck.goaway.class_path);
    }
  } else if(!strcasecmp(directive, "PCKgoawayhost")) {
    if(path) {
      pck.goaway.host_path        = pck_expand_tilde((char*)path);
    } else {
      pck.goaway.host_path        = pck_expand_tilde(_PATH_GOAWAYHOST);
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: goaway_host_path is [%s]", 
	       pck.goaway.host_path);
    }
  }
  
  return;
}

/*
 * Block-out module before entering password-check module.
 * DNS and GoAway checks are done here.
 */
void pck_block_out(char *directive, 
		   char *logging_mark,
		   char *index)
{
  config_rec *pck_tmp;
  char **save, **arg_temp, *p;
  int check_flag, yesno;
  const char *dummy = "DUMMY";
  
  /* 
   * anonymous and "PCK" is true? 
   */
  if( !pck.remote.anonname || !pck.enable )
    return;

  /*
   * If index="class" and class-name is not defined, quit here.
   */
  if( !pck.remote.class_defined && !strcasecmp(index, "class") ) {
    if( pck.enable )
      log_pri (L_ERROR, "%s Quit. Class not defined?",
	     logging_mark);
    return;
  }

  /*
   * We initialize various flags.
   */
  check_flag  =  NO_NEED_TO_CHECK;
  check_flag |=  NONE_DEFINED;
  check_flag &= ~NEED_TO_CHECK;
  check_flag &= ~CHECKED_ONCE_FILE;
  check_flag &= ~NO_MORE_CHECK;
  check_flag &= ~NONE_DEFINED;
  check_flag &= ~YESNO;
  check_flag &= ~DNS_PATH_DEFINED;
  check_flag &= ~OVERRIDE_DEFINED;

  pck_tmp = find_config( main_server->conf, 
			 CONF_PARAM,
			 directive,
			 FALSE);

  /*
   * When specified "directive" is _NOT_ found,
   * return normally without doing anything.
   * Initialization is already done by pwcheck_child_init().
   */
  if(pck_tmp) {
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: \"%s\" FOUND.", directive);
  } else {
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: \"%s\" NOT FOUND ??", directive);
    pck.index = dummy;	/* Avoid excess syslog messages */
    return;
  }
  
  pck.index = index;

  /*
   * "pck_buf0" is reduced to a remote access class name.
   */
  if ( !strcasecmp(pck.index, "class") ) {
    (void) strcpy(pck_buf0,pck.remote.class);
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: CLASS check!!");
    /*
     * "pck_buf0" is reduced to a remote host name.
     */
  } else if ( !strcasecmp(pck.index, "host" ) ) {
    (void) strcpy(pck_buf0,pck.remote.name);
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: HOST  check!!");
    /*
     * By default, "pck_buf0" is reduced to a remote acess class name.
     */
  } else {
    (void) strcpy(pck_buf0,pck.remote.class);
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: CLASS check!! <default>");
  }

  (void) pck_upper(pck_buf0);

  while(pck_tmp) {	/* while-loop-1 */

    if( pck.debug & PCK_DEBUG_NORMAL )
      log_pri (L_TRACE, "---: Check entry found:[%s]", pck_buf0);
    
    if(!pck_tmp->argv[0]) {
      yesno = TRUE;
      if( pck.debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: argv.0 = YES");
    } else {
      yesno = FALSE;
      if( pck.debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: argv.0 = NO");
    }

    if( pck.debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: No. of arguments = [%d]", pck_tmp->argc);

    /*
     * Reduce "class regex", "file path" and "override" here.
     *
     * pck_buf1 .........."class regex"   <---- argv[1]
     * pck_buf2 .........."file path"     <---- argv[2]
     * pck_buf3 .........."override"      <---- argv[3]
     *
     */

    /*
     * Arguments are "class", "path", "override"
     *
     * Example.  argv[1]     argv[2]     argv[3]
     *           !local*  /msgs/message  override
     */
    if( pck_tmp && pck_tmp->argc > 3 ) {
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: Argv.1,2,3 [%s][%s][%s]", 
	       (char *)pck_tmp->argv[1], 
	       (char *)pck_tmp->argv[2], 
	       (char *)pck_tmp->argv[3]);
      (void) strcpy(pck_buf1,pck_tmp->argv[1]);	/* "class"  */
      pck_buf2 = pck_tmp->argv[2];		/* "path"   */
      pck_buf3 = pck_tmp->argv[3];		/* "override" */
    /*
     * Arguments are "class", "path"
     *
     * Example.  argv[1]     argv[2]
     *           !local*  /msgs/message
     *           !local*  "override"
     */
    } else if( pck_tmp && pck_tmp->argc > 2 ) {
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: Argv.1,2 [%s][%s]", 
	       (char *)pck_tmp->argv[1], 
	       (char *)pck_tmp->argv[2]);
      (void) strcpy(pck_buf1,pck_tmp->argv[1]);		/* "class" */
      /*
       * If "/" mark not found in second argument,
       * it is *not* an absolute path of a named file!
       * We regard it as "override" and default path is
       * set to pck_buf2.
       */
      if(pck_istrstr(0, pck_tmp->argv[2], "/") == strlen(pck_tmp->argv[2])) {
	(void) pck_block_sub1(directive, &pck_buf2);	/* "path" */
	pck_buf3 = pck_tmp->argv[2];			/* "override" */
	/*
	 * "/" mark (a named file path) found!!
	 * We regard it as default path and "overide" feature
	 * is not activated.
	 */
      } else {
	pck_buf2 = pck_tmp->argv[2];	/* "path" */
	pck_buf3 = "not-override";	/* "override" */
      }
    /*
     * Argument is only "class"
     * Set default path to pck_buf2 and "overide" feature
     * is not activated.
     *
     * Example.  argv[1]
     *           !local*
     */
    } else if( pck_tmp && pck_tmp->argc > 1 ) {
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: Argv.1 [%s]", 
	       (char *)pck_tmp->argv[1]);
      (void) strcpy(pck_buf1,pck_tmp->argv[1]);		/* "class" */
      (void) pck_block_sub1(directive, &pck_buf2);	/* "path" */
      pck_buf3 = "not-override";			/* "override" */
    } else {
      (void) strcpy(pck_buf1,"*");			/* "class" */
      (void) pck_block_sub1(directive, &pck_buf2);	/* "path" */
      pck_buf3 = "not-override";			/* "override" */
    }

    if( pck.debug & PCK_DEBUG_NORMAL )
      log_pri (L_TRACE, (yesno)? 
	     "---: Entry - Class:Path:OVR[yes][%s][%s][%s]" :
	     "---: Entry - Class:Path:OVR [no][%s][%s][%s]",
	     pck_buf1, pck_buf2, pck_buf3);

    if(pck_buf1) {	/* if-loop-1 */

      if(pck_tmp->argv[1]) {
	if( pck.debug & PCK_DEBUG_HIGH )
	  log_pri (L_TRACE, "---: argv.1 = [%s]", 
		 (char *)pck_tmp->argv[1]);
	if(pck_tmp->argv[2]) {
	  if( pck.debug & PCK_DEBUG_HIGH )
	    log_pri (L_TRACE, "---: argv.2 = [%s]", 
		   (char *)pck_tmp->argv[2]);
	  if(pck_tmp->argv[3]) {
	    if( pck.debug & PCK_DEBUG_HIGH )
	      log_pri (L_TRACE, "---: argv.3 = [%s]", 
		     (char *)pck_tmp->argv[3]);
	  }
	}
      }
      
      (void) strcpy(pck_temp,pck_buf1);
      for ( save = arg_temp = pck_comma_split(pck_temp); 
	    (p = *arg_temp++) != NULL; ) {
	
	/*
	 * Normally, "p" denotes access class in a regular
	 * expression form.  If "p" denotes IP address in a
	 * regular expression form, pck_buf0 should not be used
	 * to compare with "p".  If this is the case, we use
	 * "pck.remote.ip" instead.
	 */
	(void) pck_upper(p);

	if( pck.debug & PCK_DEBUG_HIGH )
	  log_pri (L_TRACE, "---: argv.1.target_class = [%s]", p);
	
	if ( *p == '!' ) { /* !!!! */
	  while (*p == '!') p++;
	  if ( !pck_fnmatch(p, 
			    IsDigitAd(p, pck.index)?
			    (char*)pck.remote.ip : pck_buf0,
			    FNM_PERIOD2) ) {
	    log_pri (L_INFO, (yesno)?
		   "%s match [!%s] <%s> NO NEED TO CHECK (flag: yes)" :
		   "%s match [!%s] <%s> NEED TO CHECK (flag: no)",
		   logging_mark, 
		   p, 
		   IsDigitAd(p, pck.index)?
		   (char*)pck.remote.ip : pck_buf0);
	    if( !(check_flag & CHECKED_ONCE_FILE) ) {
	      check_flag |=  CHECKED_ONCE_FILE;
	      if(yesno) {
		check_flag &= ~NEED_TO_CHECK;
		check_flag |=  YESNO;
	      } else {
		check_flag |=  NEED_TO_CHECK;
		check_flag &= ~YESNO;
	      }
	      check_flag |= DNS_PATH_DEFINED;
	      (void) pck_path_define(directive, pck_buf2);
	      check_flag |= OVERRIDE_DEFINED;
	      (void) pck_override_define(directive, pck_buf3);
	    }
	  } else {
	    log_pri (L_INFO, (yesno)? 
		   "%s no match [!%s] <%s> NEED TO CHECK (flag: yes)" :
		   "%s no match [!%s] <%s> NO NEED TO CHECK (flag: no)",
		   logging_mark, 
		   p, 
		   IsDigitAd(p, pck.index)?
		   (char*)pck.remote.ip : pck_buf0);
	    if( !(check_flag & CHECKED_ONCE_FILE) ) {
	      check_flag |=  CHECKED_ONCE_FILE;
	      if(yesno) {
		check_flag |=  NEED_TO_CHECK;
		check_flag |=  YESNO;
	      } else {
		check_flag &= ~NEED_TO_CHECK;
		check_flag &= ~YESNO;
	      }
	      check_flag |= DNS_PATH_DEFINED;
	      (void) pck_path_define(directive, pck_buf2);
	      check_flag |= OVERRIDE_DEFINED;
	      (void) pck_override_define(directive, pck_buf3);
	    }
	  }
	} else if ( *p == '$' ) { /* $$$$ */
	  while (*p == '$') p++;
	  if ( !pck_fnmatch(p, 
			    IsDigitAd(p, pck.index)?
			    (char*)pck.remote.ip : pck_buf0,
			    FNM_PERIOD2) ) {
	    log_pri (L_INFO, (yesno)?
		   "%s match [$%s] <%s> NEED TO CHECK (flag: yes)" :
		   "%s match [$%s] <%s> NO NEED TO CHECK (flag: no)",
		   logging_mark, 
		   p, 
		   IsDigitAd(p, pck.index)?
		   (char*)pck.remote.ip : pck_buf0);
	    if( !(check_flag & NO_MORE_CHECK) ) {
	      check_flag |=  CHECKED_ONCE_FILE;
	      check_flag |=  NO_MORE_CHECK;
	      if(yesno) {
		check_flag |=  NEED_TO_CHECK;
		check_flag |=  YESNO;
	      } else {
		check_flag &= ~NEED_TO_CHECK;
		check_flag &= ~YESNO;
	      }
	      check_flag |= DNS_PATH_DEFINED;
	      (void) pck_path_define(directive, pck_buf2);
	      check_flag |= OVERRIDE_DEFINED;
	      (void) pck_override_define(directive, pck_buf3);
	    }
	  }
	} else { /* |||| */
	  if ( !pck_fnmatch(p, 
			    IsDigitAd(p, pck.index)?
			    (char*)pck.remote.ip : pck_buf0,
			    FNM_PERIOD2) ) {
	    log_pri (L_INFO, (yesno)?
		   "%s match [%s] <%s> (flag: yes)":
		   "%s match [%s] <%s> (flag: no)",
		   logging_mark, 
		   p, 
		   IsDigitAd(p, pck.index)?
		   (char*)pck.remote.ip : pck_buf0);
	    if( !(check_flag & CHECKED_ONCE_FILE) ) {
	      check_flag |=  CHECKED_ONCE_FILE;
	      if(yesno) {
		check_flag |=  NEED_TO_CHECK;
		check_flag |=  YESNO;
	      } else {
		check_flag &= ~NEED_TO_CHECK;
		check_flag &= ~YESNO;
	      }
	      check_flag |= DNS_PATH_DEFINED;
	      (void) pck_path_define(directive, pck_buf2);
	      check_flag |= OVERRIDE_DEFINED;
	      (void) pck_override_define(directive, pck_buf3);
	    }
	  }
	}
      }
    }	/* endof if-loop-1 */
    
    pck_tmp = find_config_next( pck_tmp,
				pck_tmp->next,
				CONF_PARAM,
				directive,
				FALSE);
  }	/* edd of while-loop-1 */
  
  /*
   * User's class matches `$' expression.
   */
  if( (check_flag & NO_MORE_CHECK) ) {
    log_pri (L_INFO, (check_flag & YESNO)?
	   "%s target [%s] match (`$') NEED TO CHECK" :
	   "%s target [%s] match (`$') NO NEED TO CHECK",
	   logging_mark, 
	   pck_buf0);
    if(check_flag & YESNO) 
      check_flag =  NEED_TO_CHECK;
    else
      check_flag =  NO_NEED_TO_CHECK;
    (void)  pck_block_sub3(directive, check_flag, "$ERRMSG");
    return;
  }

  if( (check_flag & NONE_DEFINED) ) {
    /*
     * When `PCKreverseDNS' does not exist in proftpd.conf,
     * we leave here and skip DNS checks.
     */
    log_pri (L_INFO, "%s target [%s] not specified in proftpd.conf",
	   logging_mark, 
	   pck_buf0);
    check_flag  = NO_NEED_TO_CHECK;
    (void) pck_block_sub2(directive, check_flag);
    return;
  } else {
    /*
     * User's class matches some regular expression.
     * Access permission is stored in check_flag here.
     * Other flags should be stripped off before return.
     */
    if( (check_flag & CHECKED_ONCE_FILE) ) {
      log_pri (L_INFO, (check_flag & YESNO)?
	     "%s target [%s] match (flag: yes)":
	     "%s target [%s] match (flag: no)",
	     logging_mark, 
	     pck_buf0);
      check_flag &= ~NONE_DEFINED;
      check_flag &= ~NO_MORE_CHECK;
      check_flag &= ~CHECKED_ONCE_FILE;
      check_flag &= ~YESNO;
      check_flag &= ~DNS_PATH_DEFINED;
      check_flag &= ~OVERRIDE_DEFINED;
      if ( pck.debug & PCK_DEBUG_FULL )
	log_pri (L_TRACE, "---: return flag? [%d]",check_flag);

      (void)  pck_block_sub3(directive, check_flag, "ERRMSG");
      return;
    } else {
      /*
       * Although class regex is specified in `PCKreverseDNS'
       * of proftpd.conf, we cannot find user's access class there.
       * In this case we leave here and skip DNS checks.
       */
      log_pri (L_INFO, "%s target [%s] does not match anything",
	     logging_mark, 
	     pck_buf0);
      check_flag = NO_NEED_TO_CHECK;
      (void) pck_block_sub2(directive, check_flag);
      return;
    }
  }
  
  log_pri (L_ERROR, "ERR: DNS something wrong?  NO DNS checks done.");
  log_pri (L_ERROR, "ERR: Check your proftpd.conf...");
  check_flag = NO_NEED_TO_CHECK;
  (void) pck_block_sub2(directive, check_flag);
  return;
  
}

void pck_block_sub1(char *directive, const char **pck_buf2)
{
  if(!strcasecmp(directive, "PCKserverdeadclass")) {
    *pck_buf2 = _PATH_SERVERDEADCLASS;
  } else if(	!strcasecmp(directive, "PCKserverdeadhost")) {
    *pck_buf2 = _PATH_SERVERDEADHOST;
  } else if(!strcasecmp(directive, "PCKreverseDNSclass")) {
    *pck_buf2 = _PATH_REVERSEDNSCLASS;
  } else if(!strcasecmp(directive, "PCKreverseDNShost")) {
    *pck_buf2 = _PATH_REVERSEDNSHOST;
  } else if(!strcasecmp(directive, "PCKmatchDNSclass")) {
    *pck_buf2 = _PATH_MATCHDNSCLASS;
  } else if(!strcasecmp(directive, "PCKmatchDNShost")) {
    *pck_buf2 = _PATH_MATCHDNSHOST;
  } else if(!strcasecmp(directive, "PCKgoawayclass")) {
    *pck_buf2 = _PATH_GOAWAYCLASS;
  } else if(!strcasecmp(directive, "PCKgoawayhost")) {
    *pck_buf2 = _PATH_GOAWAYHOST;
  }
  return;
}

void pck_block_sub2(char *directive, int check_flag)
{
  if(!strcasecmp(directive, "PCKserverdeadclass")) {
    pck.server_dead.class  = check_flag;
  } else if(!strcasecmp(directive, "PCKserverdeadhost")) {
    pck.server_dead.host   = check_flag;
  } else if(!strcasecmp(directive, "PCKreverseDNSclass")) {
    pck.reverse_dns.class  = check_flag;
  } else if(!strcasecmp(directive, "PCKreverseDNShost")) {
    pck.reverse_dns.host   = check_flag;
  } else if(!strcasecmp(directive, "PCKmatchDNSclass")) {
    pck.matching_dns.class = check_flag;
  } else if(!strcasecmp(directive, "PCKmatchDNShost")) {
    pck.matching_dns.host  = check_flag;
  } else if(!strcasecmp(directive, "PCKgoawayclass")) {
    pck.goaway.class       = check_flag;
  } else if(!strcasecmp(directive, "PCKgoawayhost")) {
    pck.goaway.host = check_flag;
  }
  return;
}

void pck_block_sub3(char *directive, int check_flag, char *logging_mark)
{
  if(!strcasecmp(directive, "PCKserverdeadclass")) {
    pck.server_dead.class  = check_flag;
    if( pck.debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: %s path (serverdead_class) [%s]",
	     logging_mark, pck.server_dead.class_path);
  } else if(!strcasecmp(directive, "PCKserverdeadhost")) {
    pck.server_dead.host   = check_flag;
    if( pck.debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: %s path (serverdead_host) [%s]",
	     logging_mark, pck.server_dead.host_path);
  } else if(!strcasecmp(directive, "PCKreverseDNSclass")) {
    pck.reverse_dns.class  = check_flag;
    if( pck.debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: %s path (reverse_dns_class) [%s]",
	     logging_mark, pck.reverse_dns.class_path);
  } else if(!strcasecmp(directive, "PCKreverseDNShost")) {
    pck.reverse_dns.host   = check_flag;
    if( pck.debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: %s path (reverse_dns_host) [%s]",
	     logging_mark, pck.reverse_dns.host_path);
  } else if(!strcasecmp(directive, "PCKmatchDNSclass")) {
    pck.matching_dns.class = check_flag;
    if( pck.debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: %s path (matching_dns_class) [%s]",
	     logging_mark, pck.matching_dns.class_path);
  } else if(!strcasecmp(directive, "PCKmatchDNShost")) {
    pck.matching_dns.host  = check_flag;
    if( pck.debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: %s path (matching_dns_host) [%s]",
	     logging_mark, pck.matching_dns.host_path);
  } else if(!strcasecmp(directive, "PCKgoawayclass")) {
    pck.goaway.class       = check_flag;
    if( pck.debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: %s path (goaway_class) [%s]",
	     logging_mark, pck.goaway.class_path);
  } else if(!strcasecmp(directive, "PCKgoawayhost")) {
    pck.goaway.host        = check_flag;
    if( pck.debug & PCK_DEBUG_FULL )
      log_pri (L_TRACE, "---: %s path (goaway_host) [%s]",
	     logging_mark, pck.goaway.host_path);
  }
  return;
}

void pck_override_define(char *directive, const char *pck_argv)
{

  if(!strcasecmp(directive,  "PCKreverseDNShost")) {
    if(!strcasecmp(pck_argv, "override")) {
      pck.reverse_dns.host_override    = TRUE;
    } else {
      pck.reverse_dns.host_override    = FALSE;
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: 4th argument is not \"override\" [%s]",
	       pck_argv);
    }
  } else if(!strcasecmp(directive, "PCKreverseDNSclass")) {
    if(!strcasecmp(pck_argv,       "override") ) {
      pck.reverse_dns.class_override   = TRUE;
    } else {
      pck.reverse_dns.class_override   = FALSE;
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: 4th argument is not \"override\" [%s]",
	       pck_argv);
    }
  } else if(!strcasecmp(directive, "PCKmatchDNShost")) {
    if(!strcasecmp(pck_argv,       "override")) {
      pck.matching_dns.host_override   = TRUE;
    } else {
      pck.matching_dns.host_override   = FALSE;
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: 4th argument is not \"override\" [%s]",
	       pck_argv);
    }
  } else if(!strcasecmp(directive, "PCKmatchDNSclass")) {
    if(!strcasecmp(pck_argv,       "override") ) {
      pck.matching_dns.class_override  = TRUE;
    } else {
      pck.matching_dns.class_override  = FALSE;
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: 4th argument is not \"override\" [%s]",
	       pck_argv);
    }
  }
  
  return;
}

#if defined(PW_CHECK_DNS)	/* compile only for DNS check */
int pck_has_reverse_dns ()
{
  char *fqdn;
  char *ip;
  int has_reverse_dns_class = TRUE;	/* By default, DNS is valid! */
  int has_reverse_dns_host  = TRUE;	/* By default, DNS is valid! */

  if( !pck.remote.anonname || !pck.enable )	/* anonymous and "PCK" is true? */
    return TRUE;

  ip = (char *)pck.remote.ip;

  /*
   * Reverse DNS check
   */
  if(!strcasecmp(pck.index, "class"))
    log_pri (L_INFO, (pck.reverse_dns.class)?
	   "DNS: Reverse__DNS_Class check? [yes] (%s)" :
	   "DNS: Reverse__DNS_Class check? [no]  (%s)", 
	   pck.remote.class);
  
  if(!strcasecmp(pck.index, "host"))
  log_pri (L_INFO, (pck.reverse_dns.host)?
	 "DNS: Reverse__DNS_Host  check? [yes] (%s)" :
	 "DNS: Reverse__DNS_Host  check? [no]  (%s)", 
	 pck.remote.name);
  
  if( pck.reverse_dns.class || pck.reverse_dns.host ) {

    /* the old code checks remote_address for 0.0.0.0 and if there's a   */
    /* match, sets the hostname to 'localhost' - so let's do that for    */
    /* compatibility and perhaps unknown resolvers which behave this way */

#if defined(PCK_AY_DEBUG)	/* (Debugging purpose:@@) */
    ip = "255.255.255.255";
#endif

    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: IP address = [%s]",ip);

    if ( !strcmp(ip, "0.0.0.0") || !strcmp(ip, "255.255.255.255") ) {
      fqdn = "localhost";
#if 0
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: 0.0.0.0 LH:RH:IP:FQDN [%s]:[%s]:[%s]:[%s]",
	       pck.local.name, pck.remote.name, ip, fqdn);
      (void) pck_lookup_name((char *)pck.local.name,&ip);
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: 0.0.0.0 LH:RH:IP:FQDN [%s]:[%s]:[%s]:[%s]",
	       pck.local.name, pck.remote.name, ip, fqdn);
      if( pck.reverse_dns.class )
	has_reverse_dns_class  = pck_lookup_ip(ip,&fqdn);
      if( pck.reverse_dns.host )
	has_reverse_dns_host   = pck_lookup_ip(ip,&fqdn);
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: 0.0.0.0 LH:RH:IP:FQDN [%s]:[%s]:[%s]:[%s]",
	       pck.local.name, pck.remote.name, ip, fqdn);
#else
      if( pck.reverse_dns.class )
	has_reverse_dns_class  = TRUE;
      if( pck.reverse_dns.host )
	has_reverse_dns_host   = TRUE;
#endif
    } else {
      if( pck.reverse_dns.class )
	has_reverse_dns_class  = pck_lookup_ip(ip,&fqdn);
      if( pck.reverse_dns.host )
	has_reverse_dns_host   = pck_lookup_ip(ip,&fqdn);
    }

    if( pck.debug & PCK_DEBUG_NORMAL ) {
      if(has_reverse_dns_class || has_reverse_dns_host) {
	log_pri (L_TRACE, "---: Has Reverse DNS Remote_FQDN:IP [%s]:[%s]",
	       fqdn, ip);
	log_pri (L_TRACE, "---: Has Reverse DNS Remote:Local   [%s]:[%s]",
	       pck.remote.name, pck.local.name);
      } else {
	log_pri (L_TRACE, "---: No Reverse DNS? Remote_FQDN:IP [%s]:[%s]",
	       fqdn, ip);
	log_pri (L_TRACE, "---: No Reverse DNS? Remote:Local   [%s]:[%s]",
	       pck.remote.name, pck.local.name);
      }
    }
    
  }

  if( pck.debug & PCK_DEBUG_NORMAL ) {
    if(!strcasecmp(pck.index, "class"))
      log_pri (L_TRACE, (has_reverse_dns_class)?
	     "---: Has Reverse_Dns_Class? [yes]" :
	     "---: Has Reverse_Dns_Class? [no]" );
    if(!strcasecmp(pck.index, "host"))
      log_pri (L_TRACE, (has_reverse_dns_host)?
	     "---: Has Reverse_Dns_Host?  [yes]" :
	     "---: Has Reverse_Dns_Host?  [no]" );
  }

#if defined(PCK_AY_DEBUG)	/* Force FALSE (Debugging purpose:@@) */
  has_reverse_dns_class = FALSE;
  has_reverse_dns_host  = FALSE;
#endif

  if( pck.reverse_dns.class )
    return has_reverse_dns_class;

  if( pck.reverse_dns.host )
    return has_reverse_dns_host;

  return TRUE;
}

int pck_has_matching_dns ()
{
  char *fqdn;
  char *ip;
  int has_matching_dns_class = TRUE;	/* By default, DNS is valid! */
  int has_matching_dns_host  = TRUE;	/* By default, DNS is valid! */

  if( !pck.remote.anonname || !pck.enable )	/* anonymous and "PCK" is true? */
    return TRUE;

  ip = (char *)pck.remote.ip;

  /*
   * DNS matching check
   */
  if(!strcasecmp(pck.index, "class"))
    log_pri (L_INFO, (pck.matching_dns.class)?
	   "DNS: Matching_DNS_Class check? [yes] (%s)" :
	   "DNS: Matching_DNS_Class check? [no]  (%s)", 
	   pck.remote.class);
  
  if(!strcasecmp(pck.index, "host"))
    log_pri (L_INFO, (pck.matching_dns.host)?
	   "DNS: Matching_DNS_Host  check? [yes] (%s)" :
	   "DNS: Matching_DNS_Host  check? [no]  (%s)", 
	   pck.remote.name);

  if( pck.matching_dns.class || pck.matching_dns.host ) {

    /* the old code checks remote_address for 0.0.0.0 and if there's a   */
    /* match, sets the hostname to 'localhost' - so let's do that for    */
    /* compatibility and perhaps unknown resolvers which behave this way */

#if defined(PCK_AY_DEBUG)	/* (Debugging purpose:@@) */
    ip = "255.255.255.255";
#endif

    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: IP address = [%s]",ip);

    if ( !strcmp(ip, "0.0.0.0") || !strcmp(ip, "255.255.255.255") ) {
      fqdn = "localhost";
#if 0
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: 0.0.0.0 LH:RH:IP:FQDN [%s]:[%s]:[%s]:[%s]",
	       pck.local.name, pck.remote.name, ip, fqdn);
      (void) pck_lookup_name((char *)pck.local.name,&ip);
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: 0.0.0.0 LH:RH:IP:FQDN [%s]:[%s]:[%s]:[%s]",
	       pck.local.name, pck.remote.name, ip, fqdn);
      (void) pck_lookup_ip(ip,&fqdn);
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: 0.0.0.0 LH:RH:IP:FQDN [%s]:[%s]:[%s]:[%s]",
	       pck.local.name, pck.remote.name, ip, fqdn);
      if( pck.matching_dns.class )
	has_matching_dns_class = pck_check_name_for_ip(fqdn,ip);
      if( pck.matching_dns.host )
	has_matching_dns_host  = pck_check_name_for_ip(fqdn,ip);
      if( pck.debug & PCK_DEBUG_HIGH )
	log_pri (L_TRACE, "---: 0.0.0.0 LH:RH:IP:FQDN [%s]:[%s]:[%s]:[%s]",
	       pck.local.name, pck.remote.name, ip, fqdn);
#else
      if( pck.matching_dns.class )
	has_matching_dns_class = TRUE;
      if( pck.matching_dns.host )
	has_matching_dns_host  = TRUE;
#endif
    } else {
      (void) pck_lookup_ip(ip,&fqdn);
      if( fqdn ) {
	if( pck.matching_dns.class )
	  has_matching_dns_class = pck_check_name_for_ip(fqdn,ip);
	if( pck.matching_dns.host )
	  has_matching_dns_host  = pck_check_name_for_ip(fqdn,ip);
      } else {
	if( pck.matching_dns.class )
	  has_matching_dns_class = FALSE;
	if( pck.matching_dns.host )
	  has_matching_dns_host  = FALSE;
      }
    }
    
    if( pck.debug & PCK_DEBUG_NORMAL ) {
      if(has_matching_dns_class || has_matching_dns_host ) {
	log_pri (L_TRACE, "---: Has Reverse DNS Remote_FQDN:IP [%s]:[%s]",
	       fqdn, ip);
	log_pri (L_TRACE, "---: Has Reverse DNS Remote:Local   [%s]:[%s]",
	       pck.remote.name, pck.local.name);
      } else {
	log_pri (L_TRACE, "---: No Reverse DNS? Remote_FQDN:IP [%s]:[%s]",
	       fqdn, ip);
	log_pri (L_TRACE, "---: No Reverse DNS? Remote:Local   [%s]:[%s]",
	       pck.remote.name, pck.local.name);
      }
    }

  }

  if( pck.debug & PCK_DEBUG_NORMAL ) {
    if(!strcasecmp(pck.index, "class"))
      log_pri (L_TRACE, (has_matching_dns_class)?
	     "---: Has Matching_DNS_Class? [yes]" :
	     "---: Has Matching_DNS_Class? [no]" );
    if(!strcasecmp(pck.index, "host"))
      log_pri (L_TRACE, (has_matching_dns_host)?
	     "---: Has Matching_DNS_Host?  [yes]" :
	     "---: Has Matching_DNS_Host?  [no]" );
  }
  
#if defined(PCK_AY_DEBUG)	/* Force FALSE (Debugging purpose:@@) */
  has_matching_dns_class = FALSE;
  has_matching_dns_host  = FALSE;
#endif

  if( pck.matching_dns.class )
    return has_matching_dns_class;

  if( pck.matching_dns.host )
    return has_matching_dns_host;

  return TRUE;

}
#endif	/* compile only for DNS check */

int pck_server_dead_class_kickout()
{
  int kick_out = FALSE;		/* By default, do not kick out! */
  
#if defined(PCK_AY_DEBUG)	/* Force TRUE (Debugging purpose:@@) */
  pck.server_dead.class  = TRUE; 
#endif

  /*
   * Kick out if user's class matches server_dead_class.
   */
  if( pck.remote.anonname    &&	/* anonymous? */
      pck.enable             &&	/* "PCK" is true? */
      pck.server_dead.class )	/* Server Dead (Class)? */
    {
      if( pck_core_display_file("530", 
				(char *)pck.server_dead.class_path) ) {
	log_pri (L_ERROR, "ERR: Cannot Open SERVERDEADCLASS file %s: %s",
	       pck.server_dead.class_path, strerror(errno));
	send_response("530",
		      "FTP SERVER SHUTDOWN (by \"class\"). "
		      "Please try again later...");
      }
      log_pri (L_INFO, "PCK: Kick out (Server_Dead_Class check) <%s>",
	     pck.remote.class);
      kick_out = TRUE;
      return kick_out;
    }
  return kick_out;
}

int pck_server_dead_host_kickout()
{
  int kick_out = FALSE;		/* By default, do not kick out! */
  
#if defined(PCK_AY_DEBUG)	/* Force TRUE (Debugging purpose:@@) */
  pck.server_dead.host  = TRUE; 
#endif

  /*
   * Kick out if user's host matches server_dead_host.
   */
  if( pck.remote.anonname    &&	/* anonymous? */
      pck.enable             &&	/* "PCK" is true? */
      pck.server_dead.host )	/* Server Dead (Host)? */
    {
      if( pck_core_display_file("530", 
				(char *)pck.server_dead.host_path) ) {
	log_pri (L_ERROR, "ERR: Cannot Open SERVERDEADHOST file %s: %s",
	       pck.server_dead.host_path, strerror(errno));
	send_response("530",
		      "FTP SERVER SHUTDOWN (by \"host\"). "
		      "Please try again later...");
      }
      log_pri (L_INFO, "PCK: Kick out (Server_Dead_Host check) <%s>",
	     pck.remote.name);
      kick_out = TRUE;
      return kick_out;
    }
  return kick_out;
}

int pck_goaway_class_kickout()
{
  int kick_out = FALSE;		/* By default, do not kick out! */
  
#if defined(PCK_AY_DEBUG)	/* Force TRUE (Debugging purpose:@@) */
  pck.goaway.class  = TRUE; 
#endif

  /*
   * Kick out if user's class matches goaway_class.
   */
  if( pck.remote.anonname    &&	/* anonymous? */
      pck.enable             &&	/* "PCK" is true? */
      pck.goaway.class )	/* Go Away? */
    {
      if( pck_core_display_file("530", 
				(char *)pck.goaway.class_path) ) {
	log_pri (L_ERROR, "ERR: Cannot Open GOAWAYCLASS file %s: %s",
	       pck.goaway.class_path, strerror(errno));
	send_response("530",
		      "Server denies anonymous login. "
		      "(Go_Away_Class Match)");
      }
      log_pri (L_INFO, "PCK: Kick out (Go_Away_Class check) <%s>",
	     pck.remote.class);
      kick_out = TRUE;
      return kick_out;
    }
  return kick_out;
}

int pck_goaway_host_kickout()
{
  int kick_out = FALSE;		/* By default, do not kick out! */
  
#if defined(PCK_AY_DEBUG)	/* Force TRUE (Debugging purpose:@@) */
  pck.goaway.host  = TRUE; 
#endif

  /*
   * Kick out if user's host matches goaway_host.
   */
  if( pck.remote.anonname    &&	/* anonymous? */
      pck.enable             &&	/* "PCK" is true? */
      pck.goaway.host )		/* Go Away? */
    {
      if( pck_core_display_file("530", 
				(char *)pck.goaway.host_path) ) {
	log_pri (L_ERROR, "ERR: Cannot Open GOAWAYHOST file %s: %s",
	       pck.goaway.host_path, strerror(errno));
	send_response("530",
		      "Server denies anonymous login. "
		      "(Go_Away_Host Match)");
      }
      log_pri (L_INFO, "PCK: Kick out (Go_Away_Host check) <%s>",
	     pck.remote.name);
      kick_out = TRUE;
      return kick_out;
    }
  return kick_out;
}

int pck_has_reverse_dns_class_kickout()
{
  int kick_out = FALSE;		/* By default, do not kick out! */
  char *blank="";
  
  /*
   * Kick out if NO reverse DNS entry found.
   */
  if( pck.remote.anonname    &&	/* anonymous? */
      pck.enable             &&	/* "PCK" is true? */
      pck.reverse_dns.class )	/* Do reverse_DNS check? */
    {
      /*
       * DO NOT kick out (override=TRUE!!)
       */
      if(pck.reverse_dns.class_override) {
	if( pck_override_display_file("230",
				      (char *)pck.reverse_dns.class_path) ) {
	  log_pri (L_ERROR, "ERR: Cannot Open REVERSEDNSCLASS file %s: %s",
		 pck.reverse_dns.class_path, strerror(errno));
	  add_response("230",blank);
	  add_response("230","##############################################");
	  add_response("230","PERMISSION GRANTED...");
	  add_response("230","YOUR HOST, HOWEVER, HAS DNS PROBLEMS !!");
	  add_response("230","FIX IT OR ASK YOUR SYSTEM ADMINISTRATOR.");
	  add_response("230","                (REVERSE DNS ENTRY NOT FOUND?)");
	  add_response("230","##############################################");
	  add_response("230",blank);
	}
	log_pri (L_INFO, "OVR: Warning (Reverse_DNS_Class check) <%s>",
	       pck.remote.class);
	kick_out = FALSE;
      /*
       * kick out (override=FALSE!!)
       */
      } else {
	if( pck_core_display_file("530",
				  (char *)pck.reverse_dns.class_path) ) {
	  log_pri (L_ERROR, "ERR: Cannot Open REVERSEDNSCLASS file %s: %s",
		 pck.reverse_dns.class_path, strerror(errno));
	  send_response("530",
			"Server denies anonymous login. "
			"Reverse_DNS_Class check failed...");
	}
	log_pri (L_INFO, "PCK: Kick out (Reverse_DNS_Class check) <%s>",
	       pck.remote.class);
	kick_out = TRUE;
      }
    }
  return kick_out;
}

int pck_has_reverse_dns_host_kickout()
{
  int kick_out = FALSE;		/* By default, do not kick out! */
  char *blank="";

  /*
   * Kick out if NO reverse DNS entry found.
   */
  if( pck.remote.anonname    &&	/* anonymous? */
      pck.enable             &&	/* "PCK" is true? */
      pck.reverse_dns.host )	/* Do reverse_DNS check? */
    {
      /*
       * DO NOT kick out (override=TRUE!!)
       */
      if(pck.reverse_dns.host_override) {
	if( pck_override_display_file("230",
				      (char *)pck.reverse_dns.host_path) ) {
	  log_pri (L_ERROR, "ERR: Cannot Open REVERSEDNSHOST file %s: %s",
		 pck.reverse_dns.host_path, strerror(errno));
	  add_response("230",blank);
	  add_response("230","##############################################");
	  add_response("230","PERMISSION GRANTED...");
	  add_response("230","YOUR HOST, HOWEVER, HAS DNS PROBLEMS !!");
	  add_response("230","FIX IT OR ASK YOUR SYSTEM ADMINISTRATOR.");
	  add_response("230","                (REVERSE DNS ENTRY NOT FOUND?)");
	  add_response("230","##############################################");
	  add_response("230",blank);
	}
	log_pri (L_INFO, "OVR: Warning (Reverse_DNS_Host check) <%s>",
	       pck.remote.name);
	kick_out = FALSE;
      /*
       * kick out (override=FALSE!!)
       */
      } else {
	if( pck_core_display_file("530",
				  (char *)pck.reverse_dns.host_path) ) {
	  log_pri (L_ERROR, "ERR: Cannot Open REVERSEDNSHOST file %s: %s",
		 pck.reverse_dns.host_path, strerror(errno));
	  send_response("530",
			"Server denies anonymous login. "
			"Reverse_DNS_Host check failed...");
	}
	log_pri (L_INFO, "PCK: Kick out (Reverse_DNS_Host check) <%s>",
	       pck.remote.name);
	kick_out = TRUE;
      }
    }
  return kick_out;
}

int pck_has_matching_dns_class_kickout()
{
  int kick_out = FALSE;		/* By default, do not kick out! */
  char *blank="";
  
  /*
   * Kick out if forward and reverse lookups do not match.
   */
  if( pck.remote.anonname    &&	/* anonymous? */
      pck.enable             &&	/* "PCK" is true? */
      pck.matching_dns.class )	/* Do matching_DNS check? */
    {
      /*
       * DO NOT kick out (override=TRUE!!)
       */
      if(pck.matching_dns.class_override) {
	if( pck_override_display_file("230",
				      (char *)pck.matching_dns.class_path) ) {
	  log_pri (L_ERROR, "ERR: Cannot Open MATCHDNSCLASS file %s: %s",
		 pck.matching_dns.class_path, strerror(errno));
	  add_response("230",blank);
	  add_response("230","##############################################");
	  add_response("230","PERMISSION GRANTED...");
	  add_response("230","YOUR HOST, HOWEVER, HAS DNS PROBLEMS !!");
	  add_response("230","FIX IT OR ASK YOUR SYSTEM ADMINISTRATOR.");
	  add_response("230","       (FORWARD AND REVERSE LOOKUP NOT MATCH?)");
	  add_response("230","##############################################");
	  add_response("230",blank);
	}
	log_pri (L_INFO, "OVR: Warning (Matching_DNS_Class check) <%s>",
	       pck.remote.class);
	kick_out = FALSE;
      /*
       * kick out (override=FALSE!!)
       */
      } else {
	if( pck_core_display_file("530",
				  (char *)pck.matching_dns.class_path) ) {
	  log_pri (L_ERROR, "ERR: Cannot Open MATCHDNSCLASS file %s: %s",
		 pck.matching_dns.class_path, strerror(errno));
	  send_response("530",
			"Server denies anonymous login. "
			"Matching_DNS_Host check failed...");
	}
	log_pri (L_INFO, "PCK: Kick out (Matching_DNS_Class check) <%s>",
	       pck.remote.class);
	kick_out = TRUE;
      }
    }
  return kick_out;
}

int pck_has_matching_dns_host_kickout()
{
  int kick_out = FALSE;		/* By default, do not kick out! */
  char *blank="";
  
  /*
   * Kick out if forward and reverse lookups do not match.
   */
  if( pck.remote.anonname    &&	/* anonymous? */
      pck.enable             &&	/* "PCK" is true? */
      pck.matching_dns.host )	/* Do matching_DNS check? */
    {
      /*
       * DO NOT kick out (override=TRUE!!)
       */
      if(pck.matching_dns.host_override) {
	if( pck_override_display_file("230",
				      (char *)pck.matching_dns.host_path) ) {
	  log_pri (L_ERROR, "ERR: Cannot Open MATCHDNSHOST file %s: %s",
		 pck.matching_dns.host_path, strerror(errno));
	  add_response("230",blank);
	  add_response("230","##############################################");
	  add_response("230","PERMISSION GRANTED...");
	  add_response("230","YOUR HOST, HOWEVER, HAS DNS PROBLEMS !!");
	  add_response("230","FIX IT OR ASK YOUR SYSTEM ADMINISTRATOR.");
	  add_response("230","       (FORWARD AND REVERSE LOOKUP NOT MATCH?)");
	  add_response("230","##############################################");
	  add_response("230",blank);
	}
	log_pri (L_INFO, "OVR: Warning (Matching_DNS_Host check) <%s>",
	       pck.remote.name);
	kick_out = FALSE;
      /*
       * kick out (override=FALSE!!)
       */
      } else {
	if( pck_core_display_file("530",
				  (char *)pck.matching_dns.host_path) ) {
	  log_pri (L_ERROR, "ERR: Cannot Open MATCHDNSHOST file %s: %s",
		 pck.matching_dns.host_path, strerror(errno));
	  send_response("530",
			"Server denies anonymous login. "
			"Matching_DNS_Host check failed...");
	}
	log_pri (L_INFO, "PCK: Kick out (Matchin_DNS_Host check) <%s>",
	       pck.remote.name);
	kick_out = TRUE;
      }
    }
  return kick_out;
}

void pck_entrance()
{
  config_rec *pck_tmp;
  
  /*
   * do `password-check' if "PCK" directive is activated.
   */
  if ( pck.debug & PCK_DEBUG_NORMAL ) {
    log_pri (L_TRACE, "---: password-check start.  valid=[%d]",pck.valid);
  }
  
  /*
   * `pck.local.name_server' is defined here.
   * If `PCKnameserver' directive is not found in `conf',
   * `hostname' is set as `pck.local.name_server'.
   */
  if(get_param_int(main_server->conf,"PCKnameserver",FALSE) != -1) {
    pck_tmp = find_config(main_server->conf, 
			  CONF_PARAM, 
			  "PCKnameserver", 
			  TRUE);
    if(pck_tmp->argv[0]) {
      pck.local.name_server = pck_tmp->argv[0];
    }
  }
  
  if( !(pck.local.name_server && *pck.local.name_server )) {
    pck.local.name_server = pck.local.name;
  }
  
  if( pck.debug & PCK_DEBUG_HIGH ) {
    log_pri (L_TRACE, "---: PCKnameserver is [%s]", pck.local.name_server);
    log_pri (L_TRACE, "---: hostname      is [%s]", pck.local.name);
  }

  /*
   * `pck.pwcheck.host_path' & `pck.pwcheck.class_path' is
   * defined here.  If `PCKinvalidpwhost/class' directive is not
   * found in `proftpd.conf', _PATH_INVALIDPWHOST/CLASS in
   * pwcheck.h is set as `pck.path.invalidpw'.  
s   */
  if(get_param_int(main_server->conf,"PCKinvalidpwhost",FALSE) != -1) {
    pck_tmp = find_config(main_server->conf, 
			  CONF_PARAM, 
			  "PCKinvalidpwhost", 
			  TRUE);
    if(pck_tmp->argv[0]) {
      pck.pwcheck.host_path = pck_expand_tilde((char*)pck_tmp->argv[0]);
    }
  }
  
  if( !(pck.pwcheck.host_path && *pck.pwcheck.host_path )) {
    pck.pwcheck.host_path = pck_expand_tilde(_PATH_INVALIDPWHOST);
  }
  
  if( pck.debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: PCKinvalidpwhost is [%s]", 
	   pck.pwcheck.host_path);

  if(get_param_int(main_server->conf,"PCKinvalidpwclass",FALSE) != -1) {
    pck_tmp = find_config(main_server->conf, 
			  CONF_PARAM, 
			  "PCKinvalidpwclass", 
			  TRUE);
    if(pck_tmp->argv[0]) {
      pck.pwcheck.class_path = pck_expand_tilde((char*)pck_tmp->argv[0]);
    }
  }
  
  if( !(pck.pwcheck.class_path && *pck.pwcheck.class_path )) {
    pck.pwcheck.class_path = pck_expand_tilde(_PATH_INVALIDPWCLASS);
  }
  
  if( pck.debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: PCKinvalidpwclass is [%s]", 
	   pck.pwcheck.class_path);

  /*
   * `pck.path.errmsg' file is defined here.
   * If `PCKinvalidpw' directive is not found in `conf',
   * _PATH_PWDERRMSG in pathnames.h is set as `pck.path.errmsg'.
   */
  if(get_param_int(main_server->conf,"PCKerrmsg",FALSE) != -1) {
    pck_tmp = find_config(main_server->conf, 
			  CONF_PARAM, 
			  "PCKerrmsg", 
			  TRUE);
    if(pck_tmp->argv[0]) {
      pck.path.errmsg = pck_expand_tilde((char*)pck_tmp->argv[0]);
    }
  }
  
  if( !(pck.path.errmsg && *pck.path.errmsg )) {
    pck.path.errmsg = pck_expand_tilde(_PATH_PWDERRMSG);
  }
  
  if( pck.debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: PCKerrmsg    is [%s]", pck.path.errmsg);

  /*
   * update `dot-check' mode
   */
  switch(get_param_int(main_server->conf,"PCKdotcheck",FALSE))
    {
    case FALSE: 
      pck.dot_check = FALSE; 
      if ( pck.debug & PCK_DEBUG_NORMAL ) 
	log_pri (L_TRACE, "---: dot-check OFF");
      break;
    case TRUE: 
      pck.dot_check = TRUE; 
      if ( pck.debug & PCK_DEBUG_NORMAL ) 
	log_pri (L_TRACE, "---: dot-check ON");
      break;
    default: 
      pck.dot_check = TRUE; 
      if ( pck.debug & PCK_DEBUG_NORMAL ) 
	log_pri (L_TRACE, "---: dot-check ON <default>");
      break;
    }
  
  /*
   * update `no-more-check' mode
   */
  switch(get_param_int(main_server->conf,"PCKnomorecheck",FALSE))
    {
    case FALSE: 
      pck.no_more_check = FALSE; 
      if ( pck.debug & PCK_DEBUG_NORMAL ) 
	log_pri (L_TRACE, "---: no-more-check OFF");
      break;
    case TRUE: 
      pck.no_more_check = TRUE; 
      if ( pck.debug & PCK_DEBUG_NORMAL ) 
	log_pri (L_TRACE, "---: no-more-check ON");
      break;
    default: 
      pck.no_more_check = TRUE; 
      if ( pck.debug & PCK_DEBUG_NORMAL ) 
	log_pri (L_TRACE, "---: no-more-check ON <default>");
      break;
    }
  
  return;
}


int pck_exit()
{
  int authcode = Valid;	/* By default, enable user's login */
  
  if ( pck.debug & PCK_DEBUG_NORMAL ) {
    log_pri (L_TRACE, "---: password-check done.  valid=[%d]",pck.valid);
  }
  
  /*
   * define "PCKdenymessage" message if it exists in conf file.
   */
  if(get_param_int(main_server->conf,"PCKdenymessage",FALSE) != -1) {
    
    pck.denymsg = (char *)get_param_ptr(main_server->conf,
				       "PCKdenymessage",FALSE);
  }
  
  /*
   * define "PCKdenymessage" message if it exists in conf file.
   */
  if(get_param_int(main_server->conf,"PCKgrantmessage",FALSE) != -1) {
    pck.grantmsg = (char *)get_param_ptr(main_server->conf,
					"PCKgrantmessage",FALSE);
  }
  
  /*
   * issue "PCKdenymessage" or "PCKgrantmessage"
   * if it exists in `conf' file.
   */
  if ( !pck.valid ) {		/* password not valid! (deny) */
    log_pri (L_INFO, "PCK: %s", pck.denymsg); /* deny */
    /*
     * open and display error message
     */
    if( pck_core_display_file("530", (char *)pck.path.errmsg) ) {
      log_pri (L_ERROR, "ERR: Cannot Open PWDERRMSG file %s: %s",
	     pck.path.errmsg, strerror(errno));
      send_response("530",
		    "Password-check failed... " 
		    "Server denies anonymous login.");
    }
    authcode = Not_Valid;
    return authcode;
    /*goto auth_failure;*/
  } else {			/* password valid! (accept) */
    /*
     * open and display grant message
     */
#if 0
    log_pri (L_INFO, "PCK: %s",pck.grantmsg); /* accept */
#endif
  }
  
  return authcode;
}

/*
 * Return code = 0  .... normal
 * Return code = -1 .... File not foud
 */
int pck_override_display_file(const char *code, char *file)
{
  register FILE *infile;
  char *pointer;
  int return_code = 0;
  char *blank="";

  infile = fopen(file, "r");

  if (infile != NULL) {
    while (fgets(pck_buf0, sizeof(pck_buf0), infile)) {
      if ((pointer = strchr(pck_buf0, '\n')) != NULL)
	*pointer = '\0';
#if 0
      pck_msg_massage(pck_buf0, pck_buf1, sizeof(pck_buf0));
      pck_lreply(atoi(code), "%s", pck_buf0);
#else
      /* Suppress cookie replacement to avoid errors */
      (void) strcpy(pck_buf0, pck_remove_one_char(pck_buf0,'%'));
      add_response(code, pck_buf0);
#endif
    }

    fclose(infile);

    if(!strcasecmp(code, "530")){
      pck_reply(atoi(code), "");
    } else {
#if 0
      pck_lreply(atoi(code), "Administrative Override. Permission granted.");
      pck_lreply(atoi(code), "");
#else
      add_response(code,blank);
      add_response(code,"Administrative Override. Permission granted.");
      add_response(code,blank);
#endif
    }
  } else {
    return_code = -1; /* File not found */
  }
  return return_code;
}

int pck_core_display_file(const char *numeric, const char *fn)
{
  fsdir_t *fp;
  char buf[1024] = {'\0'};
  int len, max, fd, classes_enabled;
  unsigned long fs_size = 0;
  pool *p;
  xaset_t *s;
  char *outs, *mg_time, mg_size[12] = {'\0'}, mg_max[12] = "unlimited";
  char mg_class_limit[12] = {'\0'}, mg_cur[12] = {'\0'},
       mg_xfer_bytes[12] = {'\0'}, mg_cur_class[12] = {'\0'};
  char mg_xfer_units[12] = {'\0'}, config_class_users[128] = {'\0'}, *user;
  short first = 1;

#if defined(HAVE_SYS_STATVFS_H) || defined(HAVE_SYS_VFS_H)
  fs_size = get_fs_size((char*)fn);
  snprintf(mg_size, sizeof(mg_size), "%lu", fs_size);
#endif

  if((fp = fs_open_canon(fn,O_RDONLY,&fd)) == NULL)
    return -1;

  p = make_sub_pool(permanent_pool);

  s = (session.anon_config ? session.anon_config->subset : main_server->conf);

  mg_time = fmt_time(time(NULL));
  snprintf(mg_size, sizeof(mg_size), "%lu",fs_size);
  max = get_param_int(s,"MaxClients",FALSE);
  snprintf(mg_cur, sizeof(mg_cur), "%d",(int)get_param_int(main_server->conf,
          "CURRENT-CLIENTS",FALSE)+1);

  if((classes_enabled = get_param_int(CURRENT_CONF,"Classes",FALSE)) < 0)
    classes_enabled = 0;
  
  if (classes_enabled && session.class && session.class->name) {
	snprintf(config_class_users,sizeof(config_class_users),"%s-%s","CURRENT-CLIENTS-CLASS",session.class->name);
	snprintf(mg_cur_class,sizeof(mg_cur_class),"%d",(int)get_param_int(main_server->conf,config_class_users,FALSE)+1);
	snprintf(mg_class_limit, sizeof(mg_class_limit), "%u",
		 session.class->max_connections);
  } else {
	mg_cur_class[0] = 0;
	snprintf(mg_class_limit, sizeof(mg_class_limit), "%u",max);
  }
   
  snprintf(mg_xfer_bytes, sizeof(mg_xfer_bytes), "%lu",
	   session.total_bytes >> 10);
  
  snprintf(mg_xfer_units, sizeof(mg_xfer_units), "%luB",
	   session.total_bytes);
  
  if(session.total_bytes >= 10240) {
    snprintf(mg_xfer_units, sizeof(mg_xfer_units), "%lukB",
	     session.total_bytes >> 10);
  } else if ((session.total_bytes >> 10) >= 10240) {
    snprintf(mg_xfer_units, sizeof(mg_xfer_units), "%luMB",
	     session.total_bytes >> 20);
  } else if ((session.total_bytes >> 20) >= 10240) {
    snprintf(mg_xfer_units, sizeof(mg_xfer_units), "%luGB",
	     session.total_bytes >> 30);
  }

  if(max != -1)
    snprintf(mg_max, sizeof(mg_max), "%d",max);

  user = (char*)get_param_ptr(main_server->conf,"USER",FALSE);
  if (user == 0)
      user = "";
   
  while(fs_gets(buf,sizeof(buf),fp,fd) != NULL) {
    buf[sizeof(buf)-1] = '\0';

    len = strlen(buf);

    while(len && (buf[len-1] == '\r' || buf[len-1] == '\n')) {
      buf[len-1] = '\0';
      len--;
    }

    outs = sreplace(p,buf,
		    "%i", pck.remote.ip,
		    "%c", pck.local.cname,
		    "%P", (pck.remote.anonname ? 
			   (*pck.remote.pass ? pck.remote.pass : "(none)") :
			   "(hidden)"),
		    "%T", mg_time,
		    "%F", mg_size,
		    "%C", (session.cwd[0] ? session.cwd : "(none)"),
		    "%R", (session.c && session.c->remote_name ?
			   session.c->remote_name : "(unknown)"),
		    "%L", main_server->ServerFQDN,
		    "%u", session.ident_user,
		    "%U", user,
		    "%k", mg_xfer_units,
		    "%K", mg_xfer_bytes,
		    "%M", mg_max,
		    "%N", mg_cur,
		    "%E", main_server->ServerAdmin,
		    "%V", main_server->ServerName,
		    "%x", (classes_enabled && session.class) ?
		    session.class->name : "",
		    "%y", mg_cur_class,
		    "%z", mg_class_limit,
		    NULL);
    
    if(first) {
      send_response_raw("%s-%s",numeric,outs);
      first=0;
    } else {
      if(MultilineRFC2228)
        send_response_raw("%s-%s",numeric,outs);
      else
        send_response_raw(" %s",outs);
      }
  }

  fs_close(fp,fd);
  return 0;
}

/*
 *  Split text into comma-separated fields.  Return an allocated
 *  NULL-terminated array of the fields within the modified argument that
 *  the caller is expected to save or free.  We don't use strchr() since
 *  the text is expected to be either relatively short or "comma-dense."
 */
char **
pck_comma_split( char *text )
{
  register int    i;
  register char   *p;
  register char   **av;
  char            **save;

  for (i = 2, p = text; *p; p++)
    if (*p == ',')
      i++;

  for (av = save = NEW(char*, i), *av++ = p = text; *p; )
    if (*p == ',') {
      *p++ = '\0';
      *av++ = p;
    }
    else
      p++;

  *av = NULL;
  return save;
}

/*
 * Construct a filename from the NULL terminated list of parts.
 * Tilde replacement is supported here.
 * Never fail to terminate with "NULL" character.
 *
 * Example:
 * char *out;
 * out = pck_make_filename("~/DIR1","DIR2","DIR3","DIR4","FILENAME",NULL);
 * out ===> [/home/star/DIR1/DIR2/DIR3/DIR4/FILENAME]
 *
 */
char *
pck_make_filename( const char *first_part, ... )
{
  va_list arg_ptr ;
  size_t n;
  const char *s;
  char *name, *home = NULL, *p;

  va_start( arg_ptr, first_part ) ;
  n = strlen(first_part)+1;

  while( (s=va_arg(arg_ptr, const char *)) )
    n += strlen(s) + 1;
  
  va_end(arg_ptr);
  
  if( *first_part == '~' && first_part[1] == '/' ) {
#if 1
    /*
     * Get home_dir of "ftp" user.
     */
    home = pck_get_home_dir("ftp");
    
    if( *home )
      n += strlen(home);
#else
    /*
     * Get home_dir of the user who owns the process.
     * (Normally, "nobody" in proftpd)
     */
    home = NULL;
    if( *first_part == '~' && first_part[1] == '/'
	&& (home = getenv("HOME")) && *home )
      n += strlen(home);
#endif
  } else if( *first_part == '.' && first_part[1] == '/' ) {
    
    /* Force "TRUE" if no "PCKmsgprefix" is found! */
    if( !pck.msg_prefix )
      pck.msg_prefix = TRUE;

    home = pck_get_system_dir();

    if( *home )
      n += strlen(home);
  }

  name = (char *)malloc(n);

  p = home ?
    pck_stpcpy(pck_stpcpy(name,home), first_part+1) :
    pck_stpcpy(name, first_part);

  va_start( arg_ptr, first_part ) ;

  while( (s=va_arg(arg_ptr, const char *)) )
    p = pck_stpcpy(pck_stpcpy(p,"/"), s);

  va_end(arg_ptr);

  return name;
}

char *
pck_get_system_dir()
{
  if( !pck.msg_prefix ) return NULL;

  if( pck.path.msg_prefix && *pck.path.msg_prefix )
    return (char *)pck.path.msg_prefix;
  else
    return NULL;
}

char *
pck_expand_tilde(char *path)
{
  char *expanded_path = NULL;
  
  if ( *path != '~' && *path != '.' )
    return path;

  expanded_path = pck_make_filename(path, NULL);
  return expanded_path;
}

char *
pck_get_home_dir(char *user_name)
{
#  if defined HAVE_GETPWNAM_R || defined _LIBC
  size_t buflen = sysconf (_SC_GETPW_R_SIZE_MAX);
  char *pwtmpbuf = (char *) alloca (buflen);
  struct passwd pwbuf, *p;

  if (__getpwnam_r (user_name, &pwbuf, pwtmpbuf, buflen, &p) >= 0)
    home_dir = p->pw_dir;
  else
    home_dir = NULL;
#else
  char *home_dir;
  struct passwd *p = getpwnam (user_name);
  
  if (p != NULL)
    home_dir = p->pw_dir;
  else
    home_dir = NULL;
#endif

  return home_dir;
}

#ifndef HAVE_STPCPY
char *
pck_stpcpy( char *a, const char *b )
{
  while( *b )
    *a++ = *b++;
  *a = 0;

  return (char*)a;
}
#endif

void pck_display_login_message()
{
  char *dummy;
  char *blank="";

  /* OK. we display a grant message to the remote user finally */
  if( pck.remote.anonname      && 
      pck.enable  ) {

    /* "PCKdisplaylogin" is TRUE */
    if( pck.display_login ) {
      if( pck_core_display_file("230",
				(char *)pck.path.display_login) ) {
	log_pri (L_ERROR, "ERR: Cannot Open DISPLAYLOGIN file %s: %s",
	       pck.path.display_login, strerror(errno));
	add_response("230",blank);
	add_response("230","Welcome to %s ftp server!",
		     pck.local.cname);
	add_response("230","Everything you do is logged... So be nice!");
      }

    /* "PCKdisplaylogin" is FALSE */
    } else {

      dummy = (char *) get_param_ptr( main_server->conf,
				     "DisplayLogin",
				     FALSE);

      /* "DisplayLogin" is FALSE */
      if( !dummy ) {
	add_response("230",blank);
	add_response("230","Welcome to %s ftp server!!",
		     pck.local.cname);
	add_response("230","Everything you do is logged... So be nice!!");
      }

    }
    
  }
  return;
}

void pck_respond_grant_message(cmd_rec *command_rec) 
{
  char *blank="";

  add_response(R_230,blank);

  if( pck.debug & PCK_DEBUG_NONE ) {	/* NOT REACHED */
    add_response(R_DUP,"##############################################");
    add_response(R_DUP,"ACCESS GRANTED...");
    add_response(R_DUP,"SUCCESSFULLY CONNECTED TO %s.",
		 command_rec->server->ServerFQDN);
    add_response(R_DUP,"PLEASE CONTACT %s",
		 command_rec->server->ServerAdmin);
    add_response(R_DUP,"       FOR MORE INFORMATION...");
    add_response(R_DUP,"##############################################");
    add_response(R_DUP,blank);
  }
  
  return;
}

void pck_respond_deny_message(cmd_rec *command_rec) 
{
  char *blank="";

  add_response_err(R_530,blank);

  if( pck.debug & PCK_DEBUG_NONE ) {	/* NOT REACHED */
    add_response_err(R_DUP,"##############################################");
    add_response_err(R_DUP,"ACCESS DENIED...");
    add_response_err(R_DUP,"UNABLE TO CONNECT TO %s.",
		     command_rec->server->ServerFQDN);
    add_response_err(R_DUP,"PLEASE CONTACT %s",
		     command_rec->server->ServerAdmin);
    add_response_err(R_DUP,"       FOR MORE INFORMATION...");
    add_response_err(R_DUP,"##############################################");
    add_response_err(R_DUP,blank);
  }
  
  return;
}

/*
 * This function was borrowed from module/mod_auth.c.
 */
static void pck_build_group_arrays(pool *p, struct passwd *pw, char *name,
			       array_header **gids, array_header **groups)
{
  struct group *gr;
  array_header *xgids,*xgroups;

#ifdef HAVE_INITGROUPS
  /* Optimization for large groups, especially.
   */
  int i, total;
  
  i = total = 0;

  /* If we don't have a password file, we create an empty array and punt.
   */
  if(!pw) {
    if(!name || !(pw = (struct passwd *) auth_getpwnam(p, name))) {
      *gids = make_array(p, 2, sizeof(gid_t));
      *groups = make_array(p, 2, sizeof(char *));
      return;
    }
  }
  
  /* We need root privileges for initgroups to succeed.
   */
  PRIVS_ROOT;
  (void) initgroups((const char*)pw->pw_name, (gid_t)pw->pw_gid);
  PRIVS_RELINQUISH;
  
  /* Allocate the appropriate space.
   */
  total = getgroups(0, 0);
  xgids = make_array(p, total, sizeof(gid_t));
  xgroups = make_array(p, 2, sizeof(char *));
  
  /* Populate the array of GIDs.
   */
  xgids->nelts = getgroups(total, (GETGROUPS_T *) xgids->elts);
  
  /* Now populate the names of the groups.
   */
  for(i = 0; i < total; i++) {
    if((gr = (struct group *) auth_getgrgid(p, ((gid_t *) xgids->elts)[i])))
      *((char **) push_array(xgroups)) = pstrdup(p, gr->gr_name);
  }
  
#else /* No HAVE_INITGROUPS */

  char **gr_mem;
  
  /* Allocate space.
   */
  xgids = make_array(p, 2, sizeof(gid_t));
  xgroups = make_array(p, 2, sizeof(char *));
  
  /* Nothing there...punt.
   */
  if(!pw) {
    if(!name || !(pw = (struct passwd *) auth_getpwnam(p, name))) {
      *gids = xgids;
      *groups = xgroups;
      return;
    }
  }
  
  /* Populate the first group name.
   */
  if((gr = auth_getgrgid(p, pw->pw_gid)) != NULL)
    *((char **) push_array(xgroups)) = pstrdup(p, gr->gr_name);
  
  auth_setgrent(p);
  
  /* This is where things get slow, expensive, and ugly.
   * Loop through everything, checking to make sure we haven't already added
   * it.  This is why we have getgroups() and company.
   */
  while((gr = auth_getgrent(p)) != NULL && gr->gr_mem)
    for(gr_mem = gr->gr_mem; *gr_mem; gr_mem++) {
      if(!strcmp(*gr_mem, pw->pw_name)) {
        *((int *) push_array(xgids)) = (int) gr->gr_gid;
        if(pw->pw_gid != gr->gr_gid)
          *((char **) push_array(xgroups)) = pstrdup(p, gr->gr_name);
        break;
      }
    }
#endif /* HAVE_INITGROUPS */
  
  *gids = xgids;
  *groups = xgroups;
}

/*
 * This function was borrowed from module/mod_auth.c.
 */
static config_rec *pck_anonymous_group(pool *p, char *user)
{
  config_rec *c;
  int ret = 0;
  
  pck_build_group_arrays(p,NULL,user,&session.gids,&session.groups);
  c = find_config(main_server->conf,CONF_PARAM,"AnonymousGroup",FALSE);
  
  if(c) do {
    ret = group_expression((char**)c->argv);
  } while(!ret && (c = find_config_next(c,
					c->next,
					CONF_PARAM,
					"AnonymousGroup",
					FALSE)) != NULL);
  
  return ret ? c : NULL;
}

/*
 * Determine user's access tables.
 * This function was borrowed from module/mod_auth.c.
 */
config_rec *pck_resolve_user(pool *pool, 
			     char **user, 
			     char **ournamep,
			     char **anonnamep) {
  
  config_rec *conf, *top_conf;
  char *ourname,*anonname = NULL;
  int is_alias = 0, force_anon = 0;
  
  /* Precendence rules:
   *   1. Search for UserAlias directive.
   *   2. Search for Anonymous directive.
   *   3. Normal user login
   */
  
  ourname = (char*) get_param_ptr(main_server->conf, "UserName", FALSE);
  
  if (ournamep && ourname)
    *ournamep = ourname; 
  
  conf = find_config(main_server->conf, CONF_PARAM, "UserAlias", TRUE);
  
  if (conf) do {
    if (!strcmp(conf->argv[0], "*") || !strcmp(conf->argv[0], *user)) {
      is_alias = 1;
      break;
    }  
    
  } while ((conf = find_config_next(conf, conf->next, CONF_PARAM,
				    "UserAlias", TRUE)) != NULL);
  
  /* if AuthAliasOnly is set, ignore this one and continue */
  top_conf = conf;
  
  while (conf && conf->parent &&
	 find_config(conf->parent->set, CONF_PARAM, "AuthAliasOnly", FALSE)) {
    
    is_alias = 0;
    find_config_set_top(top_conf);
    conf = find_config_next(conf, conf->next, CONF_PARAM, "UserAlias", TRUE);
    
    if (conf && (!strcmp(conf->argv[0], "*") || !strcmp(conf->argv[0], *user)))
      is_alias = 1;
  }
  
  if (conf) {
    *user = conf->argv[1];
    
    /* If the alias is applied inside an <Anonymous> context, we have found
     * our anon block
     */
    
    if (conf->parent && conf->parent->config_type == CONF_ANON)
      conf = conf->parent;
    
    else
      conf = NULL;
  }
  
  /* Next, search for an anonymous entry */
  
  if (!conf)
    conf = find_config(main_server->conf, CONF_ANON, NULL, FALSE);
  
  else
    find_config_set_top(conf);
  
  if (conf) do {
    anonname = (char*) get_param_ptr(conf->subset, "UserName", FALSE);
    
    if (!anonname)
      anonname = ourname;
    
    if (anonname && !strcmp(anonname,*user)) {
      if (anonnamep)
	*anonnamep = anonname;
      break;
    }
  } while ((conf = find_config_next(conf, 
				    conf->next, 
				    CONF_ANON, 
				    NULL,
				    FALSE)) != NULL);
  if(!conf) {
    conf = pck_anonymous_group(pool, *user);
    
    if(conf)
      force_anon = 1;
  }

  if (!is_alias && !force_anon) {
    
    if (find_config((conf ? conf->subset :
		     main_server->conf), CONF_PARAM, "AuthAliasOnly", FALSE)) {
      
      if (conf && conf->config_type == CONF_ANON)
        conf = NULL;
      else
        *user = NULL;
      
      if (*user && find_config(main_server->conf, 
			       CONF_PARAM, 
			       "AuthAliasOnly",
			       FALSE))
        *user = NULL;
      
      if ((!user || !conf) && anonnamep)
        *anonnamep = NULL;
    }
  }
  
  return conf;
}

void pck_path_define_pck(char *directive, 
			 char dest[],
			 const char *path)
{
  if(!strcasecmp(directive, "PCKcheckmodehost")) {
    if(path) {
      (void) strcpy(dest, pck_expand_tilde((char*)path));
    } else {
      /* tilde replacement is already done for pck.pwcheck.host_path! */
      (void) strcpy(dest, pck.pwcheck.host_path);
    }
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: %s path is [%s]", 
	     directive, dest);

  } else if(!strcasecmp(directive, "PCKcheckmodeclass")) {
    if(path) {
      (void) strcpy(dest, pck_expand_tilde((char*)path));
    } else {
      /* tilde replacement is already done for pck.pwcheck.class_path! */
      (void) strcpy(dest, pck.pwcheck.class_path);
    }
    if( pck.debug & PCK_DEBUG_HIGH )
      log_pri (L_TRACE, "---: %s path is [%s]", 
	     directive, dest);
  }
  return;
}

void pck_init_checkmode()
{
      
  (void) strcpy((char *)pck.check_mode[0], "file+dns");
  (void) strcpy((char *)pck.check_mode[1], "enforce");
  (void) strcpy((char *)pck.check_mode[2], "remove");
  (void) strcpy((char *)pck.check_mode[3], "*");
  (void) strcpy((char *)pck.check_mode[4], "");

  return;
}

void
pck_dump(PCK_PTR x)
{
  register FILE *infile;
  
  if( !x->dump )
    return;
  
  log_pri (L_INFO, "DMP: PCK parameters written to [%s]", x->path.dump);

  /* open file in "append" mode */
  infile = fopen(x->path.dump, "a");
  
  if (infile != NULL) {
    fprintf(infile, "###################################");
    fprintf(infile, "###################################\n");
    fprintf(infile, "enable                     = [%s]\n", 
	    YesNo(x->enable));
    fprintf(infile, "valid                      = [%s]\n", 
	    YesNo(x->valid));
    fprintf(infile, "debug                      = [%s]\n", 
	    YesNo(x->debug));
    fprintf(infile, "dot_check                  = [%s]\n", 
	    YesNo(x->dot_check));
    fprintf(infile, "no_more_check              = [%s]\n", 
	    YesNo(x->no_more_check));

    /* server dead */
    fprintf(infile, "server_dead.class          = [%s]\n", 
	    YesNo(x->server_dead.class));
    fprintf(infile, "server_dead.class_path     = [%s]\n", 
	    x->server_dead.class_path);
    fprintf(infile, "server_dead.host           = [%s]\n", 
	    YesNo(x->server_dead.host));
    fprintf(infile, "server_dead.host_path      = [%s]\n", 
	    x->server_dead.host_path);

    /* reverse_dns */
    fprintf(infile, "reverse_dns.class          = [%s]\n", 
	    YesNo(x->reverse_dns.class));
    fprintf(infile, "reverse_dns.class_override = [%s]\n", 
	    YesNo(x->reverse_dns.class_override));
    fprintf(infile, "reverse_dns.class_path     = [%s]\n", 
	    x->reverse_dns.class_path);
    fprintf(infile, "reverse_dns.host           = [%s]\n", 
	    YesNo(x->reverse_dns.host));
    fprintf(infile, "reverse_dns.host_override  = [%s]\n", 
	    YesNo(x->reverse_dns.host_override));
    fprintf(infile, "reverse_dns.host_path      = [%s]\n", 
	    x->reverse_dns.host_path);

    /* matching_dns */
    fprintf(infile, "matching_dns.class         = [%s]\n", 
	    YesNo(x->matching_dns.class));
    fprintf(infile, "matching_dns.class_override= [%s]\n", 
	    YesNo(x->matching_dns.class_override));
    fprintf(infile, "matching_dns.class_path    = [%s]\n", 
	    x->matching_dns.class_path);
    fprintf(infile, "matching_dns.host          = [%s]\n", 
	    YesNo(x->matching_dns.host));
    fprintf(infile, "matching_dns.host_override = [%s]\n", 
	    YesNo(x->matching_dns.host_override));
    fprintf(infile, "matching_dns.host_path     = [%s]\n", 
	    x->matching_dns.host_path);

    /* goaway */
    fprintf(infile, "goaway.class               = [%s]\n", 
	    YesNo(x->goaway.class));
    fprintf(infile, "goaway.class_path          = [%s]\n", 
	    x->goaway.class_path);
    fprintf(infile, "goaway.host                = [%s]\n", 
	    YesNo(x->goaway.host));
    fprintf(infile, "goaway.host_path           = [%s]\n", 
	    x->goaway.host_path);

    /* pwcheck */
    fprintf(infile, "pwcheck.class_path         = [%s]\n", 
	    x->pwcheck.class_path);
    fprintf(infile, "pwcheck.host_path          = [%s]\n", 
	    x->pwcheck.host_path);

    /* path */
    fprintf(infile, "path.invalidpw             = [%s]\n", 
	    x->path.invalidpw);
    fprintf(infile, "path.errmsg                = [%s]\n", 
	    x->path.errmsg);
    fprintf(infile, "path.display_login         = [%s]\n", 
	    x->path.display_login);
    fprintf(infile, "path.dump                  = [%s]\n", 
	    x->path.dump);

    fprintf(infile, "display_login              = [%s]\n", 
	    YesNo(x->display_login));
    fprintf(infile, "dump                       = [%s]\n", 
	    YesNo(x->dump));
    fprintf(infile, "denymsg                    = [%s]\n", 
	    x->denymsg);
    fprintf(infile, "grantmsg                   = [%s]\n", 
	    x->grantmsg);

    /* remote info */
    fprintf(infile, "remote.name                = [%s]\n", 
	    x->remote.name);
    fprintf(infile, "remote.ip                  = [%s]\n", 
	    x->remote.ip);
    fprintf(infile, "remote.class               = [%s]\n", 
	    x->remote.class);
    fprintf(infile, "remote.anonname            = [%s]\n", 
	    x->remote.anonname);
    fprintf(infile, "remote.pass                = [%s]\n", 
	    x->remote.pass);
    fprintf(infile, "remote.class_defined       = [%s]\n", 
	    YesNo(x->remote.class_defined));

    /* local info */
    fprintf(infile, "local.name                 = [%s]\n", 
	    x->local.name);
    fprintf(infile, "local.ip                   = [%s]\n", 
	    x->local.ip);
    fprintf(infile, "local.cname                = [%s]\n", 
	    x->local.cname);
    fprintf(infile, "local.name_server          = [%s]\n", 
	    x->local.name_server);

    fprintf(infile, "index                      = [%s]\n", 
	    x->index);
    fprintf(infile, "check_mode[0]              = [%s]\n", 
	    x->check_mode[0]);
    fprintf(infile, "check_mode[1]              = [%s]\n", 
	    x->check_mode[1]);
    fprintf(infile, "check_mode[2]              = [%s]\n", 
	    x->check_mode[2]);
    fprintf(infile, "check_mode[3]              = [%s]\n", 
	    x->check_mode[3]);
    fprintf(infile, "check_mode[4]              = [%s]\n", 
	    x->check_mode[4]);
    fprintf(infile, "check_order[0]             = [%s]\n", 
	    x->check_order[0]);
    fprintf(infile, "check_order[1]             = [%s]\n", 
	    x->check_order[1]);
    fprintf(infile, "check_order[2]             = [%s]\n", 
	    x->check_order[2]);
    fprintf(infile, "check_order[3]             = [%s]\n", 
	    x->check_order[3]);
    fprintf(infile, "check_order[4]             = [%s]\n", 
	    x->check_order[4]);
    fprintf(infile, "dns_order[0]               = [%s]\n", 
	    x->dns_order[0]);
    fprintf(infile, "dns_order[1]               = [%s]\n", 
	    x->dns_order[1]);
  } else {
    log_pri (L_ERROR, "ERR: Cannot Open DUMP file %s: %s",
	   x->path.dump, strerror(errno));
    return;
  }
    
  fclose(infile);

  return;
}

void
pck_free()
{

  if( !pck.remote.anonname || !pck.enable ) 
    return;

  if( pck.debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: pck_xfree - dispose memory! (in)");

  if(pck.local.ip)     pck_dispose(pck.local.ip);
  if(pck.remote.ip)    pck_dispose(pck.remote.ip);
  if(pck.remote.name)  pck_dispose(pck.remote.name);
  if(pck.remote.class) pck_dispose(pck.remote.class);

  if( pck.debug & PCK_DEBUG_HIGH )
    log_pri (L_TRACE, "---: pck_xfree - dispose memory! (out)");

  return;
}

void 
pck_xfree(void *ptr)
{
  if (ptr == NULL) {
    log_pri (L_ERROR, "ERR: NULL pointer given as argument for pck_xfree...");
#if 1
    return;
#else
    exit(1);
#endif
  }

  (void) free(ptr);

  return;
}

MODRET
pck_add_bool (cmd_rec * cmd)
{
  int b;
  config_rec *c;

  CHECK_ARGS (cmd, 1);	/* minimum requirement: 1 argument! */

  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);

  if(cmd->argc < 2 || cmd->argc > 2)	/* only 1 argument is allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");

  if((b = get_boolean(cmd, 1)) == -1)
    CONF_ERROR(cmd, "Expected boolean argument...");

  c = add_config_param (cmd->argv[0], 1, (void *)b);

  c->flags |= CF_MERGEDOWN;

  return HANDLED (cmd);
}

MODRET
pck_add_bool_str (cmd_rec * cmd)
{
  int b;
  config_rec *c;

  CHECK_ARGS (cmd, 1);	/* minimum requirement: 1 argument! */

  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);

  if(cmd->argc < 2 || cmd->argc > 3)	/* 1 or 2 arguments are allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");

  if((b = get_boolean(cmd,1)) == -1)
    CONF_ERROR(cmd,"Expected boolean argument.");

  /*
   * Arguments (cmd->argv[1]) is _NOT_ stored
   * when cmd->argv[0] is "FALSE"!!
   */
  if(b && cmd->argc == 3) {
    c = add_config_param(cmd->argv[0],2,(void*)!b,NULL);
    c->argv[1] = pstrdup(permanent_pool,cmd->argv[2]);
  } else
    c = add_config_param(cmd->argv[0],1,(void*)!b);

  c->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);

}

MODRET
pck_add_bool_str_ugly (cmd_rec * cmd)
{
  int b;
  config_rec *c;

  CHECK_ARGS (cmd, 1);	/* minimum requirement: 1 argument! */

  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);

  if(cmd->argc < 2 || cmd->argc > 3)	/* 1 or 2 arguments are allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");

  if((b = get_boolean(cmd,1)) == -1)
    CONF_ERROR(cmd,"Expected boolean argument.");

  /*
   * Arguments (cmd->argv[1]) is stored
   * even if cmd->argv[0] is "FALSE"!!
   */
#if 1
  if(cmd->argc == 3) {
#else
  if(b && cmd->argc == 3) {
#endif
    c = add_config_param(cmd->argv[0],2,(void*)!b,NULL);
    c->argv[1] = pstrdup(permanent_pool,cmd->argv[2]);
  } else
    c = add_config_param(cmd->argv[0],1,(void*)!b);

  c->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);

}

MODRET
pck_add_bool_str2 (cmd_rec * cmd)
{
  int b;
  config_rec *c;

  CHECK_ARGS (cmd, 1);	/* minimum requirement: 1 argument! */

  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);

  if(cmd->argc < 2 || cmd->argc > 4)	/* 1,2 or 3 arguments are allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");

  if((b = get_boolean(cmd,1)) == -1)
    CONF_ERROR(cmd,"Expected boolean argument.");

  /*
   * Arguments (cmd->argv[1]-cmd->argv[2]) are _NOT_ stored
   * when cmd->argv[0] is "FALSE"!!
   */
  if(b && cmd->argc == 4) {		/* cmd->argv[0] SHOULD be boolean! */
    c = add_config_param(cmd->argv[0],3,(void*)!b,NULL);
    c->argv[1] = pstrdup(permanent_pool,cmd->argv[2]);
    c->argv[2] = pstrdup(permanent_pool,cmd->argv[3]);
  } else if(b && cmd->argc == 3) {	/* cmd->argv[0] SHOULD be boolean! */
    c = add_config_param(cmd->argv[0],2,(void*)!b,NULL);
    c->argv[1] = pstrdup(permanent_pool,cmd->argv[2]);
  } else
    c = add_config_param(cmd->argv[0],1,(void*)!b);

  c->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);

}

MODRET
pck_add_bool_str2_ugly (cmd_rec * cmd)
{
  int b;
  config_rec *c;

  CHECK_ARGS (cmd, 1);	/* minimum requirement: 1 argument! */

  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);

  if(cmd->argc < 2 || cmd->argc > 4)	/* 1,2 or 3 arguments are allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");

  if((b = get_boolean(cmd,1)) == -1)
    CONF_ERROR(cmd,"Expected boolean argument.");

  /*
   * Arguments (cmd->argv[1]-cmd->argv[2]) are stored
   * even if cmd->argv[0] is "FALSE"!!
   */
#if 1
  if(cmd->argc == 4) {
#else
  if(b && cmd->argc == 4) {		/* cmd->argv[0] SHOULD be boolean! */
#endif
    c = add_config_param(cmd->argv[0],3,(void*)!b,NULL);
    c->argv[1] = pstrdup(permanent_pool,cmd->argv[2]);
    c->argv[2] = pstrdup(permanent_pool,cmd->argv[3]);
#if 1
  } else if(cmd->argc == 3) {
#else
  } else if(b && cmd->argc == 3) {	/* cmd->argv[0] SHOULD be boolean! */
#endif
    c = add_config_param(cmd->argv[0],2,(void*)!b,NULL);
    c->argv[1] = pstrdup(permanent_pool,cmd->argv[2]);
  } else
    c = add_config_param(cmd->argv[0],1,(void*)!b);

  c->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);

}

MODRET
pck_add_bool_str3 (cmd_rec * cmd)
{
  int b;
  config_rec *c;

  CHECK_ARGS (cmd, 1);	/* minimum requirement: 1 argument! */

  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);

  if(cmd->argc < 2 || cmd->argc > 5)	/* 1-4 arguments are allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");

  if((b = get_boolean(cmd,1)) == -1)
    CONF_ERROR(cmd,"Expected boolean argument.");

  /*
   * Arguments (cmd->argv[1]-cmd->argv[3]) are _NOT_ stored
   * when cmd->argv[0] is "FALSE"!!
   */
  if(b && cmd->argc == 5) {		/* cmd->argv[0] SHOULD be boolean! */
    c = add_config_param(cmd->argv[0],4,(void*)!b,NULL);
    c->argv[1] = pstrdup(permanent_pool,cmd->argv[2]);
    c->argv[2] = pstrdup(permanent_pool,cmd->argv[3]);
    c->argv[3] = pstrdup(permanent_pool,cmd->argv[4]);
  } else if(b && cmd->argc == 4) {	/* cmd->argv[0] SHOULD be boolean! */
    c = add_config_param(cmd->argv[0],3,(void*)!b,NULL);
    c->argv[1] = pstrdup(permanent_pool,cmd->argv[2]);
    c->argv[2] = pstrdup(permanent_pool,cmd->argv[3]);
  } else if(b && cmd->argc == 3) {	/* cmd->argv[0] SHOULD be boolean! */
    c = add_config_param(cmd->argv[0],2,(void*)!b,NULL);
    c->argv[1] = pstrdup(permanent_pool,cmd->argv[2]);
  } else
    c = add_config_param(cmd->argv[0],1,(void*)!b);

  c->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);

}

MODRET
pck_add_bool_str3_ugly (cmd_rec * cmd)
{
  int b;
  config_rec *c;

  CHECK_ARGS (cmd, 1);	/* minimum requirement: 1 argument! */

  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);

  if(cmd->argc < 2 || cmd->argc > 5)	/* 1-4 arguments are allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");

  if((b = get_boolean(cmd,1)) == -1)
    CONF_ERROR(cmd,"Expected boolean argument.");

  /*
   * Arguments (cmd->argv[1]-cmd->argv[3]) are stored
   * even if cmd->argv[0] is "FALSE"!!
   */
#if 1
  if(cmd->argc == 5) {
#else
  if(b && cmd->argc == 5) {		/* cmd->argv[0] SHOULD be boolean! */
#endif
    c = add_config_param(cmd->argv[0],4,(void*)!b,NULL);
    c->argv[1] = pstrdup(permanent_pool,cmd->argv[2]);
    c->argv[2] = pstrdup(permanent_pool,cmd->argv[3]);
    c->argv[3] = pstrdup(permanent_pool,cmd->argv[4]);
#if 1
  } else if(cmd->argc == 4) {
#else
  } else if(b && cmd->argc == 4) {	/* cmd->argv[0] SHOULD be boolean! */
#endif
    c = add_config_param(cmd->argv[0],3,(void*)!b,NULL);
    c->argv[1] = pstrdup(permanent_pool,cmd->argv[2]);
    c->argv[2] = pstrdup(permanent_pool,cmd->argv[3]);
#if 1
  } else if(cmd->argc == 3) {
#else
  } else if(b && cmd->argc == 3) {	/* cmd->argv[0] SHOULD be boolean! */
#endif
    c = add_config_param(cmd->argv[0],2,(void*)!b,NULL);
    c->argv[1] = pstrdup(permanent_pool,cmd->argv[2]);
  } else
    c = add_config_param(cmd->argv[0],1,(void*)!b);

  c->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);

}

MODRET
pck_add_str1 (cmd_rec * cmd)
{
  CHECK_ARGS (cmd, 1);	/* minimum requirement: 1 argument! */
  
  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);
  
  if(cmd->argc < 2 || cmd->argc > 2)	/* only 1 argument is allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");
  
  add_config_param_str (cmd->argv[0], 
			1,
			(void *) cmd->argv[1]);
  
  return HANDLED (cmd);
}
 
MODRET
pck_add_str2 (cmd_rec * cmd)
{
  CHECK_ARGS (cmd, 2);	/* minimum requirement: 2 arguments! */
  
  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);
  
  if(cmd->argc < 3 || cmd->argc > 3)	/* only 2 arguments are allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");
  
  add_config_param_str (cmd->argv[0], 
			2,
			(void *) cmd->argv[1],
			(void *) cmd->argv[2]);
  
  return HANDLED (cmd);
}

MODRET
pck_add_str3 (cmd_rec * cmd)
{
  CHECK_ARGS (cmd, 3);	/* minimum requirement: 3 arguments! */

  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);

  if(cmd->argc < 4 || cmd->argc > 4)	/* only 3 arguments are allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");
  
  add_config_param_str (cmd->argv[0], 
			3,
			(void *) cmd->argv[1],
			(void *) cmd->argv[2],
			(void *) cmd->argv[3]);
  
  return HANDLED (cmd);
}

MODRET
pck_add_str4 (cmd_rec * cmd)
{
  CHECK_ARGS (cmd, 4);	/* minimum requirement: 4 arguments! */

  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);

  if(cmd->argc < 5 || cmd->argc > 5)	/* only 4 arguments are allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");

  add_config_param_str (cmd->argv[0], 
			4,
			(void *) cmd->argv[1],
			(void *) cmd->argv[2],
			(void *) cmd->argv[3],
			(void *) cmd->argv[4]);
  
  return HANDLED (cmd);
}

MODRET
pck_add_str5 (cmd_rec * cmd)
{
  CHECK_ARGS (cmd, 5);	/* minimum requirement: 5 arguments! */
  
  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);
  
  if(cmd->argc < 6 || cmd->argc > 6)	/* only 5 arguments are allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");
  
  add_config_param_str (cmd->argv[0], 
			5,
			(void *) cmd->argv[1],
			(void *) cmd->argv[2],
			(void *) cmd->argv[3],
			(void *) cmd->argv[4],
			(void *) cmd->argv[5]);
  
  return HANDLED (cmd);
}

MODRET
pck_add_str45 (cmd_rec * cmd)
{
  CHECK_ARGS (cmd, 4);	/* minimum requirement: 4 arguments! */
  
  CHECK_CONF (cmd, CONF_ROOT|CONF_VIRTUAL|CONF_ANON|CONF_DIR|CONF_GLOBAL);
  
  if(cmd->argc < 5 || cmd->argc > 6)	/* 4 or 5 arguments are allowed */
    CONF_ERROR(cmd,"Invalid number of arguments");
  
  if( cmd->argc == 6 ) {	/* 5 arguments */
    add_config_param_str (cmd->argv[0], 
			  5,
			  (void *) cmd->argv[1],
			  (void *) cmd->argv[2],
			  (void *) cmd->argv[3],
			  (void *) cmd->argv[4],
			  (void *) cmd->argv[5]);
  } else if( cmd->argc == 5 ) {	/* 4 arguments */
    add_config_param_str (cmd->argv[0], 
			  4,
			  (void *) cmd->argv[1],
			  (void *) cmd->argv[2],
			  (void *) cmd->argv[3],
			  (void *) cmd->argv[4]);
  }

  return HANDLED (cmd);
}

/*
 * This is the main handler for this module according to which
 * the server allows/disallows user's anonymous session. If the
 * return value of `pck_core_handler' is FALSE, server denies
 * user's anonymous login.  
 */
MODRET 
pck_command_handler(cmd_rec *command_rec)
{
  char *user, *our_name, *anon_name = NULL;
  config_rec *conf = NULL;
  
  /*
   * "user", "our_name", "anon_name(=pck.remote.anonname)",
   * and "pass(=pck.remote.pass)" are defined here.  
   */
  user = (char *) get_param_ptr(command_rec->server->conf, C_USER, FALSE);
  conf = pck_resolve_user(command_rec->pool, &user, &our_name, &anon_name);

  pck.remote.anonname = anon_name;		/* anonymous access? */
  pck.remote.pass     = command_rec->arg;	/* store user's password */

  if( pck.debug & PCK_DEBUG_NONE ) {	/* NOT REACHED */
    log_pri (L_TRACE, "---: user     =[%s]", user);
    log_pri (L_TRACE, "---: pass     =[%s]", pck.remote.pass);
    log_pri (L_TRACE, "---: our_name =[%s]", our_name);
    log_pri (L_TRACE, "---: anon_name=[%s]", pck.remote.anonname);
  }
  
  /* invoke main module to allow/disallow login */
  if( !pck_core_handler(session.c, session.class, conf) ) {
    pck_respond_deny_message(command_rec);
    pck_dump(&pck);
    pck_free();
    return ERROR(command_rec);		/* deny login! */
  } else {
    pck_display_login_message();
    pck_respond_grant_message(command_rec);
    pck_dump(&pck);
    pck_free();
    return DECLINED(command_rec);	/* grant login? */
  }
}

static int
pwcheck_child_init ()
{
  (void) memset ((void *)&pck, (int)0, sizeof (pck));

  if(get_param_int(TOPLEVEL_CONF, "PCK", FALSE) == TRUE)
    pck.enable		= TRUE;	/* T(do password-check) F(not)    */

  pck.valid		= TRUE;	/* T(accept)            F(reject) */
  pck.debug		= FALSE;/* T(do debug)          F(not)    */
  pck.dot_check		= TRUE;	/* T(do dotcheck)       F(not)    */
  pck.no_more_check	= TRUE;	/* T(do nomorecheck)    F(not)    */

  /* 
   * Set next parameters to FALSE for fail-safe!
   * These settings would be overridden by configurations
   * in "proftpd.conf".
   */
  pck.remote.class_defined		= FALSE;/* True(define class)   */
  pck.display_login			= FALSE;/* True(display welcome)*/
  pck.server_dead.class			= FALSE;/* True(shutdown)       */
  pck.server_dead.host			= FALSE;/* True(shutdown)       */
  pck.goaway.class			= FALSE;/* True(goaway check)   */
  pck.goaway.host			= FALSE;/* True(goaway check)   */
  pck.reverse_dns.class			= FALSE;/* True(revDNS check)   */
  pck.reverse_dns.host			= FALSE;/* True(revDNS check)   */
  pck.reverse_dns.class_override	= FALSE;/* True(override)       */
  pck.reverse_dns.host_override		= FALSE;/* True(override)       */
  pck.matching_dns.class		= FALSE;/* True(matchDNS check) */
  pck.matching_dns.host			= FALSE;/* True(matchDNS check) */
  pck.matching_dns.class_override	= FALSE;/* True(override)       */
  pck.matching_dns.host_override	= FALSE;/* True(override)       */
  pck.dump				= FALSE;/* True(dump "pck")     */

  if (!(pck.denymsg	= get_param_ptr (TOPLEVEL_CONF, 
					    "PCKdenymessage",  FALSE)))
    pck.denymsg		= "Anonymous ftp REFUSED!!!";

  if (!(pck.grantmsg	= get_param_ptr (TOPLEVEL_CONF, 
					    "PCKgrantmessage", FALSE)))
    pck.grantmsg	= "Anonymous ftp GRANTED!!!";

  if (!(pck.local.name_server	= get_param_ptr (TOPLEVEL_CONF, 
						 "PCKnameserver", FALSE)))
    pck.local.name_server	= "localhost";
  
  pck.index		= "class";
  pck.remote.name	= NULL;
  pck.remote.ip		= NULL;
  pck.remote.anonname	= NULL;
  pck.remote.pass	= NULL;
  pck.local.ip		= NULL;
  pck.local.cname	= NULL;

  /*
   * Following members of "PCK" have their default path settings 
   * in "pwcheck.h". (see `#define' statements)
   *
   * NO DEFAULTS ARE GIVEN HERE!
   *
   * pck.path.display_login;
   * pck.path.invalidpw;
   * pck.path.errmsg;
   * pck.path.dump;
   * pck.server_dead.path;
   * pck.server_dead.class_path;
   * pck.server_dead.host_path;
   * pck.reverse_dns.class_path;
   * pck.reverse_dns.host_path;
   * pck.matching_dns.class_path;
   * pck.matching_dns.host_path;
   * pck.goaway.class_path;
   * pck.goaway.host_path;
   *
   */

  /* 
   * Default check mode. 
   * This is overridden by pck_init_checkmode().
   */
  (void) strcpy((char *)pck.check_mode[0], "file+dns");
  (void) strcpy((char *)pck.check_mode[1], "enforce");
  (void) strcpy((char *)pck.check_mode[2], "remove");
  (void) strcpy((char *)pck.check_mode[3], "*");
  (void) strcpy((char *)pck.check_mode[4], "");

  /* Default PCK check order */
  (void) strcpy((char *)pck.check_order[0], "dead");
  (void) strcpy((char *)pck.check_order[1], "goaway");
  (void) strcpy((char *)pck.check_order[2], "reverse");
  (void) strcpy((char *)pck.check_order[3], "match");
  (void) strcpy((char *)pck.check_order[4], "password");

  /* Default DNS check order */
  (void) strcpy((char *)pck.dns_order[0], "mx");
  (void) strcpy((char *)pck.dns_order[1], "a");

  /* Default hostname */
  (void) strcpy(hostname,"localhost");

  return 0;
}

static cmdtable pwcheck_command[] = {
  { PRE_CMD, C_PASS, G_NONE, pck_command_handler, FALSE, FALSE },
  { 0, NULL }
};

static conftable pwcheck_config[] = {

  /* Argument : only one Boolean */
  { "PCK",			pck_add_bool,		NULL },
  { "PCKdotcheck",		pck_add_bool,		NULL },
  { "PCKnomorecheck",		pck_add_bool,		NULL },

  /* Argument : one Boolean [+ one string] */
  { "PCKdebug",			pck_add_bool_str,	NULL },
  { "PCKdump",			pck_add_bool_str,	NULL },
  { "PCKdisplaylogin",		pck_add_bool_str,	NULL },
  { "PCKmsgprefix",		pck_add_bool_str,	NULL },

  /* Argument : one Boolean [+ one or two strings] */
  { "PCKserverdeadhost",	pck_add_bool_str2_ugly,	NULL },
  { "PCKserverdeadclass",	pck_add_bool_str2_ugly,	NULL },

  /* Argument : one Boolean [+ one or two strings] */
  { "PCKgoawayhost",		pck_add_bool_str2_ugly,	NULL },
  { "PCKgoawayclass",		pck_add_bool_str2_ugly,	NULL },

  /* Argument : one Boolean [+ one, two or three strings] */
  { "PCKreverseDNShost",	pck_add_bool_str3_ugly,	NULL },
  { "PCKreverseDNSclass",	pck_add_bool_str3_ugly,	NULL },

  /* Argument : one Boolean [+ one, two or three strings] */
  { "PCKmatchDNShost",		pck_add_bool_str3_ugly,	NULL },
  { "PCKmatchDNSclass",		pck_add_bool_str3_ugly,	NULL },

  /* Argument : four or five strings */
  { "PCKcheckmodehost",		pck_add_str45,		NULL },
  { "PCKcheckmodeclass",	pck_add_str45,		NULL },

  /* Argument : five strings */
  { "PCKcheckorder",		pck_add_str5,		NULL },

  /* Argument : one string */
  { "PCKdenymessage",		pck_add_str1,		NULL },
  { "PCKgrantmessage",		pck_add_str1,		NULL },
  { "PCKinvalidpwhost",		pck_add_str1,		NULL },
  { "PCKinvalidpwclass",	pck_add_str1,		NULL },
  { "PCKerrmsg",		pck_add_str1,		NULL },
  { "PCKnameserver",		pck_add_str1,		NULL },
  { "PCKlocalcname",		pck_add_str1,		NULL },

  /* Argument : two strings */
  { "PCKdnsorder",		pck_add_str2,		NULL },

  /* check use */
  { "PCKdebug1",		pck_add_bool_str_ugly,	NULL },
  { "PCKdebug2",		pck_add_bool_str2,	NULL },
  { "PCKdebug3",		pck_add_bool_str3,	NULL },
  { "PCKdebug4",		pck_add_str2,		NULL },
  { "PCKdebug5",		pck_add_str3,		NULL },
  { "PCKdebug6",		pck_add_str4,		NULL },

  { NULL,			NULL,			NULL }

};

module pwcheck_module = {
  NULL,			/* Pointer to the next module (Always NULL) */
  NULL,			/* Pointer to the previous module (Always NULL) */
  0x20,			/* Module API version 2.0 */
  "pwcheck",		/* Module name */
  pwcheck_config,	/* Configuration handler table */
  pwcheck_command,	/* Command handler table */
  NULL,			/* No authentication handler table */
  NULL,			/* No initialization function */
  pwcheck_child_init	/* Post-fork "child mode" initialization function */
};

#endif /* end of PW_CHECK_FILE || PW_CHECK_DNS */

/*
 * Modules of `password-check-kit' end here.
 */
